#include "pre.h"
#include "mbranch.h"
#include "mgroup.h"
#include "mcgate.h"
#include <malloc.h>
#include <stdio.h>
#include <string.h>
#include <string>
#include <cstring>

using namespace std;

#define NLAlloc(type,num) (type *)calloc(num,sizeof(type))


MGroup::MGroup(MosList *moshead_, NodeList *nodehead_, InoutList *inputhead_, InoutList *outputhead_)
{
	moshead = moshead_;
	nodehead = nodehead_;
	inputhead = inputhead_;
	outputhead = outputhead_;
	transhead = NULL;
//	wirehead = NULL;

	p_branch_list = NULL;
	n_branch_list = NULL;
	pass_transistor_list = NULL;
	left_transistor_list = NULL;
	cmos_gate_list = NULL;
//	special_gate_list = NULL;
}

MGroup::~MGroup()
{
	if(moshead) delete moshead;
	if(nodehead) delete nodehead;
	if(inputhead) delete inputhead;
	if(outputhead) delete outputhead;
	if(transhead) delete transhead;
//	if(wirehead) delete wirehead;

	if(p_branch_list) delete p_branch_list;
	if(n_branch_list) delete n_branch_list;
	if(pass_transistor_list) delete pass_transistor_list;
	if(left_transistor_list) delete left_transistor_list;
	if(cmos_gate_list) delete cmos_gate_list;
//	if(special_gate_list) delete special_gate_list;
}

int
MGroup::return_number(char *name)
{
        char *st = NLAlloc(char, strlen(name));
        strcpy(st, name);
        char *ret=NULL;
        ret = NLAlloc(char, strlen(st));
        int counter = 0;

        bool begin_f=false;
        bool end_f=false;

        int i=0;
        while(st[i]!='\0')
        {
                if(st[i]>=48 && st[i]<=57 && st[i]!='\0')
                {
                        begin_f = true;
                        ret[counter]=st[i];
                        counter+=1;
                }
                if(begin_f && st[i]!='\0' && ( st[i]<48 || st[i]>57 ))
                {
                        end_f = true;
                        break;
                }
                i+=1;
        }

        ret[counter] =  '\0';

        if(begin_f && !end_f)
        {
                return atoi(ret);
        }
        else
        {
                return -1;
        }
}

char *
MGroup::get_vdd(MosList *moshead)
{
   char *vdd;
   MosList *curr = NULL;
   for(curr=moshead; curr!=NULL; curr=curr->next)
   {
        if(curr->nptype)
        {
          vdd = NLAlloc(char,strlen(curr->b));
          strcpy(vdd, curr->b);
          break;
        }
   }
   return vdd;
}

char *
MGroup::get_gnd(MosList *moshead)
{
   char *gnd;
   MosList *curr = NULL;
   for(curr=moshead; curr!=NULL; curr=curr->next)
   {
        if(!curr->nptype)
        {
          gnd = NLAlloc(char,strlen(curr->b));
          strcpy(gnd, curr->b);
          break;
        }
   }
   return gnd;
}

int
MGroup::query_n_branch(char *name)
{
   int exist_flag = 0;
   if(n_branch_list==NULL)
	return 0;
   for(MBranch *cur_branch=n_branch_list; cur_branch!=NULL; cur_branch=cur_branch->next)
   {
	for(MosList *cur_mos=cur_branch->mos_list; cur_mos!=NULL; cur_mos=cur_mos->next)
	{
	   if(!strcmp(cur_mos->mosname, name))
	   {
		exist_flag = 1;
		return 1;
	   }
	}
   }
   
   if(!exist_flag)
	return 0;
}

int
MGroup::query_p_branch(char *name)
{
   int exist_flag = 0;
   if(p_branch_list==NULL)
	return 0;
   for(MBranch *cur_branch=p_branch_list; cur_branch!=NULL; cur_branch=cur_branch->next)
   {
        for(MosList *cur_mos=cur_branch->mos_list; cur_mos!=NULL; cur_mos=cur_mos->next)
        {
           if(!strcmp(cur_mos->mosname, name))
	   {
              exist_flag = 1;
	      return 1;
	   }
        }
   }

   if(!exist_flag)
        return 0;
}

MBranch *
MGroup::insert_branch(MBranch *head, MBranch *new_branch)
{
   MBranch *current;
   if(head==NULL)
   {
	head = new_branch;
	head->next = NULL;
   }
   else
   {
	current = new_branch;
	current->next = head->next;
	head->next = current;
   }
   return head;
}

// This fuction is added by Yang Lei, Oct 25, 2002.
void
MGroup::undelete_mosfet(MosList *mlist)
{
	for(MosList *curr=mlist; curr!=NULL; curr=curr->next)
	{
		for(MosList *cup=moshead; cup!=NULL; cup=cup->next)
		{
			if(!strcmp(curr->mosname, cup->mosname))
			{
				cup->in_branch = false;
				break;
			}
		}
	}
}
	


// Revised by Yang Lei, Oct 25, 2002.
// Add: if the branch recognition failed, recover the mosfets in moshead;
void
MGroup::generate_branch_list()
{
   char *gnd;
   gnd = get_gnd(moshead);
   char *vdd;
   vdd = get_vdd(moshead);

   NodeList *gnd_node, *vdd_node;

   for(NodeList *curr_node=nodehead; curr_node!=NULL; curr_node=curr_node->next)
   {
	if(!strcmp(curr_node->nodename, gnd))
	   gnd_node = curr_node;
	if(!strcmp(curr_node->nodename, vdd))
	   vdd_node = curr_node;
   }


   for(DeviceList *cur_device=gnd_node->devicelist; cur_device!=NULL; cur_device=cur_device->next)
   {
	bool branch_recog_success;
//	printf("%s \n", cur_device->devicename);
	MosList *first_n_mos;
	if(!query_n_branch(cur_device->devicename))	
	{
	   MBranch *new_branch;
	   new_branch = new MBranch(0, moshead, nodehead);	

	   for(MosList *cur_m=moshead; cur_m!=NULL; cur_m=cur_m->next)
	   {
		if(!strcmp(cur_device->devicename, cur_m->mosname))
		{
		   first_n_mos = cur_m;
//		   printf("A new branch from this mos: %s \n", cur_device->devicename);
		   break;
		}
	   }

	   branch_recog_success = new_branch->generate_branch(first_n_mos,0);
	   new_branch->back_up_mos_list();  // Revised
/*
	   for(MosList *cur_b=new_branch->mos_list; cur_b!=NULL; cur_b=cur_b->next)
		printf("%s  ", cur_b->mosname);
	   printf("\n");	
*/
	   if(branch_recog_success)
	   	   n_branch_list = insert_branch(n_branch_list, new_branch);
	   else
		   undelete_mosfet(new_branch->mos_list);
	}
   }

   for(DeviceList *cur_device=vdd_node->devicelist; cur_device!=NULL; cur_device=cur_device->next)
   {
     bool branch_recog_success;
        MosList *first_p_mos;
        if(!query_p_branch(cur_device->devicename))
        {
           MBranch *new_branch;
           new_branch = new MBranch(1, moshead, nodehead);

           for(MosList *cur_m=moshead; cur_m!=NULL; cur_m=cur_m->next)
           {
                if(!strcmp(cur_device->devicename, cur_m->mosname))
                {
                   first_p_mos = cur_m;
//	           printf("A new branch from this mos: %s \n", cur_device->devicename);
                   break;
                }
           }

           branch_recog_success = new_branch->generate_branch(first_p_mos,1);
	   	   new_branch->back_up_mos_list();  // Revised
/*
           for(MosList *cur_b=new_branch->mos_list; cur_b!=NULL; cur_b=cur_b->next)
                printf("%s  ", cur_b->mosname);
           printf("\n");
*/
		   if(branch_recog_success)
               p_branch_list = insert_branch(p_branch_list, new_branch);
		   else
		   	   undelete_mosfet(new_branch->mos_list);
	}
   }
}

void
MGroup::group_branch_list()
{
   if(p_branch_list->next != NULL)
   {
     for(MBranch *cur_branch=p_branch_list; cur_branch!=NULL; cur_branch=cur_branch->next)
     {
//		MosList *cu=cur_branch->mos_list;
//	        printf("Deal with: %s \n",cu->mosname); 
	MBranch *curr, *prev, *temp;
	prev=cur_branch;
	curr=cur_branch->next;	
	while(curr!=NULL)
	{
	   if(!strcmp(curr->pn_node, cur_branch->pn_node))
	   {
		for(MosList *cur_mos=curr->mos_list; cur_mos!=NULL; cur_mos=cur_mos->next)
		   cur_branch->mos_list = cur_branch->insert_mos(cur_branch->mos_list,cur_mos);
		for(BranchNodeList *cur_node=curr->node_list; cur_node!=NULL; cur_node=cur_node->next)
		   cur_branch->node_list = cur_branch->insert_node(cur_branch->node_list,cur_node->nodename);
		prev->next = curr->next;
		curr = prev->next;
/*
		temp = prev->next;
		prev = temp;
		curr = temp->next;
*/
	   }
	   else
	   {
	   prev=curr;
	   curr=curr->next;
	   }
	}
/*
	for(MBranch *cur_br=p_branch_list; cur_br!=NULL; cur_br=cur_br->next)
	{
		MosList *cu1=cur_br->mos_list;
	        printf("%s ",cu1->mosname); 
	}
	printf("\n");
*/
      }
   }



   if(n_branch_list->next != NULL)
   {
     for(MBranch *cur_branch=n_branch_list; cur_branch!=NULL; cur_branch=cur_branch->next)
     {
        MBranch *curr, *prev, *temp;
        prev=cur_branch;
        curr=cur_branch->next;
        while(curr!=NULL)
        {
           if(!strcmp(curr->pn_node, cur_branch->pn_node))
           {
                for(MosList *cur_mos=curr->mos_list; cur_mos!=NULL; cur_mos=cur_mos->next)
                   cur_branch->mos_list = cur_branch->insert_mos(cur_branch->mos_list,cur_mos);
		for(BranchNodeList *cur_node=curr->node_list; cur_node!=NULL; cur_node=cur_node->next)
                   cur_branch->node_list = cur_branch->insert_node(cur_branch->node_list,cur_node->nodename);
                prev->next = curr->next;
		curr = prev->next;
/*
                temp = prev->next;
                prev = temp;
                curr = temp->next;
*/
           }
           else
           {
           prev=curr;
           curr=curr->next;
           }
        }
/*
        for(MBranch *cur_br=n_branch_list; cur_br!=NULL; cur_br=cur_br->next)
        {
                MosList *cu1=cur_br->mos_list;
                printf("%s ",cu1->mosname);
        }
        printf("\n");
*/
      } 
   }    
}

MosList *
MGroup::insert_mos(MosList *head, MosList *mos)
{
   MosList *current;
   if(head==NULL)
   {
        head = NLAlloc(MosList,1);
        head->mosname = NLAlloc(char,strlen(mos->mosname));
        strcpy(head->mosname,mos->mosname);
        head->d = NLAlloc(char,strlen(mos->d));
        strcpy(head->d,mos->d);
        head->g = NLAlloc(char,strlen(mos->g));
        strcpy(head->g,mos->g);
        head->s = NLAlloc(char,strlen(mos->s));
        strcpy(head->s,mos->s);
        head->nptype = mos->nptype;
	head->is_instead = mos->is_instead;
        head->b = NLAlloc(char,strlen(mos->b));
        strcpy(head->b,mos->b);
        head->modelname = NLAlloc(char,strlen(mos->modelname));
        strcpy(head->modelname,mos->modelname);
        head->lvalue = mos->lvalue;
        head->wvalue = mos->wvalue;
        head->next = NULL;
   }
   else
   {
        current = NLAlloc(MosList,1);
        current->mosname = NLAlloc(char, strlen(mos->mosname));
        strcpy(current->mosname,mos->mosname);
        current->d = NLAlloc(char, strlen(mos->d));
        strcpy(current->d,mos->d);
        current->g = NLAlloc(char, strlen(mos->g));
        strcpy(current->g,mos->g);
        current->s = NLAlloc(char, strlen(mos->s));
        strcpy(current->s,mos->s);
        current->nptype = mos->nptype;
	current->is_instead = mos->is_instead;
        current->b = NLAlloc(char, strlen(mos->b));
        strcpy(current->b,mos->b);
        current->modelname = NLAlloc(char, strlen(mos->modelname));
        strcpy(current->modelname,mos->modelname);
        current->lvalue = mos->lvalue;
        current->wvalue = mos->wvalue;
        current->next = head->next;
        head->next = current;
   }
   return head;
}


void
MGroup::exchange_d_s(MosList *mos)
{
    char *temp=NLAlloc(char, strlen(mos->d));
    strcpy(temp, mos->d);
    mos->d = mos->s;
    mos->s = temp;
}

int
MGroup::query_is_pnnode(char *name)
{
    for(NodeList *cur_n=nodehead; cur_n!=NULL; cur_n=cur_n->next)
    {
	if(!strcmp(cur_n->nodename,name))
	{
	    return cur_n->pn_node;
	}
    }
}

int
MGroup::query_is_gnode(char *name)
{
    int p=0;
    for(NodeList *cur_n=nodehead; cur_n!=NULL; cur_n=cur_n->next)
    {
        if(!strcmp(cur_n->nodename,name))
        {
            for(DeviceList *cur_d=cur_n->devicelist; cur_d!=NULL; cur_d=cur_d->next)
            {
                if(cur_d->pintype==GP)
                {
                    p=1;
                }
            }
        }
    }
    return p;
}


void
MGroup::generate_pass_left_list()
{
    int is_pn_1, is_g_1;
    int is_pn_2, is_g_2;

    for(MosList *curr_mos=moshead; curr_mos!=NULL; curr_mos=curr_mos->next)
    {
//	if(!query_n_branch(curr_mos->mosname) && !query_p_branch(curr_mos->mosname))
	if(!curr_mos->in_branch)  // Revised by Yang Lei, Oct 25, 2002.
	{
	    if(query_is_pnnode(curr_mos->d) && query_is_gnode(curr_mos->s))
			pass_transistor_list = insert_mos(pass_transistor_list,curr_mos);
	    else if(query_is_pnnode(curr_mos->s) && query_is_gnode(curr_mos->d))
	    {
	        exchange_d_s(curr_mos);
			pass_transistor_list = insert_mos(pass_transistor_list,curr_mos);
	    }
        else
            left_transistor_list = insert_mos(left_transistor_list, curr_mos);
	}
    }	
}


TransList *
MGroup::insert_trans(TransList *head,const char *name,  char *input, char *output, char *p_g, char *n_g)
{
    TransList *current;
    if(head==NULL)
    {
	head = NLAlloc(TransList, 1);
//	head->type = type;
        head->transname = NLAlloc(char, strlen(name));
	strcpy(head->transname, name);
	head->input = NLAlloc(char, strlen(input));
	strcpy(head->input, input);
	head->output = NLAlloc(char, strlen(output));
	strcpy(head->output, output);
	head->p_g = NLAlloc(char, strlen(p_g));
	strcpy(head->p_g, p_g);
	head->n_g = NLAlloc(char, strlen(n_g));
	strcpy(head->n_g, n_g);
	head->next = NULL;
    }
    else
    {
	current = NLAlloc(TransList,1);
//	current->type = type;
        current->transname = NLAlloc(char, strlen(name));
	strcpy(current->transname, name);
	current->input = NLAlloc(char, strlen(input));
	strcpy(current->input, input);
	current->output = NLAlloc(char, strlen(output));
	strcpy(current->output, output);
	current->p_g = NLAlloc(char, strlen(p_g));
	strcpy(current->p_g, p_g);
	current->n_g = NLAlloc(char, strlen(n_g));	
	strcpy(current->n_g, n_g);
	current->next = head->next;
	head->next = current;
    }
    return head;
}

void
MGroup::undelete_mos_in_pass()
{
	for(MosList *curr=pass_transistor_list; curr!=NULL; curr=curr->next)
		curr->in_branch = false;
}

// Revised by Yang Lei, Oct 25, 2002.
// Add: Only handle the double type transmission gate, discarding the single one.
void
MGroup::generate_trans_list()
{
	string s1 = "trans";
	char s2[20];
	int counter = 0;

	undelete_mos_in_pass();
    for(MosList *curr_mos=pass_transistor_list; curr_mos!=NULL; curr_mos=curr_mos->next)
    {
	 	if(!curr_mos->in_branch)
		{
			for(MosList *curr=curr_mos->next; curr!=NULL; curr=curr->next)
			{
				if(!curr->in_branch)
				{
	    		  if((!strcmp(curr_mos->d, curr->d) && !strcmp(curr_mos->s, curr->s)) || (!strcmp(curr_mos->d, curr->s) && !strcmp(curr_mos->s,curr->d)))
				  {

					sprintf(s2, "%d", counter); 
					string s3 = s1 + s2; 
					counter += 1;

					if(curr_mos->nptype==1 && curr->nptype==0)
					{
		    			transhead = insert_trans(transhead,s3.c_str(),curr_mos->d, curr_mos->s, curr_mos->g, curr->g);
						curr_mos->in_branch = true;
						curr->in_branch = true;
						break;
					}
					if(curr_mos->nptype==0 && curr->nptype==1)
					{
		    			transhead = insert_trans(transhead,s3.c_str(),curr_mos->d, curr_mos->s, curr->g, curr_mos->g);
						curr_mos->in_branch = true;
						curr->in_branch = true;
						break;
					}
//						counter += 1;
				  }
				}
			}
		}
	}

	num_of_trans = counter;

	for(MosList *cl = pass_transistor_list; cl!=NULL; cl=cl->next)
	{
		if(!cl->in_branch)
			left_transistor_list = insert_mos(left_transistor_list, cl);
	}
}
	    
int
MGroup::counter(MBranch *branch)
{
   int num = 0;
   for(MosList *cur_mos=branch->mos_list; cur_mos!=NULL; cur_mos=cur_mos->next)
   {
	num = num+1;
   }
   return num;
}

void
MGroup::generate_num_for_branch()   // Revised, this must appear in Main function.
{
	for(MBranch *cu=n_branch_list; cu!=NULL; cu=cu->next)
		cu->num_of_mos = counter(cu);
	for(MBranch *cu=p_branch_list; cu!=NULL; cu=cu->next)
		cu->num_of_mos = counter(cu);
}

char *
MGroup::generate_add(char g1, char g2)
{
    char *new_string = NLAlloc(char, 6);
    strcpy(new_string, "a | b");
    new_string[0] = g1;
    new_string[1] = ' ';
    new_string[2] = '|';
    new_string[3] = ' ';
    new_string[4] = g2;
    new_string[5] = '\0';
    return new_string;
}

char *
MGroup::generate_add(char *g1, char g2)
{
    char *new_string = NLAlloc(char, strlen(g1)+4);
    strcpy(new_string,g1);
    int k = strlen(new_string);
    new_string[k] = ' ';
    new_string[k+1] = '|';
    new_string[k+2] = ' ';
    new_string[k+3] = g2;
    new_string[k+4] = '\0';
    return new_string;
}

char *
MGroup::generate_add(char *g1, char *g2)
{
    const char *add = " | ";
    int len1 = strlen(g1);
    int len2 = strlen(g2);
    char *new_string = NLAlloc(char, len1+len2+3);
    strcpy(new_string,g1);
    strcat(new_string, add);
    strcat(new_string, g2);
    return new_string;
}

char *
MGroup::generate_mult(char g1, char g2)
{
    char *new_string = NLAlloc(char, 6);
    strcpy(new_string, "a & b");
    new_string[0] = g1;
    new_string[1] = ' ';
    new_string[2] = '&';
    new_string[3] = ' ';
    new_string[4] = g2;
    new_string[5] = '\0';
    return new_string;
}

char *
MGroup::generate_mult(char *g1, char g2)
{
    char *new_string = NLAlloc(char, strlen(g1)+4);
    strcpy(new_string,g1);
    int k = strlen(new_string);
    new_string[k] = ' ';
    new_string[k+1] = '&';
    new_string[k+2] = ' ';
    new_string[k+3] = g2;
    new_string[k+4] = '\0';
    return new_string;
}

char *
MGroup::generate_mult(char *g1, char *g2)
{
    const char *mult = " & ";
    int len1 = strlen(g1);
    int len2 = strlen(g2);
    char *new_string = NLAlloc(char, len1+len2+3);
    strcpy(new_string,g1);
    strcat(new_string,mult);
    strcat(new_string,g2);
    return new_string;
}


char *
MGroup::generate_parenthes(char *g)
{
    const char *left = "(";
    const char *right = ")";
    char *new_string=NLAlloc(char,strlen(g)+2);
    strcpy(new_string,left);
    strcat(new_string, g);
    strcat(new_string, right);
    return new_string;
}

char *
MGroup::generate_mark(char *g)
{
    const char *mark = "~";
    char *new_string=NLAlloc(char, strlen(g)+1);
    strcpy(new_string,mark);
    strcat(new_string, g);
    return new_string;
}

int
MGroup::is_series_node(char *nodename, MBranch *curr_branch)
{
    int flag = 0;
    int counter=0;
    MosList *mos=curr_branch->mos_list;
    for(; mos!=NULL; mos=mos->next)
    {
	if(!strcmp(nodename, mos->d) || !strcmp(nodename, mos->s))
	    counter += 1;
    }

    if(counter==2)
	flag = 1;

    return flag;
}

char 
MGroup::change_g(MBranch *n_branch, MosList *mos1, MosList *mos2, char turn, int add_or_flag)
{
    char local = turn;

    if(mos1->is_instead && mos2->is_instead)
    {
	if(add_or_flag)
	    mos1->g = generate_add(mos1->g, mos2->g);
	else
	    mos1->g = generate_mult(mos1->g, mos2->g);
    }

    if(mos1->is_instead && !mos2->is_instead)
    {
	n_branch->name_list = n_branch->insert_name(n_branch->name_list, mos2->g, local);
	if(add_or_flag)
	    mos1->g = generate_add(mos1->g, local);
	else
	    mos1->g = generate_mult(mos1->g, local);
	local+=1;
    }

    if(!mos1->is_instead && mos2->is_instead)
    {
	n_branch->name_list = n_branch->insert_name(n_branch->name_list, mos1->g, local);
	if(add_or_flag)
	    mos1->g = generate_add(mos2->g, local);
	else
	    mos1->g = generate_mult(mos2->g, local);
	local+=1;
	mos1->is_instead = 1;
    }

    if(!mos1->is_instead && !mos2->is_instead)
    {
	n_branch->name_list = n_branch->insert_name(n_branch->name_list, mos1->g, local);
	n_branch->name_list = n_branch->insert_name(n_branch->name_list, mos2->g, local+1);
	char local1 = local+1;
	if(add_or_flag)
	    mos1->g = generate_add(local, local1);
	else
	    mos1->g = generate_mult(local, local1);
	local+=2;
	mos1->is_instead = 1;
    }
    
    return local;
}

// Insert the gatename in the gatenamelist, added by Yang Lei, Nov 20, 2002
GateNameList *
MGroup::insert_gatename(GateNameList *head, char *name){
  GateNameList *current;
  if(head==NULL)
  {
        head=NLAlloc(GateNameList,1);
        head->name = NLAlloc(char,strlen(name));
        strcpy(head->name,name);
        head->next = NULL;
  }
  else {
        current=NLAlloc(GateNameList,1);
        current->name = NLAlloc(char,strlen(name));
        strcpy(current->name,name);
        current->next = head->next;
        head->next = current;
  }
  return head;
}

RelationList *
MGroup::insert_relation(RelationList *head, int type, GateNameList *gnlist){
  RelationList *current;
  if(head==NULL)
  {
        head=NLAlloc(RelationList,1);
		head->type = type;
		head->gatename_list = gnlist;
        head->next = NULL;
  }
  else {
        current=NLAlloc(RelationList,1);
		current->type = type;
		current->gatename_list = gnlist;
        current->next = head->next;
        head->next = current;
  }
  return head;
}

RelationList *
MGroup::query_alone_list(RelationList *rlist)
{
	if(rlist==NULL)
		return NULL;
	else
	{
		for(RelationList *curr=rlist; curr!=NULL; curr=curr->next)
		{
			if(curr->type==0)
				return curr;
		}
	}
	return NULL;
}

// End adding.

    

char
MGroup::parallel_recognition(MBranch *n_branch, char turn)
{
    char retur=turn;
    for(MosList *curr_mos=n_branch->mos_list; curr_mos!=NULL; curr_mos=curr_mos->next)
    {
	int flag =0;
	MosList *curr_pointer, *prev_pointer;
	prev_pointer = curr_mos;
	curr_pointer = curr_mos->next;

	GateNameList *gnlist = NULL; // New added line, Nov 20,2002

	while(curr_pointer!=NULL)
	{
	    if((!strcmp(curr_mos->d, curr_pointer->d) && !strcmp(curr_mos->s, curr_pointer->s)) || (!strcmp(curr_mos->d, curr_pointer->s) && !strcmp(curr_mos->s, curr_pointer->d)))
	    {
		flag = 1;

// Added by Lei Yang, Nov 20,2002
		if(!curr_mos->is_instead)
			gnlist = insert_gatename(gnlist, curr_mos->g);		
		if(!curr_pointer->is_instead)
			gnlist = insert_gatename(gnlist, curr_pointer->g);		
// End


		retur = change_g(n_branch, curr_mos, curr_pointer, retur, 1);

		prev_pointer->next = curr_pointer->next;
		curr_pointer = prev_pointer->next;
	    }
	    else
	    {
		prev_pointer=curr_pointer;
		curr_pointer=curr_pointer->next;
	    }
	}
// Added by Lei Yang, Nov, 20,2002
		if(gnlist==NULL)
			;
		else
		{
			if(gnlist->next == NULL)
			{
				RelationList *here = query_alone_list(n_branch->relation_list);
				if(here==NULL)
				{
					n_branch->relation_list = insert_relation(n_branch->relation_list,0,gnlist);
				}
				else
				{
					here->gatename_list = insert_gatename(here->gatename_list,gnlist->name);
				}
			}
			else
				n_branch->relation_list = insert_relation(n_branch->relation_list,1,gnlist);
		}
// End
	if(flag){
	curr_mos->g=generate_parenthes(curr_mos->g);
	}
    }
    return retur;
}


char
MGroup::series_recognition(MBranch *n_branch, char turn)
{
    char *GND;
	char *VDD;
    GND = get_gnd(moshead);
    VDD = get_vdd(moshead);
    char retur=turn;
    for(MosList *curr_mos=n_branch->mos_list; curr_mos!=NULL; curr_mos=curr_mos->next)
    {
        int flag =0;
        MosList *curr_pointer, *prev_pointer;
        prev_pointer = curr_mos;
        curr_pointer = curr_mos->next;

		GateNameList *gnlist = NULL; // New added line, Nov 20,2002

        while(curr_pointer!=NULL)
        {
	    if((!strcmp(curr_mos->d, curr_pointer->d) && strcmp(curr_mos->d, GND) && strcmp(curr_mos->d, VDD) && strcmp(curr_mos->d, n_branch->pn_node) && strcmp(curr_mos->s, curr_pointer->s) && is_series_node(curr_mos->d,n_branch)) 
			|| (!strcmp(curr_mos->s, curr_pointer->s) && strcmp(curr_mos->s, GND) && strcmp(curr_mos->s, VDD) && strcmp(curr_mos->s, n_branch->pn_node) && strcmp(curr_mos->d, curr_pointer->d) && is_series_node(curr_mos->s, n_branch)) 	
			|| (!strcmp(curr_mos->d, curr_pointer->s) && strcmp(curr_mos->d, GND) && strcmp(curr_mos->d, VDD) && strcmp(curr_mos->d, n_branch->pn_node) && strcmp(curr_mos->s, curr_pointer->d) && is_series_node(curr_mos->d, n_branch)) 
			|| (!strcmp(curr_mos->s, curr_pointer->d) && strcmp(curr_mos->s, GND) && strcmp(curr_mos->s, VDD) && strcmp(curr_mos->s, n_branch->pn_node) && is_series_node(curr_mos->s, n_branch) && strcmp(curr_mos->d, curr_pointer->s)))	
            {
			flag=1;
	        if(!strcmp(curr_mos->d, curr_pointer->d) && strcmp(curr_mos->s, curr_pointer->s))
		     strcpy(curr_mos->d, curr_pointer->s);
	        else if(!strcmp(curr_mos->s, curr_pointer->s) && strcmp(curr_mos->d, curr_pointer->d))
		     strcpy(curr_mos->s, curr_pointer->d);
	        else if(!strcmp(curr_mos->d, curr_pointer->s) && strcmp(curr_mos->s, curr_pointer->d)) 
	       	     strcpy(curr_mos->d, curr_pointer->d);
	        else if(!strcmp(curr_mos->s, curr_pointer->d) && strcmp(curr_mos->d, curr_pointer->s))
		     strcpy(curr_mos->s, curr_pointer->s);
			else
					;

// Added by Lei Yang, Nov 20,2002
		if(!curr_mos->is_instead)
			gnlist = insert_gatename(gnlist, curr_mos->g);		
		if(!curr_pointer->is_instead)
			gnlist = insert_gatename(gnlist, curr_pointer->g);		
// End


				retur = change_g(n_branch, curr_mos, curr_pointer, retur, 0);

                prev_pointer->next = curr_pointer->next;
                curr_pointer = prev_pointer->next;
            }
            else
            {
                prev_pointer=curr_pointer;
                curr_pointer=curr_pointer->next;
            }
        }
// Added by Lei Yang, Nov, 20,2002
		if(gnlist==NULL)
			;
		else
		{
			if(gnlist->next == NULL)
			{
				RelationList *here = query_alone_list(n_branch->relation_list);
				if(here==NULL)
				{
					n_branch->relation_list = insert_relation(n_branch->relation_list,0,gnlist);
				}
				else
				{
					here->gatename_list = insert_gatename(here->gatename_list,gnlist->name);
				}
			}
			else
				n_branch->relation_list = insert_relation(n_branch->relation_list,1,gnlist);
		}
// End
	if(flag){
	curr_mos->g=generate_parenthes(curr_mos->g);
	}
    }
    return retur;
}


void
MGroup::n_branch_recognition()
{
    const char *model_name = "N_Model_";
    char model_add = 'a';
    for(MBranch *curr_n=n_branch_list; curr_n!=NULL; curr_n=curr_n->next)
    {
		if(curr_n->num_of_mos == 1)
		{
			MosList *mos;
			mos=curr_n->mos_list;
			curr_n->name_list = curr_n->insert_name(curr_n->name_list, mos->g, 'a');
			mos->g=generate_mark("a");	
			curr_n->headername = NLAlloc(char, 4);
			strcpy(curr_n->headername, "INV");
		}
		else if(curr_n->num_of_mos>1 && curr_n->num_of_mos<=6)
		{
			char input_name;
			input_name='a';
			MosList *list=curr_n->mos_list;
			while(counter(curr_n)!=1)
			{
			    input_name = parallel_recognition(curr_n,input_name); 

			    input_name = series_recognition(curr_n, input_name);
			}

			list->g = generate_mark(list->g);

			MosList *currmos=curr_n->mos_list;
                    if(!strcmp(currmos->g, "~(a | b)"))
                    {
                        curr_n->headername = NLAlloc(char, 6);
                        strcpy(curr_n->headername, "NOR2");
                    }
                    else if(!strcmp(currmos->g, "~(a & b)"))
                    {
                        curr_n->headername = NLAlloc(char, 6);
                        strcpy(curr_n->headername, "NAND2");
                    }
                    else if(!strcmp(currmos->g, "~(a | b | c)") || !strcmp(currmos->g, "~((a | b) | c)"))
                    {
                        curr_n->headername = NLAlloc(char, 6);
                        strcpy(curr_n->headername, "NOR3");
                    }
                    else if(!strcmp(currmos->g, "~(a & b & c)") || !strcmp(currmos->g, "~((a & b) & c)"))
                    {
                        curr_n->headername = NLAlloc(char, 6);
                        strcpy(curr_n->headername, "NAND3");
                    }
                    else if(!strcmp(currmos->g, "~((a | b) & c)"))
                    {
                        curr_n->headername = NLAlloc(char, 6);
                        strcpy(curr_n->headername, "OAI12");
                    }
                    else if(!strcmp(currmos->g, "~((a & b) | c)"))
                    {
                        curr_n->headername = NLAlloc(char, 6);
                        strcpy(curr_n->headername, "AOI12");
                    }
                    else if(!strcmp(currmos->g, "~((a | b) | (c | d))") || !strcmp(currmos->g, "~((a | b | c) | d)") || !strcmp(currmos->g, "~(a | b | c | d)"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "NOR4");
                    }
                    else if(!strcmp(currmos->g, "~((a & b) & (c & d))") || !strcmp(currmos->g, "~((a & b & c) & d)") || !strcmp(currmos->g, "~(a & b & c & d)"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "NAND4");
                    }
                    else if(!strcmp(currmos->g, "~((a & b) | (c & d))"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "AOI22");
                    }
                    else if(!strcmp(currmos->g, "~((a | b) & (c | d))"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "OAI22");
                    }
//                    else if(!strcmp(currmos->g, "~((a & b & c) | d)"))
                    else if(!strcmp(currmos->g, "~((a & b & c) | d)") || !strcmp(currmos->g, "~(((a & b) & c) |d)"))   // Revised by Yang Lei, Oct 26.
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "AOI13");
                    }
                    else if(!strcmp(currmos->g, "~((a | b | c) & d)") || !strcmp(currmos->g, "(((a | b) | c) & d)"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "OAI13");
                    }
                    else if(!strcmp(currmos->g, "~(((a | b) & c) | d)"))
                    {
                            curr_n->headername = NLAlloc(char, 8);
                            strcpy(curr_n->headername, "OAOI112");
		    		}
//                    else if(!strcmp(currmos->g, "~((a & b) | c | d)"))
                    else if(!strcmp(currmos->g, "~((a | b) | (c & d))"))
                    {
                            curr_n->headername = NLAlloc(char, 8);
                            strcpy(curr_n->headername, "AOOI112");
                    }
//                    else if(!strcmp(currmos->g, "~((a | b) & c & d)"))
                    else if(!strcmp(currmos->g, "~((a | b) & c & d)") || !strcmp(currmos->g, "~((a | b) & (c & d))") || !strcmp(currmos->g, "~(((a | b) & c) & d)"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "OAAI112");
                    }
                    else if(!strcmp(currmos->g, "~(((a & b) | c) & d)"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "AOAI112");
                    }

// The following is added by Yang Lei, Oct, 26, 2002

                    else if(!strcmp(currmos->g, "~((a & b) | (c & d & e))") || !strcmp(currmos->g, "~((a & b) | ((c & d) & e))") || !strcmp(currmos->g,"~((a & b & c) | (d & e))") || !strcmp(currmos->g, "~(((a & b) & e) | (c & d))"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "AOI23");
                    }
                    else if(!strcmp(currmos->g, "~((a | b) | (c & d & e))") || !strcmp(currmos->g, "~((a | b) | ((c & d) & e))"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "AOI311");
                    }
                    else if(!strcmp(currmos->g, "~((a & b) | (c & d) | e)") || !strcmp(currmos->g, "~((a & b) | e | (c & d))"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "AOI221");
                    }
                    else if(!strcmp(currmos->g, "~((a | b | c) & (d | e))") || !strcmp(currmos->g, "~((a | b) & (c | d | e))"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "OAI23");
                    }
                    else if(!strcmp(currmos->g, "~((a | b | c) & d & e)") || !strcmp(currmos->g, "~(((a | b | c) & d) & e)") || !strcmp(currmos->g,"~((a | b | c) & (d & e))") || !strcmp(currmos->g, "~((d & e) & (a | b | c))"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "OAI311");
                    }
                    else if(!strcmp(currmos->g, "~((a | b) & (c | d) & e)") || !strcmp(currmos->g, "~((a | b) & e & (c | d))") || !strcmp(currmos->g,"~(((a | b) & (c | d)) & e)") || !strcmp(currmos->g, "~(((a | b) & e) & (c | d))"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "OAI221");
                    }

                    else if(!strcmp(currmos->g, "~(((a | b) & c) | (d & e))") || !strcmp(currmos->g, "~(((a | b) & e) | (c & d))"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "OAOI212");
                    }
                    else if(!strcmp(currmos->g, "~(((a | b) & e) | (c | d))") || !strcmp(currmos->g, "~((a | b) | ((c | d) & e))"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "OAOI2111");
                    }
                    else if(!strcmp(currmos->g, "~((a | b) & ((c & d) | e))"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "AOAI212");
                    }
                    else if(!strcmp(currmos->g, "~(((a & b) | e) & (c & d))") || !strcmp(currmos->g, "~((a & b) & ((c & d) | e))") || !strcmp(currmos->g,"~(((a & b) | c) & d & e)") || !strcmp(currmos->g, "~((((a & b) | c) & d) & e)"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "AOAI2111");
                    }

                    else if(!strcmp(currmos->g, "~((a & b & c) | (d & e & f))") || !strcmp(currmos->g, "~(((a & b) & f) | ((c & d & e))") || !strcmp(currmos->g,"~((a & b & c) | ((d & e) & f))") || !strcmp(currmos->g, "~(((a & b) & e) | (c & d) & f))"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "AOI33");
                    }
                    else if(!strcmp(currmos->g, "~((a & b &c) | (d & e) | f)") || !strcmp(currmos->g, "~(((a & b) & e) | (c & d) | f)") || !strcmp(currmos->g,"~((a & b) | (c & d & e) | f)") || !strcmp(currmos->g, "~((a & b) | ((c & d) & e) | f)") || !strcmp(currmos->g, "~((a & b & c) | f | (d & e))") || !strcmp(currmos->g, "~(((a & b) & e) | f | (c & d))") || !strcmp(currmos->g, "~((a & b) | f | (c & d & e))") || !strcmp(currmos->g, "~((a & b) | f | ((c & d) & e))"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "AOI321");
                    }
                    else if(!strcmp(currmos->g, "~((a & b) | (c & d) | (e & f))"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "AOI222");
                    }
                    else if(!strcmp(currmos->g, "~((a | b) & (c | d) & (e | f))"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "OAI222");
                    }
                    else if(!strcmp(currmos->g, "~((a | b | c) & (d & e & f))"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "OAI33");
                    }
                    else if(!strcmp(currmos->g, "~((a | b | c) & (d | e) & f)") || !strcmp(currmos->g, "~((a | b) & (c | d | e) & f)") || !strcmp(currmos->g,"~((a | b | c) & f (d | e))") || !strcmp(currmos->g, "~((a | b) & f & (c | d | e))"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "OAI321");
                    }
		    		else if(!strcmp(currmos->g, "~a"))
		    		{
						;
		    		}
		    		else
		    		{           
	                	char input_name;
                        input_name='a';
                        MosList *list=curr_n->mos_list;
                        while(counter(curr_n)!=1)
                        {
                            input_name = parallel_recognition(curr_n,input_name);

                            input_name = series_recognition(curr_n, input_name);
                        }
                        list->g = generate_mark(list->g);
						curr_n->greatthan4 = 1;
                        curr_n->headername = NLAlloc(char, strlen(model_name)+2);
                        strcpy(curr_n->headername, model_name);
                        int k = strlen(curr_n->headername);
                        curr_n->headername[k]=model_add;
                        curr_n->headername[k+1]='\0';
                        model_add+=1;
						printf("This n_branch contains more than 7 mosfets, can't handle\n");
						printf("Branch name = %s, Logic function is: \n", curr_n->headername, list->g);
		     		}
		}
    }
}
				
void
MGroup::p_branch_recognition()
{
    const char *model_name = "P_Model_";
    char model_add = 'a';
    for(MBranch *curr_n=p_branch_list; curr_n!=NULL; curr_n=curr_n->next)
    {
		if(curr_n->num_of_mos == 1)
		{
			MosList *mos;
			mos=curr_n->mos_list;
			curr_n->name_list = curr_n->insert_name(curr_n->name_list, mos->g, 'a');
			mos->g=generate_mark("a");	
			curr_n->headername = NLAlloc(char, 4);
			strcpy(curr_n->headername, "INV");
		}
		else if(curr_n->num_of_mos>1 && curr_n->num_of_mos<=6)
		{
			char input_name;
			input_name='a';
			MosList *list=curr_n->mos_list;
			while(counter(curr_n)!=1)
			{
			    input_name = parallel_recognition(curr_n,input_name); 

			    input_name = series_recognition(curr_n, input_name);

			}

			list->g = generate_mark(list->g);

			MosList *currmos=curr_n->mos_list;
                    if(!strcmp(currmos->g, "~(a | b)"))
                    {
                        curr_n->headername = NLAlloc(char, 6);
                        strcpy(curr_n->headername, "NOR2");
                    }
                    else if(!strcmp(currmos->g, "~(a & b)"))
                    {
                        curr_n->headername = NLAlloc(char, 6);
                        strcpy(curr_n->headername, "NAND2");
                    }
                    else if(!strcmp(currmos->g, "~(a | b | c)") || !strcmp(currmos->g, "~((a | b) | c)"))
                    {
                        curr_n->headername = NLAlloc(char, 6);
                        strcpy(curr_n->headername, "NOR3");
                    }
                    else if(!strcmp(currmos->g, "~(a & b & c)") || !strcmp(currmos->g, "~((a & b) & c)"))
                    {
                        curr_n->headername = NLAlloc(char, 6);
                        strcpy(curr_n->headername, "NAND3");
                    }
                    else if(!strcmp(currmos->g, "~((a | b) & c)"))
                    {
                        curr_n->headername = NLAlloc(char, 6);
                        strcpy(curr_n->headername, "OAI12");
                    }
                    else if(!strcmp(currmos->g, "~((a & b) | c)"))
                    {
                        curr_n->headername = NLAlloc(char, 6);
                        strcpy(curr_n->headername, "AOI12");
                    }
                    else if(!strcmp(currmos->g, "~((a | b) | (c | d))") || !strcmp(currmos->g, "~((a | b | c) | d)") || !strcmp(currmos->g, "~(a | b | c | d)"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "NOR4");
                    }
                    else if(!strcmp(currmos->g, "~((a & b) & (c & d))") || !strcmp(currmos->g, "~((a & b & c) & d)") || !strcmp(currmos->g, "~(a & b & c & d)"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "NAND4");
                    }
                    else if(!strcmp(currmos->g, "~((a & b) | (c & d))"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "AOI22");
                    }
                    else if(!strcmp(currmos->g, "~((a | b) & (c | d))"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "OAI22");
                    }
//                    else if(!strcmp(currmos->g, "~((a & b & c) | d)"))
                    else if(!strcmp(currmos->g, "~((a & b & c) | d)") || !strcmp(currmos->g, "~(((a & b) & c) |d)"))   // Revised by Yang Lei, Oct 26.
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "AOI13");
                    }
                    else if(!strcmp(currmos->g, "~((a | b | c) & d)") || !strcmp(currmos->g, "(((a | b) | c) & d)"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "OAI13");
                    }
                    else if(!strcmp(currmos->g, "~(((a | b) & c) | d)"))
                    {
                            curr_n->headername = NLAlloc(char, 8);
                            strcpy(curr_n->headername, "OAOI112");
		    		}
//                    else if(!strcmp(currmos->g, "~((a & b) | c | d)"))
                    else if(!strcmp(currmos->g, "~((a | b) | (c & d))"))
                    {
                            curr_n->headername = NLAlloc(char, 8);
                            strcpy(curr_n->headername, "AOOI112");
                    }
//                    else if(!strcmp(currmos->g, "~((a | b) & c & d)"))
                    else if(!strcmp(currmos->g, "~((a | b) & c & d)") || !strcmp(currmos->g, "~((a | b) & (c & d))") || !strcmp(currmos->g, "~(((a | b) & c) & d)"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "OAAI112");
                    }
                    else if(!strcmp(currmos->g, "~(((a & b) | c) & d)"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "AOAI112");
                    }

// The following is added by Yang Lei, Oct, 26, 2002

                    else if(!strcmp(currmos->g, "~((a & b) | (c & d & e))") || !strcmp(currmos->g, "~((a & b) | ((c & d) & e))") || !strcmp(currmos->g,"~((a & b & c) | (d & e))") || !strcmp(currmos->g, "~(((a & b) & e) | (c & d))"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "AOI23");
                    }
                    else if(!strcmp(currmos->g, "~((a | b) | (c & d & e))") || !strcmp(currmos->g, "~((a | b) | ((c & d) & e))"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "AOI311");
                    }
                    else if(!strcmp(currmos->g, "~((a & b) | (c & d) | e)") || !strcmp(currmos->g, "~((a & b) | e | (c & d))"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "AOI221");
                    }
                    else if(!strcmp(currmos->g, "~((a | b | c) & (d | e))") || !strcmp(currmos->g, "~((a | b) & (c | d | e))"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "OAI23");
                    }
                    else if(!strcmp(currmos->g, "~((a | b | c) & d & e)") || !strcmp(currmos->g, "~(((a | b | c) & d) & e)") || !strcmp(currmos->g,"~((a | b | c) & (d & e))") || !strcmp(currmos->g, "~((d & e) & (a | b | c))"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "OAI311");
                    }
                    else if(!strcmp(currmos->g, "~((a | b) & (c | d) & e)") || !strcmp(currmos->g, "~((a | b) & e & (c | d))") || !strcmp(currmos->g,"~(((a | b) & (c | d)) & e)") || !strcmp(currmos->g, "~(((a | b) & e) & (c | d))"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "OAI221");
                    }

                    else if(!strcmp(currmos->g, "~(((a | b) & c) | (d & e))") || !strcmp(currmos->g, "~(((a | b) & e) | (c & d))"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "OAOI212");
                    }
                    else if(!strcmp(currmos->g, "~(((a | b) & e) | (c | d))") || !strcmp(currmos->g, "~((a | b) | ((c | d) & e))"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "OAOI2111");
                    }
                    else if(!strcmp(currmos->g, "~((a | b) & ((c & d) | e))"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "AOAI212");
                    }
                    else if(!strcmp(currmos->g, "~(((a & b) | e) & (c & d))") || !strcmp(currmos->g, "~((a & b) & ((c & d) | e))") || !strcmp(currmos->g,"~(((a & b) | c) & d & e)") || !strcmp(currmos->g, "~((((a & b) | c) & d) & e)"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "AOAI2111");
                    }

                    else if(!strcmp(currmos->g, "~((a & b & c) | (d & e & f))") || !strcmp(currmos->g, "~(((a & b) & f) | ((c & d & e))") || !strcmp(currmos->g,"~((a & b & c) | ((d & e) & f))") || !strcmp(currmos->g, "~(((a & b) & e) | (c & d) & f))"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "AOI33");
                    }
                    else if(!strcmp(currmos->g, "~((a & b &c) | (d & e) | f)") || !strcmp(currmos->g, "~(((a & b) & e) | (c & d) | f)") || !strcmp(currmos->g,"~((a & b) | (c & d & e) | f)") || !strcmp(currmos->g, "~((a & b) | ((c & d) & e) | f)") || !strcmp(currmos->g, "~((a & b & c) | f | (d & e))") || !strcmp(currmos->g, "~(((a & b) & e) | f | (c & d))") || !strcmp(currmos->g, "~((a & b) | f | (c & d & e))") || !strcmp(currmos->g, "~((a & b) | f | ((c & d) & e))"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "AOI321");
                    }
                    else if(!strcmp(currmos->g, "~((a & b) | (c & d) | (e & f))"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "AOI222");
                    }
                    else if(!strcmp(currmos->g, "~((a | b) & (c | d) & (e | f))"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "OAI222");
                    }
                    else if(!strcmp(currmos->g, "~((a | b | c) & (d & e & f))"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "OAI33");
                    }
                    else if(!strcmp(currmos->g, "~((a | b | c) & (d | e) & f)") || !strcmp(currmos->g, "~((a | b) & (c | d | e) & f)") || !strcmp(currmos->g,"~((a | b | c) & f (d | e))") || !strcmp(currmos->g, "~((a | b) & f & (c | d | e))"))
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "OAI321");
                    }
		    		else if(!strcmp(currmos->g, "~a"))
		    		{
						;
		    		}
		    		else
		    		{           
	                	char input_name;
                        input_name='a';
                        MosList *list=curr_n->mos_list;
                        while(counter(curr_n)!=1)
                        {
                            input_name = parallel_recognition(curr_n,input_name);

                            input_name = series_recognition(curr_n, input_name);
                        }
                        list->g = generate_mark(list->g);
						curr_n->greatthan4 = 1;
                        curr_n->headername = NLAlloc(char, strlen(model_name)+2);
                        strcpy(curr_n->headername, model_name);
                        int k = strlen(curr_n->headername);
                        curr_n->headername[k]=model_add;
                        curr_n->headername[k+1]='\0';
                        model_add+=1;
						printf("This p_branch contains more than 7 mosfets, can't handle\n");
						printf("Branch name = %s, Logic function is: \n", curr_n->headername, list->g);
		     		}
		}
    }
}

MCGate *
MGroup::insert_cgate(MCGate *head, MCGate *newgate)
{
    MCGate *current;
    if(head==NULL)
    {
	head=NLAlloc(MCGate,1);
	head->gatename = NLAlloc(char, strlen(newgate->gatename));
	strcpy(head->gatename, newgate->gatename);
	head->n_branch = newgate->n_branch;
	head->p_branch = newgate->p_branch;
	head->cmos_gate = newgate->cmos_gate;
	head->next = NULL;
    }
    else
    {
	current=NLAlloc(MCGate,1);
	current->gatename = NLAlloc(char, strlen(newgate->gatename));
	strcpy(current->gatename, newgate->gatename);
	current->n_branch=newgate->n_branch;
	current->p_branch=newgate->p_branch;
	current->cmos_gate = newgate->cmos_gate;
	current->next = head->next;
	head->next = current;
    }
    return head;
}

void
MGroup::gate_recognition()
{
	string s1 = "gate";
	char s2[20];
	int counter = 0;
	for(MBranch *curr_n=n_branch_list; curr_n!=NULL; curr_n=curr_n->next)
	{
		for(MBranch *curr_p=p_branch_list; curr_p!=NULL; curr_p=curr_p->next)
		{
			if(!strcmp(curr_n->pn_node, curr_p->pn_node))
			{
				sprintf(s2, "%d", counter); 
				string s3 = s1 + s2;
				counter += 1;

/*
					printf("........ A new gate, pn_node = %s: \n", curr_n->pn_node);
					printf("N_number=%d, P_number=%d\n", curr_n->num_of_mos, curr_p->num_of_mos);
					printf("N_fun=%s, P_fun=%s\n", (curr_n->mos_list)->g, (curr_p->mos_list)->g);
					printf("N_fun=%s, P_fun=%s\n", curr_n->headername, curr_p->headername);
*/

				if(curr_n->num_of_mos != curr_p->num_of_mos)
				{
					MCGate *new_gate = new MCGate(s3.c_str(),curr_n, curr_p, false); 
					cmos_gate_list = insert_cgate(cmos_gate_list, new_gate);
				}
				else
				{
					if(
						  ( !strcmp(curr_n->headername,"INV") && !strcmp(curr_p->headername, "INV") )
					  ||  ( !strcmp(curr_n->headername,"NAND2") && !strcmp(curr_p->headername, "NOR2") )
					  ||  ( !strcmp(curr_n->headername,"NOR2") && !strcmp(curr_p->headername, "NAND2") )
					  ||  ( !strcmp(curr_n->headername,"NAND3") && !strcmp(curr_p->headername, "NOR3") )
					  ||  ( !strcmp(curr_n->headername,"NOR3") && !strcmp(curr_p->headername, "NAND3") )
					  ||  ( !strcmp(curr_n->headername,"OAI12") && !strcmp(curr_p->headername, "AOI12") )
					  ||  ( !strcmp(curr_n->headername,"AOI12") && !strcmp(curr_p->headername, "OAI12") )
					  ||  ( !strcmp(curr_n->headername,"NOR4") && !strcmp(curr_p->headername, "NAND4") )
					  ||  ( !strcmp(curr_n->headername,"NAND4") && !strcmp(curr_p->headername, "NOR4") )
					  ||  ( !strcmp(curr_n->headername,"AOI22") && !strcmp(curr_p->headername, "OAI22") )
					  ||  ( !strcmp(curr_n->headername,"OAI22") && !strcmp(curr_p->headername, "AOI22") )
					  ||  ( !strcmp(curr_n->headername,"AOI13") && !strcmp(curr_p->headername, "OAI13") )
					  ||  ( !strcmp(curr_n->headername,"OAI13") && !strcmp(curr_p->headername, "AOI13") )
					  ||  ( !strcmp(curr_n->headername,"OAOI112") && !strcmp(curr_p->headername, "AOAI112") )
					  ||  ( !strcmp(curr_n->headername,"AOOI112") && !strcmp(curr_p->headername, "OAAI112") )
					  ||  ( !strcmp(curr_n->headername,"OAAI112") && !strcmp(curr_p->headername, "AOOI112") )
					  ||  ( !strcmp(curr_n->headername,"AOAI112") && !strcmp(curr_p->headername, "OAOI112") )
					  ||  ( !strcmp(curr_n->headername,"AOI23") && !strcmp(curr_p->headername, "OAI23") )
					  ||  ( !strcmp(curr_n->headername,"AOI311") && !strcmp(curr_p->headername, "OAI311") )
					  ||  ( !strcmp(curr_n->headername,"AOI221") && !strcmp(curr_p->headername, "OAI221") )
					  ||  ( !strcmp(curr_n->headername,"OAI23") && !strcmp(curr_p->headername, "AOI23") )
					  ||  ( !strcmp(curr_n->headername,"OAI311") && !strcmp(curr_p->headername, "AOI311") )
					  ||  ( !strcmp(curr_n->headername,"OAI221") && !strcmp(curr_p->headername, "AOI221") )

					  ||  ( !strcmp(curr_n->headername,"OAOI212") && !strcmp(curr_p->headername, "AOAI212") )
					  ||  ( !strcmp(curr_n->headername,"OAOI2111") && !strcmp(curr_p->headername, "AOAI2111") )
					  ||  ( !strcmp(curr_n->headername,"AOAI212") && !strcmp(curr_p->headername, "OAOI212") )
					  ||  ( !strcmp(curr_n->headername,"AOAI2111") && !strcmp(curr_p->headername, "AOAI2111") )

					  ||  ( !strcmp(curr_n->headername,"AOI33") && !strcmp(curr_p->headername, "OAI33") )
					  ||  ( !strcmp(curr_n->headername,"AOI321") && !strcmp(curr_p->headername, "OAI321") )
					  ||  ( !strcmp(curr_n->headername,"AOI222") && !strcmp(curr_p->headername, "OAI222") )
					  ||  ( !strcmp(curr_n->headername,"OAI222") && !strcmp(curr_p->headername, "AOI222") )
					  ||  ( !strcmp(curr_n->headername,"OAI33") && !strcmp(curr_p->headername, "AOI33") )
					  ||  ( !strcmp(curr_n->headername,"OAI321") && !strcmp(curr_p->headername, "AOI321") ) )
					{
						MCGate *new_2_gate = new MCGate(s3.c_str(),curr_n, curr_p, true);
						cmos_gate_list = insert_cgate(cmos_gate_list, new_2_gate);
					}
					else
					{
						MCGate *new_3_gate = new MCGate(s3.c_str(),curr_n, curr_p, false);
						cmos_gate_list = insert_cgate(cmos_gate_list, new_3_gate);
					}
				}
//				counter+=1;
				break;
			}
		}
	}
	num_of_gates = counter;
}

void
MGroup::gen_hash_gate_list()
{
	hash_gate_list = new PMCGate[num_of_gates];
	for(MCGate *curr_gate=cmos_gate_list; curr_gate!=NULL; curr_gate=curr_gate->next)
	{
		hash_gate_list[return_number(curr_gate->gatename)] = curr_gate;
	}
}

PinList *
MGroup::insert_pin(PinList *head, char *name){
  PinList *current;
  if(head==NULL)
  {
        head=NLAlloc(PinList,1);
        head->NodeName = NLAlloc(char,strlen(name));
        strcpy(head->NodeName,name);
        head->next = NULL;
  }
  else {
        current=NLAlloc(PinList,1);
        current->NodeName = NLAlloc(char,strlen(name));
        strcpy(current->NodeName,name);
        current->next = head->next;
        head->next = current;
  }
  return head;
}
bool 
MGroup::query_in_pinlist(char *name, PinList *plist)
{
	for(PinList *curr=plist; curr!=NULL; curr=curr->next)
	{
		if(!strcmp(name, curr->NodeName))
			return true;
	}
	return false;
}

void
MGroup::gen_inlist_for_gate()
{
	for(int i=0; i<num_of_gates; i++)
	{
		if(hash_gate_list[i]!=NULL)
		{
			for(NameList *cn=(hash_gate_list[i]->n_branch)->name_list; cn!=NULL; cn=cn->next)
				hash_gate_list[i]->inlist = insert_pin(hash_gate_list[i]->inlist, cn->gatename);
			for(NameList *cp=(hash_gate_list[i]->p_branch)->name_list; cp!=NULL; cp=cp->next)
			{
				if(!query_in_pinlist(cp->gatename, hash_gate_list[i]->inlist))
				{
				hash_gate_list[i]->inlist = insert_pin(hash_gate_list[i]->inlist, cp->gatename);
					hash_gate_list[i]->pn_input_match = false;
				}
			}
		}
	}
}


void
MGroup::gen_hash_trans_list()
{
	hash_trans_list = new PTransList[num_of_trans];
	for(TransList *curr=transhead; curr!=NULL; curr=curr->next)
	{
		hash_trans_list[return_number(curr->transname)] = curr;
	}
}


bool 
MGroup::query_in_gate_list(char *name, PinList *plist)
{
  if(plist==NULL)
	  return false;
  else
  {
	for(PinList *curr=plist; curr!=NULL; curr=curr->next)
	{
		if(!strcmp(name, curr->NodeName))
			return true;
	}
  }
	return false;
}


void
MGroup::gen_connection_of_gates()
{
	for(int i=0; i<(num_of_gates); i++)
	{
		if(hash_gate_list[i]!=NULL)
		{
			for(NodeList *cur=nodehead; cur!=NULL; cur=cur->next)
			{
				if(!strcmp(cur->nodename, (hash_gate_list[i]->n_branch)->pn_node))
				{
					cur->output_of_gate_name = NLAlloc(char, strlen(hash_gate_list[i]->gatename));
					strcpy(cur->output_of_gate_name, hash_gate_list[i]->gatename);
					break;
				}
			}

			for(NameList *curr_n=(hash_gate_list[i]->n_branch)->name_list; curr_n!=NULL; curr_n=curr_n->next)
			{
				for(NodeList *cur=nodehead; cur!=NULL; cur=cur->next)
				{
					if(!strcmp(cur->nodename, curr_n->gatename))
					{
						if(!query_in_gate_list(hash_gate_list[i]->gatename, cur->input_of_gate_list))
							cur->input_of_gate_list = insert_pin(cur->input_of_gate_list, hash_gate_list[i]->gatename);
						break;
					}
				}
			}			

			for(NameList *curr_p=(hash_gate_list[i]->p_branch)->name_list; curr_p!=NULL; curr_p=curr_p->next)
			{
				for(NodeList *cur=nodehead; cur!=NULL; cur=cur->next)
				{
					if(!strcmp(cur->nodename, curr_p->gatename))
					{
						if(!query_in_gate_list(hash_gate_list[i]->gatename, cur->input_of_gate_list))
							cur->input_of_gate_list = insert_pin(cur->input_of_gate_list, hash_gate_list[i]->gatename);
						break;
					}
				}
			}			
		}
	}
}

void
MGroup::gen_connection_of_trans()
{
	for(int i=0; i<(num_of_trans); i++)
	{
		if(hash_trans_list[i]!=NULL)
		{
			for(NodeList *curr=nodehead; curr!=NULL; curr=curr->next)
			{
				if(!strcmp(curr->nodename, hash_trans_list[i]->output))
				{
					curr->output_of_trans_name = NLAlloc(char, strlen(hash_trans_list[i]->transname));
					strcpy(curr->output_of_trans_name, hash_trans_list[i]->transname);
				}
				if(!strcmp(curr->nodename, hash_trans_list[i]->input))
				{
					curr->input_of_trans_name = NLAlloc(char, strlen(hash_trans_list[i]->transname));
					strcpy(curr->input_of_trans_name, hash_trans_list[i]->transname);
//					if(!query_in_gate_list(hash_trans_list[i]->input, curr->input_of_trans_list))
//						curr->input_of_trans_list = insert_pin(curr->input_of_trans_list, hash_trans_list[i]->transname);
				}
			}
		}
	}
}

void
MGroup::count_input_num_of_node()
{
	for(NodeList *curr=nodehead; curr!=NULL; curr=curr->next)
	{
			if(curr->pn_node)
			{
				int res = 0;
				for(PinList *curr2=curr->input_of_gate_list; curr2!=NULL; curr2=curr2->next)
				{
					res += 1;
				}


				if(curr->input_of_trans_name != NULL)
					res += 1;
				if(curr->output_of_trans_name != NULL)
					res += 1;
		/*	
				for(PinList *curr3=curr->input_of_trans_list; curr3!=NULL; curr3=curr3->next)
				{
					res += 1;
				}
		*/	
				curr->num_of_inputs = res;
			}
			else
				curr->num_of_inputs = 0;
	}
}

/*
int
MGroup::query_inoutput(char *name)
{
    int flag = 0;
    for(InoutList *curr=inputhead; curr!=NULL; curr=curr->next)
    {
	if(!strcmp(curr->pinname, name))
	{
	    flag=1;
	    return 1;
	}
    }
    for(InoutList *curr2=outputhead; curr2!=NULL; curr2=curr2->next)
    {
	if(!strcmp(curr2->pinname, name))
	{
	    flag=1;
	    return 1;
	}
    }

    if(!flag)
	return 0;
}

WireList *
MGroup::insert_wire(WireList *head, char *name)
{
    WireList *current;
    if(head==NULL)
    {
	head=NLAlloc(WireList,1);
	head->wirename=NLAlloc(char, strlen(name));
	strcpy(head->wirename,name);
	head->next = NULL;
    }
    else
    {
	current=NLAlloc(WireList,1);
	current->wirename=NLAlloc(char, strlen(name));
	strcpy(current->wirename,name);
	current->next = head->next;
	head->next = current;
    }
    return head;
}

int
MGroup::query_wire(char *wire)
{
    int flag=0;
    for(WireList *curr_w=wirehead; curr_w!=NULL; curr_w=curr_w->next)
    {
	if(!strcmp(curr_w->wirename, wire))
	{
	    flag=1;
	    break;
	}
    }

    return flag;
}

void
MGroup::generate_wirelist()
{
    for(MBranch *curr_b=n_branch_list; curr_b!=NULL; curr_b=curr_b->next)
    {
        for(NameList *curr_name=curr_b->name_list; curr_name!=NULL; curr_name=curr_name->next)
        {
            if(!query_inoutput(curr_name->gatename))
                wirehead=insert_wire(wirehead,curr_name->gatename);
        }
    }
    for(TransList *curr_t=transhead; curr_t!=NULL; curr_t=curr_t->next)
    {
        if(!query_wire(curr_t->input) && !query_inoutput(curr_t->input))
            wirehead=insert_wire(wirehead,curr_t->input);
        if(!query_wire(curr_t->output) && !query_inoutput(curr_t->output))
            wirehead=insert_wire(wirehead,curr_t->output);
    }
}

void
MGroup::delete_wire()
{
    for(WireList *curr_w=wirehead; curr_w!=NULL; curr_w=curr_w->next)
    {
        WireList *curr, *prev;
        prev=curr_w;
        curr=prev->next;
        while(curr!=NULL)
        {
	    if(!strcmp(curr_w->wirename, curr->wirename))
	    {
		prev->next = curr->next;
		curr = prev->next;
	    }
	    else
	    {
		prev=curr;
		curr=curr->next;
	    }
	}
    }
}
*/
