//---------------------------------------------------------
//  macromodel.c
//  Data Structure and operation for every Macromodel
//
//  Lei Yang       Nov 7, 2002
//---------------------------------------------------------

#include "pre.h"
#include "mtellnp.h"
#include "mnode.h"
#include "mbranch.h"
#include "mgroup.h"
#include "mcgate.h"
#include "macromodel.h"
#include <malloc.h>
#include <stdio.h>
#include <iostream>
#include <malloc.h>
#include <stdio.h>
#include <cstring>
#include <string>
using namespace std;

extern MacroModelList *MacroModelHead;
#define NLAlloc(type,num) (type *)calloc(num,sizeof(type))

////////////////////// Convert the terminal_list to inputhead ////////////////
// convert_terminal_to_input
InoutList *
Macromodel::insert_inout(InoutList *head, char *name)
{
  InoutList *current;
  if(head==NULL)
  {
        head=NLAlloc(InoutList,1);
        head->pinname = NLAlloc(char,strlen(name));
        strcpy(head->pinname,name);
        head->next = NULL;
  }
  else {
        current=NLAlloc(InoutList,1);
        current->pinname = NLAlloc(char,strlen(name));
        strcpy(current->pinname,name);
        current->next = head->next;
        head->next = current;
  }
  return head;
}

////////////////////// Macromodel //////////////////////
// construct class macromodel

Macromodel::Macromodel(MacroModelList *amacromodel)
{
    name = NLAlloc(char, strlen(amacromodel->macromodel_name));
    strcpy(name, amacromodel->macromodel_name);
    macromodel = amacromodel;
	modelhead = amacromodel->model_list;
    inputhead = NULL;

	for(TerminalList *curr=amacromodel->terminal_list; curr!=NULL; curr=curr->next)
	{
		inputhead = insert_inout(inputhead, curr->name);
	}

	outputhead = NULL;
	wire_list = NULL;
    mos_list = amacromodel->mos_list;
    node_list = NULL;	
    gate_stack = NULL;
    node_stack = NULL;
	hash_gate_list = NULL;
	hash_trans_list = NULL;
	num_of_gates = 0;
	num_of_trans = 0;
}

//////////////////// ~Macromodel ////////////////////////
// destruct class macromodel

Macromodel::~Macromodel()
{
    if(name) delete name;
    if(macromodel) delete macromodel;
	if(modelhead) delete modelhead;
    if(inputhead) delete inputhead;
    if(outputhead) delete outputhead;
	if(wire_list) delete wire_list;
    if(mos_list) delete mos_list;
    if(node_list) delete node_list;
    if(node_stack) delete node_stack;
    if(gate_stack) delete gate_stack;
	if(hash_gate_list) delete hash_gate_list;
	if(hash_trans_list) delete hash_trans_list;
}

// Given a node string name, I will go to hash_node_list to return a *node
int
Macromodel::return_number(char *name)
{
        char *st = NLAlloc(char, strlen(name));
        strcpy(st, name);
        char *ret=NULL;
        ret = NLAlloc(char, strlen(st));
        int counter = 0;

        bool begin_f=false;
        bool end_f=false;

        int i=0;
        while(st[i]!='\0')
        {
                if(st[i]>=48 && st[i]<=57 && st[i]!='\0')
                {
                        begin_f = true;
                        ret[counter]=st[i];
                        counter+=1;
                }
                if(begin_f && st[i]!='\0' && ( st[i]<48 || st[i]>57 ))
                {
                        end_f = true;
                        break;
                }
                i+=1;
        }
        ret[counter] =  '\0';

        if(begin_f && !end_f)
        {
                return atoi(ret);
        }
        else
        {
                return -1;
        }
}

/////////////////// ~Generate the data structure of macromodel ////////////
// gen_data_structure
void
Macromodel::gen_data_structure()
{
	MTellNP *mytellnp = new MTellNP(mos_list, modelhead);
	mytellnp->mos_np_netlist();
		mos_list = mytellnp->moshead;

	MNode *mynode = new MNode(mos_list, inputhead);
	mynode->gen_nodelist();
	mynode->gen_np_char_netlist();
	mynode->delete_vdd_gnd_from_inputlist();
	mynode->generate_output_pinlist();
	mynode->generate_input_pinlist();
		mos_list = mynode->moshead;
		node_list = mynode->nodehead;
		inputhead = mynode->inputhead;
		outputhead = mynode->outputhead;

	
/*
	    	for(NodeList *cur_n=node_list; cur_n!=NULL; cur_n=cur_n->next)
	        {
		        printf("..................%s(%d): ", cur_n->nodename,cur_n->pn_node);
		        for(DeviceList *cur_d=cur_n->devicelist; cur_d!=NULL; cur_d=cur_d->next)
		            printf("%s ",cur_d->devicename);
		        printf("\n");
	    	}
*/
	

	cout<<"finish node step\n";

	MGroup *mygroup = new MGroup(mos_list, node_list, inputhead, outputhead);
	mygroup->generate_branch_list();
//		cout<<"hehe1\n";


/*
    printf("N_Branch_list...........\n");
    MBranch *cur_b=mygroup->n_branch_list;
    if(cur_b==NULL)
	printf("empty\n");
    else if(cur_b->next==NULL)
	printf("only one\n");
    else {
    for(; cur_b!=NULL; cur_b=cur_b->next)
    {
	printf("n_branch_list_i:\n");
	printf("n_or_p=%d, success=%d, pn_node=%s\n",cur_b->branch_type,cur_b->success,cur_b->pn_node);
	printf("Moslist: ");
	for(MosList *cur_mo=cur_b->mos_list; cur_mo!=NULL; cur_mo=cur_mo->next)
	    printf("%s  ", cur_mo->mosname);
	printf("\n");
	printf("Nodelist: ");
	if(cur_b->node_list==NULL)
	    printf("No internal node\n");
	else
	{
	for(BranchNodeList *cur_no=cur_b->node_list; cur_no!=NULL; cur_no=cur_no->next)
	    printf("%s  ",cur_no->nodename);
	printf("\n");
	}
    }
    }

    printf("P_Branch_list...........\n");
    MBranch *cur_bb=mygroup->p_branch_list;
    if(cur_bb==NULL)
        printf("empty\n");
    else if(cur_bb->next==NULL)
        printf("only one\n");
    else {
    for(; cur_bb!=NULL; cur_bb=cur_bb->next)
    {
        printf("n_branch_list_i:\n");
        printf("n_or_p=%d, success=%d, pn_node=%s\n",cur_bb->branch_type,cur_bb->success,cur_bb->pn_node);
        printf("Moslist: ");
        for(MosList *cur_mo=cur_bb->mos_list; cur_mo!=NULL; cur_mo=cur_mo->next)
            printf("%s  ", cur_mo->mosname);
        printf("\n");
        printf("Nodelist: ");
        if(cur_bb->node_list==NULL)
            printf("No internal node\n");
        else
        {
        for(BranchNodeList *cur_no=cur_bb->node_list; cur_no!=NULL; cur_no=cur_no->next)
            printf("%s  ",cur_no->nodename);
        printf("\n");
	}
    }
    }
*/



	mygroup->group_branch_list();
//		cout<<"hehe2\n";

	mygroup->generate_pass_left_list();
//		cout<<"hehe3\n";
	mygroup->generate_trans_list();
//		cout<<"hehe4\n";

/*
    for(TransList *curr_t=mygroup->transhead; curr_t!=NULL; curr_t=curr_t->next)
    {
        printf("begin:\n");
	printf("transname: %s",curr_t->transname);
        printf("input=%s, output=%s, p_g=%s, n_g=%s, delete=%d\n",  curr_t->input, curr_t->output, curr_t->p_g, curr_t->n_g, curr_t->ready_to_delete);
        printf("end.\n");
    }
*/
	mygroup->generate_num_for_branch();
//		cout<<"hehe5\n";
	mygroup->n_branch_recognition();
//		cout<<"hehe6\n";
	mygroup->p_branch_recognition();
//		cout<<"hehe7\n";
	mygroup->gate_recognition();
//		cout<<"hehe8\n";
	mygroup->gen_hash_gate_list();
//		cout<<"hehe9\n";


/*
if(!strcmp(macromodel->macromodel_name, "DFFP"))
{
	for(int i=0; i<num_of_gates; i++)
	{
		if(mygroup->hash_gate_list[i]!=NULL)
		{
			cout<<" A new gate ....................."<<(mygroup->hash_gate_list[i])->gatename<<endl;
			cout<<"pn_node: "<<((mygroup->hash_gate_list[i])->n_branch)->pn_node<<endl;
			cout<<"real_gate: "<<(mygroup->hash_gate_list[i])->cmos_gate<<"  p_fun: "<<((mygroup->hash_gate_list[i])->p_branch)->headername<<" n_fun: "<<((mygroup->hash_gate_list[i])->n_branch)->headername<<endl;
			cout<<endl;

		}
		else
			cout<<"gate: NULL\n";
	}
}
*/

	mygroup->gen_inlist_for_gate();
//		cout<<"hehe10\n";
	mygroup->gen_hash_trans_list();
//		cout<<"hehe11\n";
	mygroup->gen_connection_of_gates();
//		cout<<"hehe12\n";
	mygroup->gen_connection_of_trans();
//		cout<<"hehe13\n";

/*
if(!strcmp(macromodel->macromodel_name, "INV4"))
{
	for(NodeList *cun=node_list; cun!=NULL; cun=cun->next)
	{
			cout<<endl;
			cout<<"node ............................."<<cun->nodename<<endl;
			if(cun->output_of_gate_name != NULL)
				cout<<"  Output of gate: "<<cun->output_of_gate_name;
			else
				cout<<"  No output gate";

			if(cun->input_of_gate_list != NULL)
			{
				cout<<"  input of gate: ";
				for(PinList *curr=cun->input_of_gate_list; curr!=NULL; curr=curr->next)
					cout<<" "<<curr->NodeName;
				cout<<endl;
			}
			else
				cout<<"  No inputput gate";

			
			if(cun->input_of_trans_name!=NULL)
				cout<<" Input of trans: "<<cun->input_of_trans_name<<endl;
			else
				cout<<" Input of trans: NULL"<<endl;
			if(cun->output_of_trans_name!=NULL)
				cout<<" Output of trans: "<<cun->output_of_trans_name<<endl;
			else
				cout<<" Output of trans: NULL"<<endl;
	}
}
*/



		
	mygroup->count_input_num_of_node();

		inputhead = mygroup->inputhead;
		outputhead = mygroup->outputhead;
		mos_list = mygroup->moshead;
		node_list = mygroup->nodehead;
		trans_list = mygroup->transhead;
		hash_gate_list = mygroup->hash_gate_list;
		hash_trans_list = mygroup->hash_trans_list;
		num_of_gates = mygroup->num_of_gates;
		num_of_trans = mygroup->num_of_trans;
	cout<<"finish gen structure\n";
		
}

//////////////////// Insert gate in gatestack //////////////

GateStack *
Macromodel::insert_gate(GateStack *head, int type, char *name)
{
    GateStack *current;
    if(head==NULL)
    {
	head = NLAlloc(GateStack, 1);
	head->type = type;
	head->macromodel_name = NLAlloc(char, strlen(name));
	strcpy(head->macromodel_name, name);
	head->circuit_name = NULL;
	head->recognized_flag = false;
	head->search_flag = false;
	head->next = NULL;
    }
    else
    {
	current = NLAlloc(GateStack, 1);
	current->type = type;
	current->macromodel_name = NLAlloc(char, strlen(name));
	strcpy(current->macromodel_name, name);
	current->circuit_name = NULL;
	current->recognized_flag = false;
	current->search_flag = false;
	current->next = head->next;
	head->next = current;
    }
    return head;
}


/////////////////// Initialize gate stack ////////////////////////
// initialize_gate_stack
void
Macromodel::initialize_gate_stack()
{
	for(int i=0; i<num_of_gates; i++)
	{
		if(hash_gate_list[i]!=NULL)
		{
			gate_stack = insert_gate(gate_stack, 0, hash_gate_list[i]->gatename);
//			cout<<" Insert gate: "<<hash_gate_list[i]->gatename<<endl;
		}
	}
	for(int j=0; j<num_of_trans; j++)
	{
		if(hash_trans_list[j]!=NULL)
		{
			gate_stack = insert_gate(gate_stack, 1, hash_trans_list[j]->transname);
//			cout<<" Insert trans: "<<hash_trans_list[j]->transname<<endl;
		}
	}
}

//////////////////// Insert node in nodestack //////////////

NodeStack *
Macromodel::insert_node(NodeStack *head, char *name, bool terminal_flag)
{
    NodeStack *current;
    if(head==NULL)
    {
	head = NLAlloc(NodeStack, 1);
	head->macromodel_name = NLAlloc(char, strlen(name));
	strcpy(head->macromodel_name, name);
	head->circuit_name = NULL;
	head->terminal_flag = terminal_flag;
	head->recognized_flag = false;
	head->search_flag = false;
	head->next = NULL;
    }
    else
    {
	current = NLAlloc(NodeStack, 1);
	current->macromodel_name = NLAlloc(char, strlen(name));
	strcpy(current->macromodel_name, name);
	current->circuit_name = NULL;
	current->terminal_flag = terminal_flag;
	current->recognized_flag = false;
	current->search_flag = false;
	current->next = head->next;
	head->next = current;
    }
    return head;
}

////////////////// Query if this node is a input or output //////////
// query_terminal
bool
Macromodel::query_terminal(char *name)
{
	for(InoutList *curr=inputhead; curr!=NULL; curr=curr->next)
	{
		if(!strcmp(name, curr->pinname))
			return true;
	}
	for(InoutList *curr2=outputhead; curr2!=NULL; curr2=curr2->next)
	{
		if(!strcmp(name, curr2->pinname))
			return true;
	}
	return false;
}


////////////////// Initialize node stack /////////////////////////
// initialize_node_stack
void
Macromodel::initialize_node_stack()
{
	for(NodeList *curr=node_list; curr!=NULL; curr=curr->next)
	{
		bool flag=query_terminal(curr->nodename);
		node_stack = insert_node(node_stack, curr->nodename, flag);
//		cout<<" nodename: "<<curr->nodename<<" terminal flag: "<<flag<<endl;
	}
}

//////////////// locate first gate or trans-gate ///////////////
// locate_first_gate
// 1) First gate is the starting gate to do the pattern matching
// 2) step 1: fine the gate whose output connected to most num of gates.
// 3) step 2: from this gate, tracing back, until we find the gate whose inputs
//            are all primary inputs.
// Revised History: When there are a ring in backtracking, choose any of them to be the source node. 

bool
Macromodel::decide_a_gate(char *name)
{
    if(name[0]=='g' && name[1]=='a' && name[2]=='t' && name[3]=='e')
        return true;
    else
        return false;
}

// Given a named gate, go to check if there exist one gate whose output connect to this named gate.
// Return this gate founded. 

char *
Macromodel::return_gate(char *name)
{
	char *trans_name=NULL;
	bool gate_or_trans = decide_a_gate(name);
	if(gate_or_trans)
	{
		MCGate *hgate = hash_gate_list[return_number(name)];
		for(PinList *curr=hgate->inlist; curr!=NULL; curr=curr->next)
		{
			for(NodeList *curr_n=node_list; curr_n!=NULL; curr_n=curr_n->next)
			{
				if(!strcmp(curr_n->nodename, curr->NodeName))
				{
					if(curr_n->output_of_gate_name != NULL)
						return curr_n->output_of_gate_name;
					if(curr_n->output_of_trans_name !=NULL)
					{
						trans_name = NLAlloc(char, strlen(curr_n->output_of_trans_name));
						strcpy(trans_name, curr_n->output_of_trans_name);
					}
					if(curr_n->input_of_trans_name !=NULL)
					{
						trans_name = NLAlloc(char, strlen(curr_n->input_of_trans_name));
						strcpy(trans_name, curr_n->input_of_trans_name);
					}
					break;
				}
			}
		}
/*
		for(NameList *curr=(hgate->n_branch)->name_list; curr!=NULL; curr=curr->next)
		{
			for(NodeList *curr_n=node_list; curr_n!=NULL; curr_n=curr_n->next)
			{
				if(!strcmp(curr_n->nodename, curr->gatename))
				{
					if(curr_n->output_of_gate_name != NULL)
						return curr_n->output_of_gate_name;
					if(curr_n->output_of_trans_name !=NULL)
					{
						trans_name = NLAlloc(char, strlen(curr_n->output_of_trans_name));
						strcpy(trans_name, curr_n->output_of_trans_name);
					}
					break;
				}
			}
		}
		for(NameList *curr=(hgate->p_branch)->name_list; curr!=NULL; curr=curr->next)
		{
			for(NodeList *curr_n=node_list; curr_n!=NULL; curr_n=curr_n->next)
			{
				if(!strcmp(curr_n->nodename, curr->gatename))
				{
					if(curr_n->output_of_gate_name != NULL)
						return curr_n->output_of_gate_name;
					if(curr_n->output_of_trans_name !=NULL)
					{
						trans_name = NLAlloc(char, strlen(curr_n->output_of_trans_name));
						strcpy(trans_name, curr_n->output_of_trans_name);
					}
					break;
				}
			}
		}
*/
	}
	else
	{
		TransList *htrans = hash_trans_list[return_number(name)];
		for(NodeList *curr=node_list; curr!=NULL; curr=curr->next)
		{
			if(!strcmp(curr->nodename, htrans->input))
			{
					if(curr->output_of_gate_name != NULL)
						return curr->output_of_gate_name;
					if(curr->output_of_trans_name !=NULL)
					{
						trans_name = NLAlloc(char, strlen(curr->output_of_trans_name));
						strcpy(trans_name, curr->output_of_trans_name);
					}
					if(curr->input_of_trans_name !=NULL)
					{
						trans_name = NLAlloc(char, strlen(curr->input_of_trans_name));
						strcpy(trans_name, curr->input_of_trans_name);
					}
				break;
			}
		}
	}
	if(trans_name!=NULL)
		return trans_name;
	else
		return NULL;
}

/************************ Locate the first gate *********************************/
// Keep backtracking to find the first source gate

bool
Macromodel::query_in_ring_list(char *name, InoutList *ring_list)
{
  if(ring_list == NULL)
	  return false;
  else
  {
	for(InoutList *curr=ring_list; curr!=NULL; curr=curr->next)
	{
		if(!strcmp(name, curr->pinname))
		{
			return true;
		}
	}
	return false;
  }
}

MCGate *
Macromodel::locate_first_gate()
{
// If there is only one gate contained in macromodel, this gate is first_gate
  if(num_of_gates == 1)
  {
	return hash_gate_list[0];
  }
// First find the gate whose output connected to most num of gates
  else
  { 
	int stand=0;
	NodeList *dnode;
	for(NodeList *curr=node_list; curr!=NULL; curr=curr->next)
	{
	  if(curr->pn_node)
	  {
		if(curr->num_of_inputs > stand)
		{
			stand = curr->num_of_inputs;
			dnode = curr;
		}
	  }
	}

	char *rname = NULL;
	if(dnode->output_of_gate_name != NULL)
	{
		rname = NLAlloc(char, strlen(dnode->output_of_gate_name));
		strcpy(rname, dnode->output_of_gate_name);
	}

	char *bak_gate_name;

// To store all the node which has been backtracked in ring_list, if one node is find again, 
// consider it as the source node. 
// Yang Lei, Mar 30, 2003.
	InoutList *ring_list = NULL;
	int counter = 1;
	while(rname != NULL)
	{
		bak_gate_name = NLAlloc(char, strlen(rname));
		strcpy(bak_gate_name, rname);

		if(counter <= num_of_gates)  // If backtrack <= num_of_gates times, store node in ring_list. 
		{
			if(!query_in_ring_list(bak_gate_name,ring_list))
			{
				ring_list = insert_inout(ring_list, bak_gate_name);
			}
		}
		else
		{
			if(!query_in_ring_list(bak_gate_name,ring_list))
			{
				ring_list = insert_inout(ring_list, bak_gate_name);
			}
			else
			{
				break;
			}
		}

		rname = return_gate(rname);
		counter += 1;
	}

//	cout<<"name: "<<bak_gate_name<<endl;
	return hash_gate_list[return_number(bak_gate_name)];
  }
}

////////////////////// GateStack search to end ///////////////////////////
//gatestack_search_to_end (if did not finish, return 1)
bool
Macromodel::gatestack_search_to_end()
{
    int flag = true;
    for(GateStack *curr=gate_stack; curr!=NULL; curr=curr->next)
    {
		if(!curr->recognized_flag)
	    	flag = false;
    }

    if(!flag)
	return true;
    else
	return false;
}

///////////////////// NodeStack search to end //////////////////////////
// nodestack_search_to_end
bool
Macromodel::nodestack_search_to_end()
{
    int flag = true;
    for(NodeStack *curr=node_stack; curr!=NULL; curr=curr->next)
    {
		if(curr->terminal_flag && !curr->recognized_flag)
	    	flag = false;
    }

    if(!flag)
	return true;
    else
	return false;
}

/////////////////// Insert a pin in a pinlist ///////////////////////////////////
// Insert a pin in a pinlist
RGList *
Macromodel::insert_return_gate(RGList *head, char *name, int type){
  RGList *current;
  if(head==NULL)
  {
        head=NLAlloc(RGList,1);
        head->gatename = NLAlloc(char,strlen(name));
        strcpy(head->gatename,name);
		head->type = type;
		head->unique = true;
// New added
		head->unique2 = true;
        head->next = NULL;
  }
  else {
        current=NLAlloc(RGList,1);
        current->gatename = NLAlloc(char,strlen(name));
        strcpy(current->gatename,name);
		current->type = type;
		current->unique = true;
// New added
		current->unique2 = true;
        current->next = head->next;
        head->next = current;
  }
  return head;
}

/////////////////// Search from a gate //////////////////////////////////////////
// Search from a gate, the output node of this gate will go to the input
// of other gates, these gates will consist of a gate_list, return this gate_list.
RGList *
Macromodel::return_a_gate_list(char *gatename)
{
	 RGList *glist = NULL;
		MCGate *hgate = hash_gate_list[return_number(gatename)];
		for(NodeList *curr=node_list; curr!=NULL; curr=curr->next)
		{
			if(!strcmp(curr->nodename, (hgate->n_branch)->pn_node))
			{
				RGList *glist = NULL;
				for(PinList *cup=curr->input_of_gate_list; cup!=NULL; cup=cup->next)
				{
					glist = insert_return_gate(glist, cup->NodeName, 0);
				}
			
				if(curr->input_of_trans_name!=NULL)
					glist = insert_return_gate(glist, curr->input_of_trans_name, 1);
				if(curr->output_of_trans_name!=NULL)
					glist = insert_return_gate(glist, curr->output_of_trans_name, 1);
			/*
				for(PinList *cup2=curr->input_of_trans_list; cup2!=NULL; cup2=cup2->next)
				{
					glist = insert_return_gate(glist, cup2->NodeName, 1);
				}
			*/
			
				return glist;
			}
		}
}
/********************** Added by Lei Yang ******************************/
/********************** Mar 30, 2003 ***********************************/
/* Purpose: Guarantee total match between two gates(unique=0, unique1 = 1) */
char * 
Macromodel::return_circuit_name_from_nodestack(char *name)
{
	for(NodeStack *curr=node_stack; curr!=NULL; curr=curr->next)
	{
		if(curr->recognized_flag)
		{
			if(!strcmp(curr->macromodel_name, name))
				return curr->circuit_name;
		}
	}
	return NULL;
}
/***************************End of Apr 1, 2003 *************************/


/********************** Added by Lei Yang ******************************/
/********************** Mar 30, 2003 ***********************************/
/**** Purpose: trace from trans gate to other gates ********************/
char * 
Macromodel::query_name_in_nodestack(char *name)
{
	for(NodeStack *curr=node_stack; curr!=NULL; curr=curr->next)
	{
		if(curr->recognized_flag)
		{
			if(!strcmp(curr->macromodel_name, name))
				return curr->macromodel_name;
		}
	}
	return NULL;
}

char *
Macromodel::one_valid_terminal_of_trans(char *gatename)
{
	TransList *htrans = hash_trans_list[return_number(gatename)];
	if(query_name_in_nodestack(htrans->input)!=NULL && query_name_in_nodestack(htrans->output)==NULL)
		return htrans->output;
	else if(query_name_in_nodestack(htrans->input)==NULL && query_name_in_nodestack(htrans->output)!=NULL)
		return htrans->input;
	else 
		return NULL;
}

RGList *
Macromodel::return_a_trans_rglist(char *gatename)
{
	RGList *glist = NULL;
	if(one_valid_terminal_of_trans(gatename)!=NULL)
	{
		char *here = NLAlloc(char, strlen(one_valid_terminal_of_trans(gatename)));
		strcpy(here, one_valid_terminal_of_trans(gatename));
		for(NodeList *curr=node_list; curr!=NULL; curr=curr->next)
		{
			if(!strcmp(curr->nodename,here))
			{
				for(PinList *cup=curr->input_of_gate_list; cup!=NULL; cup=cup->next)
				{
					glist = insert_return_gate(glist, cup->NodeName, 0);
				}
				if(curr->output_of_gate_name!=NULL)
					glist = insert_return_gate(glist, curr->output_of_gate_name, 0);

				if(curr->input_of_trans_name!=NULL)
					glist = insert_return_gate(glist, curr->input_of_trans_name, 1);
				if(curr->output_of_trans_name!=NULL)
					glist = insert_return_gate(glist, curr->output_of_trans_name, 1);
				return glist;
			}
		}
	}
	else
		return NULL;
}

/********************* The end *****************************************/

RGList *
Macromodel::return_a_trans_output_list(char *gatename)
{
	 RGList *glist = NULL;
		TransList *htrans = hash_trans_list[return_number(gatename)];
		for(NodeList *curr=node_list; curr!=NULL; curr=curr->next)
		{
			if(!strcmp(curr->nodename, htrans->output))
			{
				RGList *glist = NULL;
				for(PinList *cup=curr->input_of_gate_list; cup!=NULL; cup=cup->next)
				{
					glist = insert_return_gate(glist, cup->NodeName, 0);
				}

				if(curr->input_of_trans_name!=NULL)
					glist = insert_return_gate(glist, curr->input_of_trans_name, 1);
				if(curr->output_of_trans_name!=NULL)
					glist = insert_return_gate(glist, curr->output_of_trans_name, 1);
			/*
				for(PinList *cup2=curr->input_of_trans_list; cup2!=NULL; cup2=cup2->next)
				{
					glist = insert_return_gate(glist, cup2->NodeName, 1);
				}
			*/
				return glist;
			}
		}
}

RGList *
Macromodel::return_a_trans_input_list(char *gatename)
{
	 RGList *glist = NULL;
		TransList *htrans = hash_trans_list[return_number(gatename)];
		for(NodeList *curr=node_list; curr!=NULL; curr=curr->next)
		{
			if(!strcmp(curr->nodename, htrans->input))
			{
				RGList *glist = NULL;
				for(PinList *cup=curr->input_of_gate_list; cup!=NULL; cup=cup->next)
				{
					glist = insert_return_gate(glist, cup->NodeName, 0);
				}

				if(curr->input_of_trans_name!=NULL)
					glist = insert_return_gate(glist, curr->input_of_trans_name, 1);
				if(curr->output_of_trans_name!=NULL)
					glist = insert_return_gate(glist, curr->output_of_trans_name, 1);
			/*
				for(PinList *cup2=curr->input_of_trans_list; cup2!=NULL; cup2=cup2->next)
				{
					glist = insert_return_gate(glist, cup2->NodeName, 1);
				}
			*/
				return glist;
			}
		}
}

bool
Macromodel::query_is_terminal(char *name)
{
	for(InoutList *curr=inputhead; curr!=NULL; curr=curr->next)
	{
		if(!strcmp(name, curr->pinname))
			return true;
	}
	for(InoutList *curr=outputhead; curr!=NULL; curr=curr->next)
	{
		if(!strcmp(name, curr->pinname))
			return true;
	}
	return false;
}

void
Macromodel::unique_rglist(RGList *glist)
{
  if(glist!=NULL)
  {
	for(RGList *curr=glist; curr!=NULL; curr=curr->next)
	{
		for(GateStack *curr_g=gate_stack; curr_g!=NULL; curr_g=curr_g->next)
		{
		  if(curr_g->recognized_flag)
		  {
			if(!strcmp(curr_g->macromodel_name, curr->gatename))
			{
				curr->unique = false;
// New added
				curr->unique2 = false;
				break;
			}
		  }
		}
	}


  if(glist->next!=NULL)
  {
	for(RGList *curr2=glist; curr2!=NULL; curr2=curr2->next)
	{
		if(curr2->unique)
		{
			for(RGList *curr3=curr2->next; curr3!=NULL; curr3=curr3->next)
			{
			  if(curr3->unique)
			  {
				if(!curr2->type && !curr3->type)
				{
					MCGate *h1 = hash_gate_list[return_number(curr2->gatename)];
					MCGate *h2 = hash_gate_list[return_number(curr3->gatename)];
					if(!strcmp((h1->n_branch)->headername, (h2->n_branch)->headername) && !strcmp((h1->p_branch)->headername, (h2->p_branch)->headername))
					{
						curr2->unique = false;
						curr3->unique = false;
// New added
						curr2->unique2 = true;
						curr3->unique2 = true;
					}
				}
				if(curr2->type && curr3->type)
				{
						curr2->unique = false;
						curr3->unique = false;
// New added
						curr2->unique2 = false;
						curr3->unique2 = false;
				}
			  }
			}
		}
	}
  }
  }
}

void
Macromodel::insert_gates_in_gatestack(char *mname, char *cname)
{
	for(GateStack *curr=gate_stack; curr!=NULL; curr=curr->next)
	{
		if(!strcmp(curr->macromodel_name, mname))
		{
			curr->circuit_name = NLAlloc(char, strlen(cname));
			strcpy(curr->circuit_name, cname);
			curr->recognized_flag = true;
//			cout<<" Insert M: "<<curr->macromodel_name<<" Insert C: "<<curr->circuit_name<<endl;
		}
	}
}

void
Macromodel::insert_node_in_nodestack(char *mname, char *cname)
{
	for(NodeStack *curr=node_stack; curr!=NULL; curr=curr->next)
	{
		if(!strcmp(curr->macromodel_name, mname))
		{
//			cout<<"M name: "<<mname<<"  C name: "<<cname<<endl;
			curr->circuit_name = NLAlloc(char, strlen(cname));
			strcpy(curr->circuit_name, cname);
			curr->recognized_flag = true;
		}
	}
}


// To check if the search method can be continued, if return true,
// can continue. Otherwise, can't continue.
bool
Macromodel::check_dead_loop()
{
	for(GateStack *curr=gate_stack; curr!=NULL; curr=curr->next)
	{
//	  if(decide_a_gate(curr->macromodel_name))
//	  {
		if(curr->recognized_flag && !curr->search_flag )
			return false;
//	  }
	}
	return true;
}

// Check if one gate has been found.
char *
Macromodel::gate_has_been_found(char *name)
{
	char *rval = NULL;
	for(GateStack *curr=gate_stack; curr!=NULL; curr=curr->next)
	{
		if(!strcmp(curr->macromodel_name, name))
		{
			if(curr->recognized_flag)
			{
				rval = NLAlloc(char, strlen(curr->circuit_name));
				strcpy(rval, curr->circuit_name);
				return rval;
			}
			else
				return NULL;
		}
	}
}

// Insert a pin in a pinlist
PinList *
Macromodel::insert_pin(PinList *head, char *name){
  PinList *current;
  if(head==NULL)
  {
        head=NLAlloc(PinList,1);
        head->NodeName = NLAlloc(char,strlen(name));
        strcpy(head->NodeName,name);
        head->next = NULL;
  }
  else {
        current=NLAlloc(PinList,1);
        current->NodeName = NLAlloc(char,strlen(name));
        strcpy(current->NodeName,name);
        current->next = head->next;
        head->next = current;
  }
  return head;
}


// Check an unfound gate, if the output of gate connect to other found gates, 
// return a gate_list of those gates.
PinList *
Macromodel::return_found_gate_list(char *gatename)
{
	PinList *rlist = NULL;
//	bool gate_or_trans=decide_a_gate(gatename);
//	if(gate_or_trans)
//	{
		MCGate *hgate = hash_gate_list[return_number(gatename)];
		for(NodeList *curr=node_list; curr!=NULL; curr=curr->next)
		{
			if(!strcmp(curr->nodename, (hgate->n_branch)->pn_node))
			{
//				cout<<curr->nodename<<endl;
				char *here=NULL;
				for(PinList *cp=curr->input_of_gate_list; cp!=NULL; cp=cp->next)
				{
//					cout<<" gate: "<<cp->NodeName<<endl;
					here = gate_has_been_found(cp->NodeName);
					if(here!=NULL)
					{
						rlist = insert_pin(rlist, here);
					}
				}
				
				char *cinput=NULL;
				char *coutput=NULL;
				if(curr->input_of_trans_name != NULL)
					cinput = gate_has_been_found(curr->input_of_trans_name);
				if(curr->output_of_trans_name != NULL)
					coutput = gate_has_been_found(curr->output_of_trans_name);
				if(cinput != NULL)
					rlist = insert_pin(rlist, cinput);
				if(coutput!=NULL)
					rlist = insert_pin(rlist, coutput);
			/*
				for(PinList *cp2=curr->input_of_trans_list; cp2!=NULL; cp2=cp2->next)
				{
					here = gate_has_been_found(cp2->NodeName);
					if(here!=NULL)
					{
						rlist = insert_pin(rlist, here);
					}
				}
			*/
			}
		}
//	}
/*
	else
	{
		TransList *htrans = hash_trans_list[return_number(gatename)];
		for(NodeList *curr=node_list; curr!=NULL; curr=curr->next)
		{
			if(!strcmp(curr->nodename, htrans->output))
			{
				char *here=NULL;
				for(PinList *cp=curr->input_of_gate_list; cp!=NULL; cp=cp->next)
				{
					here = gate_has_been_found(cp->NodeName);
					if(here!=NULL)
					{
						rlist = insert_pin(rlist,here);
					}
				}
				for(PinList *cp2=curr->input_of_trans_list; cp2!=NULL; cp2=cp2->next)
				{
					here = gate_has_been_found(cp2->NodeName);
					if(here!=NULL)
					{
						rlist = insert_pin(rlist, here);
					}
				}
			}
		}
	}
*/
	return rlist;
}

bool
Macromodel::query_node_in_nodestack(char *name)
{
	for(NodeStack *curr=node_stack; curr!=NULL; curr=curr->next)
	{
		if(curr->recognized_flag)
		{
			if(!strcmp(curr->circuit_name, name))
				return true;
		}
	}
	return false;
}

bool
Macromodel::query_gate_in_nodestack(char *name)
{
	for(GateStack *curr=gate_stack; curr!=NULL; curr=curr->next)
	{
		if(curr->recognized_flag)
		{
			if(!strcmp(curr->circuit_name, name))
				return true;
		}
	}
	return false;
}

//////// Query one node is in inputlist of the macromodel //////////
bool
Macromodel::query_in_inputlist(char *name)
{
	for(InoutList *curr=inputhead; curr!=NULL; curr=curr->next)
	{
		if(!strcmp(name, curr->pinname))
		{
			return true;
		}
	}
	return false;
}

bool
Macromodel::query_in_outputlist(char *name)
{
	for(InoutList *curr=outputhead; curr!=NULL; curr=curr->next)
	{
		if(!strcmp(name, curr->pinname))
		{
			return true;
		}
	}
	return false;
}


//// From a pin, find all the gates which connected with.
PinList *
Macromodel::return_found_gate_list_for_a_pin(char *pinname)
{
		PinList *rlist = NULL;
		for(NodeList *curr=node_list; curr!=NULL; curr=curr->next)
		{
			if(!strcmp(curr->nodename, pinname))
			{
				char *here=NULL;
				for(PinList *cp=curr->input_of_gate_list; cp!=NULL; cp=cp->next)
				{
					here = gate_has_been_found(cp->NodeName);
					if(here!=NULL)
					{
						rlist = insert_pin(rlist, here);
					}
					else
					{
						cout<<"What's going on? Some gates are not found yet! \n";
					}
				}

				char *cinput=NULL;
				char *coutput = NULL;
				if(curr->input_of_trans_name!=NULL)
					cinput = gate_has_been_found(curr->input_of_trans_name);
				if(curr->output_of_trans_name != NULL)
					coutput = gate_has_been_found(curr->output_of_trans_name);

				if(cinput != NULL)
					rlist = insert_pin(rlist, cinput);
				if(coutput!=NULL)
					rlist = insert_pin(rlist, coutput);
			/*
				for(PinList *cp2=curr->input_of_trans_list; cp2!=NULL; cp2=cp2->next)
				{
					here = gate_has_been_found(cp2->NodeName);
					if(here!=NULL)
					{
						rlist = insert_pin(rlist, here);
					}
					else
					{
						cout<<"What's going on? Some trans are not found yet! \n";
					}
				}
			*/
			}
		}
	return rlist;
}

// From here, if a node name of macromodel cell is given, I will find in the Relation_list
// of the p_branch and n_branch, find the relation_list the node belong to, and return the
// relation_list.

RelationList *
Macromodel::locate_relation_list(char *name, RelationList *rlist)
{
	for(RelationList *curr=rlist; curr!=NULL; curr=curr->next)
	{
		for(GateNameList *cg=curr->gatename_list; cg!=NULL; cg=cg->next)
		{
			if(!strcmp(name, cg->name))
				return curr;
		}
	}
	return NULL;
}

RelationList *
Macromodel::return_n_relation_list(char *name)
{
	for(NodeList *curr=node_list; curr!=NULL; curr=curr->next)
	{
		if(!strcmp(curr->nodename, name))
		{
				MCGate *hgate = hash_gate_list[return_number((curr->input_of_gate_list)->NodeName)];
				RelationList *rval = locate_relation_list(name, (hgate->n_branch)->relation_list);
				return rval;
/*
			if((curr->input_of_gate_list)==NULL || (curr->input_of_gate_list)->next!=NULL)
				cout<<"Wired, there should be only one gate connected to the node! \n";
			else if((curr->input_of_gate_list)->next == NULL)
			{
				MCGate *hgate = hash_gate_list[return_number((curr->input_of_gate_list)->NodeName)];
				RelationList *rval = locate_relation_list(name, (hgate->n_branch)->relation_list);

				return rval;
			}
			else
				;
*/
		}
	}
	return NULL;
}

RelationList *
Macromodel::return_p_relation_list(char *name)
{
	for(NodeList *curr=node_list; curr!=NULL; curr=curr->next)
	{
		if(!strcmp(curr->nodename, name))
		{
				MCGate *hgate = hash_gate_list[return_number((curr->input_of_gate_list)->NodeName)];
				RelationList *rval = NULL;
				rval = locate_relation_list(name, (hgate->p_branch)->relation_list);
				return rval;
/*
			if((curr->input_of_gate_list)==NULL || (curr->input_of_gate_list)->next!=NULL)
				cout<<"Wired, there should be only one gate connected to the node! \n";
			else if((curr->input_of_gate_list)->next == NULL)
			{
				MCGate *hgate = hash_gate_list[return_number((curr->input_of_gate_list)->NodeName)];
				RelationList *rval = NULL;
				rval = locate_relation_list(name, (hgate->p_branch)->relation_list);
				return rval;
			}
			else
				;
*/
		}
	}
	return NULL;
}

// End of the work to find the relation_list.

// From here, to verify the relation_list is valid or not. 
// What is valid, if in the relation_list, all other node(except the one we need) has been found,
// we think the list is a valid list.

GateNameList *
Macromodel::insert_gatename(GateNameList *head, char *name){
  GateNameList *current;
  if(head==NULL)
  {
        head=NLAlloc(GateNameList,1);
        head->name = NLAlloc(char,strlen(name));
        strcpy(head->name,name);
        head->next = NULL;
  }
  else {
        current=NLAlloc(GateNameList,1);
        current->name = NLAlloc(char,strlen(name));
        strcpy(current->name,name);
        current->next = head->next;
        head->next = current;
  }
  return head;
}


char *
Macromodel::return_node_from_nodestack(char *name)
{
	for(NodeStack *curr=node_stack; curr!=NULL; curr=curr->next)
	{
		if(curr->recognized_flag)
		{
			if(!strcmp(curr->macromodel_name, name))
			{
//				cout<<" t: "<<"m_name: "<<curr->macromodel_name<<"c_name: "<<curr->circuit_name<<endl;
				return curr->circuit_name;
			}
		}
	}
	return NULL;
}

GateNameList *
Macromodel::relation_list_is_valid(char *name, RelationList *rlist)
{
	GateNameList *gnlist=NULL;
	for(GateNameList *curr=rlist->gatename_list; curr!=NULL; curr=curr->next)
	{
//		cout<<" h: "<<curr->name<<endl;
		if(strcmp(name, curr->name))
		{
//			cout<<"hahaha: "<<curr->name<<endl;
			char *mre = return_node_from_nodestack(curr->name);
			if(mre != NULL)
				gnlist = insert_gatename(gnlist, mre);
			else
				return NULL;
		}
	}
	return gnlist;
}

void
Macromodel::reinitialize_macromodel()
{
	for(GateStack *cu=gate_stack; cu!=NULL; cu=cu->next)
	{
		cu->circuit_name = NULL;
		cu->recognized_flag = false;
		cu->search_flag = false;
	}
    for(NodeStack *curr=node_stack; curr!=NULL; curr=curr->next)
    {
		curr->recognized_flag = false;
		curr->search_flag = false;
		curr->circuit_name = NULL;
    }
}

PinList *
Macromodel::return_the_input_gate_list(char *name)
{
	for(NodeList *curr=node_list; curr!=NULL; curr=curr->next)
	{
		if(!strcmp(name, curr->nodename))
		{
			return curr->input_of_gate_list;
		}
	}
	return NULL;
}

///////////////////// deal with dead loop 2 ///////////////////////
// Another way to deal with dead loop
// Go to the gate_stack to see which gate did not found yet,
// if one gate did not found, go to see its input connect to which gate,
// if the "which gate" is found and search_flag=1, set search_flag=0
void
Macromodel::reset_search_flag(char *name)
{
	for(GateStack *curr=gate_stack; curr!=NULL; curr=curr->next)
	{
		if(!strcmp(name, curr->macromodel_name))
		{
			if(curr->search_flag)
			{
				curr->search_flag = false;
//				cout<<"              M name: "<<curr->macromodel_name<<" Sear falg: "<<curr->search_flag<<endl;
			}
		}
	}

}

void
Macromodel::deal_with_dead_loop2()
{
	for(GateStack *curr=gate_stack; curr!=NULL; curr=curr->next)
	{
		if(!curr->recognized_flag)
		{
//			cout<<" unreg name: "<<curr->macromodel_name<<endl;
			MCGate *mgate = hash_gate_list[return_number(curr->macromodel_name)];
			for(PinList *cp=mgate->inlist; cp!=NULL; cp=cp->next)
			{
//				cout<<"         inlist name: "<<cp->NodeName<<endl;
				for(NodeList *cn=node_list; cn!=NULL; cn=cn->next)
				{
					if(!strcmp(cn->nodename, cp->NodeName))
					{
						if(cn->output_of_gate_name!=NULL)
						{
//							cout<<"            con gate name: "<<cn->output_of_gate_name<<endl;
							reset_search_flag(cn->output_of_gate_name);
						}
					}
				}
			}
		}
	}
}

bool
Macromodel::node_not_found_yet(char *name)
{
	for(NodeStack *curr=node_stack; curr!=NULL; curr=curr->next)
	{
		if(!strcmp(name, curr->macromodel_name))
		{
			if(!curr->recognized_flag)
				return true;
			else
				return false;
			break;
		}
	}
}

bool
Macromodel::gatename_list_not_found_yet(GateNameList *gnlist)
{
	for(GateNameList *curr=gnlist; curr!=NULL; curr=curr->next)
	{
		if(!node_not_found_yet(curr->name))
			return false;
	}
	return true;
}

char *
Macromodel::find_circuit_gate_name(char *name)
{
	for(GateStack *curr=gate_stack; curr!=NULL; curr=curr->next)
	{
		if(!strcmp(name, curr->macromodel_name) && curr->recognized_flag)
			return curr->circuit_name;
	}
	return NULL;
}

void
Macromodel::insert_two_gatename_list(GateNameList *mlist, GateNameList *clist)
{
	GateNameList *currm, *prevm;
	GateNameList *currc, *prevc;
    currm=prevm=mlist;
    currc=prevc=clist;
    while(currm!=NULL && currc!=NULL)
    {
		insert_node_in_nodestack(currm->name, currc->name);
        prevm=currm;
        currm=currm->next;
        prevc=currc;
        currc=currc->next;
	}
}

// Copy a gate_stack

GateStack *
Macromodel::insert_copy_gate(GateStack *head, GateStack *gs)
{
    GateStack *current;
    if(head==NULL)
    {
	head = NLAlloc(GateStack, 1);
	head->type = gs->type;
	head->macromodel_name = NLAlloc(char, strlen(gs->macromodel_name));
	strcpy(head->macromodel_name, gs->macromodel_name);
	if(gs->circuit_name != NULL)
	{
		head->circuit_name = NLAlloc(char, strlen(gs->circuit_name));
		strcpy(head->circuit_name, gs->circuit_name);
	}
	else
		head->circuit_name = NULL;
	head->recognized_flag = gs->recognized_flag;
	head->search_flag = gs->search_flag;
	head->next = NULL;
    }
    else
    {
	current = NLAlloc(GateStack, 1);
	current->type = gs->type;
	current->macromodel_name = NLAlloc(char, strlen(gs->macromodel_name));
	strcpy(current->macromodel_name, gs->macromodel_name);
	if(gs->circuit_name != NULL)
	{
		current->circuit_name = NLAlloc(char, strlen(gs->circuit_name));
		strcpy(current->circuit_name, gs->circuit_name);
	}
	else
		current->circuit_name = NULL;
	current->recognized_flag = gs->recognized_flag;
	current->search_flag = gs->search_flag;
	current->next = head->next;
	head->next = current;
    }
    return head;
}

GateStack *
Macromodel::copy_gate_stack()
{
	GateStack *rev = NULL;
	for(GateStack *curr=gate_stack; curr!=NULL; curr=curr->next)
	{
		rev = insert_copy_gate(rev, curr);
	}
	return rev;
}

bool
Macromodel::compare_element_of_gate_stack(GateStack *ele, GateStack *new_gs)
{
	for(GateStack *curr=new_gs; curr!=NULL; curr=curr->next)
	{
		if(!strcmp(curr->macromodel_name, ele->macromodel_name))
		{
			if(curr->circuit_name == NULL)
			{
				if(ele->circuit_name==NULL && ele->recognized_flag==curr->recognized_flag && ele->search_flag==curr->search_flag)
					return true;
				else
					return false;
			}
			else
			{
			  if(ele->circuit_name!=NULL)
			  {
				if(!strcmp(ele->circuit_name,curr->circuit_name) && ele->recognized_flag==curr->recognized_flag && ele->search_flag==curr->search_flag)
					return true;
				else
					return false;
			  }
			  else
					return false;
			}
		}
	}
}


// Compare two gate_stack to see if they got into dead loop
bool
Macromodel::compare_two_gatestack(GateStack *old_gs, GateStack *new_gs)
{
	for(GateStack *curr=old_gs; curr!=NULL; curr=curr->next)
	{
		if(!compare_element_of_gate_stack(curr, new_gs))
			return false;
	}
	return true;
}

// To see if one gate have been found
bool
Macromodel::query_gate_has_been_found(char *name)
{
	for(GateStack *curr=gate_stack; curr!=NULL; curr=curr->next)
	{
		if(!strcmp(name, curr->circuit_name))
			return true;
	}
	return false;
}

void
Macromodel::set_trans_search_flag_one()
{
	for(GateStack *curr=gate_stack; curr!=NULL; curr=curr->next)
	{
		if(curr->recognized_flag)
		{
			if(!decide_a_gate(curr->macromodel_name))
				curr->search_flag = true;
		}
	}
}

void
Macromodel::set_trans_search_flag_zero()
{
	for(GateStack *curr=gate_stack; curr!=NULL; curr=curr->next)
	{
		if(curr->recognized_flag)
		{
			if(!decide_a_gate(curr->macromodel_name))
				curr->search_flag = false;
		}
	}
}

char *
Macromodel::query_macromodel_gate_name(char *name)
{
	for(GateStack *curr=gate_stack; curr!=NULL; curr=curr->next)
	{
		if(curr->recognized_flag)
		{
			if(!strcmp(curr->circuit_name, name))
				return curr->macromodel_name;
		}
	}
	return NULL;
}
/*** This function is added by Lei Yang, Mar 30,2003 **********************/
/*** This is an extrem case **********************************************/
bool
Macromodel::one_inlist_included_by_another_inlist(char *gate1, char *gate2)
{
	MCGate *small_gate = hash_gate_list[return_number(gate1)];
	MCGate *big_gate = hash_gate_list[return_number(gate2)];
	for(PinList *curr=small_gate->inlist; curr!=NULL; curr=curr->next)
	{
		for(PinList *curr2=big_gate->inlist; curr2!=NULL; curr2=curr2->next)
		{
			if(!strcmp(curr->NodeName, curr2->NodeName))
				return true;
		}
	}
	return false;
}
/********************* End add, Mar 30,2003*******************************/

