#include "pre.h"
#include "mnode.h"
#include <malloc.h>
#include <stdio.h>
#include <cstring>
using namespace std;

#define NLAlloc(type,num) (type *)calloc(num,sizeof(type))


MNode::MNode(MosList *moshead_, InoutList *inputhead_)
{
	moshead = moshead_;
	nodehead = NULL;
	inputhead = inputhead_;
	outputhead = NULL;
}

MNode::~MNode()
{
	if(moshead) delete moshead;
	if(nodehead) delete nodehead;
	if(inputhead) delete inputhead;
	if(outputhead) delete outputhead;
}

DeviceList *
MNode::insert_device(DeviceList *head, MosList *mos, PinType pintype) 
{
  DeviceList *current;
  if(head==NULL) 
  {
	head=NLAlloc(DeviceList,1);
	head->devicename = NLAlloc(char, strlen(mos->mosname));
	strcpy(head->devicename, mos->mosname);
	head->devicetype = mos->nptype;
	head->d = NLAlloc(char, strlen(mos->d));
	strcpy(head->d, mos->d);
	head->g = NLAlloc(char, strlen(mos->g));
        strcpy(head->g, mos->g);
	head->s = NLAlloc(char, strlen(mos->s));
        strcpy(head->s, mos->s);
	head->pintype = pintype;
	head->next = NULL;
  }
  else {
	current=NLAlloc(DeviceList,1);
	current->devicename = NLAlloc(char, strlen(mos->mosname));
	strcpy(current->devicename, mos->mosname);
	current->devicetype = mos->nptype;
	current->d = NLAlloc(char, strlen(mos->d));
	strcpy(current->d, mos->d);
	current->g = NLAlloc(char, strlen(mos->g));
        strcpy(current->g, mos->g);
	current->s = NLAlloc(char, strlen(mos->s));
        strcpy(current->s, mos->s);
	current->pintype = pintype;
	current->next = head->next;
	head->next = current;
  }
  return head;
}

NodeList *
MNode::insert_node(NodeList *head, DeviceList *devicelist, char *nodename){
  NodeList *current;
  if(head==NULL)
  {
	head=NLAlloc(NodeList,1);
	head->nodename = NLAlloc(char,strlen(nodename));
	strcpy(head->nodename,nodename);
	head->devicelist = devicelist;
	head->pn_node = 0;
	head->next = NULL;
  }
  else {
	current=NLAlloc(NodeList,1);
	current->nodename = NLAlloc(char,strlen(nodename));
	strcpy(current->nodename,nodename);
	current->devicelist = devicelist;
	current->pn_node = 0;
	current->next = head->next;
	head->next = current;
  }
  return head;
}

NodeList *
MNode::query_in_nodelist(char *name)
{
	for(NodeList *curr=nodehead; curr!=NULL; curr=curr->next)
	{
		if(!strcmp(name, curr->nodename))
			return curr;
	}
	return NULL;
}
	

void MNode::gen_nodelist()
{
  MosList *cur_mos;
  DeviceList *DeviceHead;
  for(cur_mos=moshead; cur_mos!=NULL; cur_mos=cur_mos->next)
  {
	NodeList *cur_node;

	if(nodehead==NULL) 
	{  
	   DeviceHead = insert_device(NULL,cur_mos,DP);
	   nodehead = insert_node(nodehead, DeviceHead, cur_mos->d);
	   DeviceHead = insert_device(NULL,cur_mos,SP);
	   nodehead = insert_node(nodehead, DeviceHead, cur_mos->s);
	   DeviceHead = insert_device(NULL,cur_mos,GP);
	   nodehead = insert_node(nodehead, DeviceHead, cur_mos->g);
	}
	else
	{
	   int d_flag=0;
	   int s_flag=0;
	   int g_flag=0;
	   for(cur_node=nodehead; cur_node!=NULL; cur_node=cur_node->next)
	   {
		if(!strcmp(cur_mos->d,cur_node->nodename))
		{
			cur_node->devicelist=insert_device(cur_node->devicelist,cur_mos,DP);
			d_flag=1;
		}
		if(!strcmp(cur_mos->s,cur_node->nodename))
		{
			cur_node->devicelist=insert_device(cur_node->devicelist,cur_mos,SP);
			s_flag=1;
		}
		if(!strcmp(cur_mos->g,cur_node->nodename))
		{
			cur_node->devicelist=insert_device(cur_node->devicelist,cur_mos,GP);
			g_flag=1;
		}
	   }
	   if(!d_flag)
	   {
		DeviceHead=insert_device(NULL,cur_mos,DP);
		nodehead=insert_node(nodehead,DeviceHead,cur_mos->d);
	   }
	   if(!s_flag)
	   {
	        DeviceHead=insert_device(NULL,cur_mos,SP);
                nodehead=insert_node(nodehead,DeviceHead,cur_mos->s);
           }
           if(!g_flag)
           {
                DeviceHead=insert_device(NULL,cur_mos,GP);
                nodehead=insert_node(nodehead,DeviceHead,cur_mos->g);
           }
	}
  }
}

void MNode::gen_np_char_netlist()
{
   NodeList *cur_node;
   DeviceList *cur_device;
   for(cur_node=nodehead; cur_node!=NULL; cur_node=cur_node->next)
   { 
	int p_flag=0;
	int n_flag=0;
	for(cur_device=cur_node->devicelist; cur_device!=NULL; cur_device=cur_device->next)
	{
	   if((cur_device->devicetype==0) && (cur_device->pintype!=GP))
		n_flag=1;
	   if((cur_device->devicetype==1) && (cur_device->pintype!=GP))
		p_flag=1;
	}
	
	if(p_flag && n_flag)
	{
		cur_node->pn_node = 1;
	}
   }
}

char *
MNode::get_vdd(MosList *moshead)
{
   char *vdd;
   MosList *curr = NULL;
   for(curr=moshead; curr!=NULL; curr=curr->next)
   {
        if(curr->nptype)
        {
          vdd = NLAlloc(char,strlen(curr->b));
          strcpy(vdd, curr->b);
          break;
        }
   }
   return vdd;
}

char *
MNode::get_gnd(MosList *moshead)
{
   char *gnd;
   MosList *curr = NULL;
   for(curr=moshead; curr!=NULL; curr=curr->next)
   {
        if(!curr->nptype)
        {
          gnd = NLAlloc(char,strlen(curr->b));
          strcpy(gnd, curr->b);
          break;
        }
   }
   return gnd;
}

void
MNode::delete_vdd_gnd_from_inputlist()
{
    char *vdd = get_vdd(moshead);
    char *gnd = get_gnd(moshead);


//    printf("%s %s", vdd, gnd);
    InoutList *curr, *prev;
    curr = inputhead->next;
    prev = inputhead;
    while(curr!=NULL)
    {
        if(!strcmp(curr->pinname, vdd) || !strcmp(curr->pinname, gnd))
        {
            prev->next = curr->next;
            curr = prev->next;
        }
        else
        {
            prev=curr;
            curr=curr->next;
        }
    }
}


InoutList *
MNode::insert_output(InoutList *head, InoutList *curr)
{
    InoutList *current;
    if(head==NULL)
    {
	head=NLAlloc(InoutList,1);
	head->pinname = NLAlloc(char, strlen(curr->pinname));
	strcpy(head->pinname, curr->pinname);
	head->next=NULL;
    }
    else
    {
	current=NLAlloc(InoutList,1);
	current->pinname = NLAlloc(char, strlen(curr->pinname));
	strcpy(current->pinname, curr->pinname);
	current->next = head->next;
	head->next = current;
    }
    return head;
}
	
void
MNode::generate_output_pinlist()
{
    for(InoutList *curr=inputhead; curr!=NULL; curr=curr->next)
    {
	int flag=0;
	for(NodeList *curr_node=nodehead; curr_node!=NULL; curr_node=curr_node->next)
	{
	    if(!strcmp(curr_node->nodename, curr->pinname))
	    {
		for(DeviceList *curr_dev=curr_node->devicelist; curr_dev!=NULL; curr_dev=curr_dev->next)
		{
		    if(curr_dev->pintype!=GP)
		    {
			flag=1;
			break;
		    }
		}
	    }
	}
	if(flag)
	    outputhead = insert_output(outputhead, curr);
    }
}

void
MNode::generate_input_pinlist()
{
    InoutList *curr_input, *prev_input;
    prev_input = inputhead;
    curr_input = prev_input->next;
    while(curr_input!=NULL)
    {
	int flag=0;
	for(InoutList *curr_out=outputhead; curr_out!=NULL; curr_out=curr_out->next)
	{
	    if(!strcmp(curr_out->pinname, curr_input->pinname))
	    {
		flag=1;
		break;
	    }
	}
	if(flag)
	{
	    prev_input->next = curr_input->next;
	    curr_input = prev_input->next;
	}
	else
	{
	    prev_input = curr_input;
	    curr_input = curr_input->next;
	}
    }
}

/*

void
MNode::generate_input_output_pinlist()
{
    InoutList *curr_input, *prev_input;
    curr_input = inputhead->next;
    prev_input = inputhead;
    while(curr_input!=NULL)
    {
	int find_it=0;
	for(NodeList *curr_node=nodehead; curr_node!=NULL; curr_node=curr_node->next)
	{
	    if(!strcmp(curr_node->nodename, curr_input->pinname))
	    {
		find_it=1;
		int flag= 0;
		for(DeviceList *curr_dev=curr_node->devicelist; curr_dev!=NULL; curr_dev=curr_dev->next)
		{
		    if(curr_dev->pintype!=GP)
			flag=1;
		}
		
		if(!flag)
		{
		    prev_input=curr_input;
		    curr_input=curr_input->next;    
		}
		else
		{
		  if(curr_input->next==NULL)
		  {
		    outputhead = insert_output(outputhead, curr_input);
		    curr_input=NULL;
		    prev_input->next = NULL;
		  }
		  else
		  {
		    outputhead = insert_output(outputhead, curr_input);
		    prev_input->next = curr_input->next;
		    curr_input = prev_input->next;
		  }
		}    
	    }
	}
	if(!find_it)
	{
	    prev_input=curr_input;
            curr_input=curr_input->next;
	}
    }
}
*/
