//------------------------------------------------------------------
//  patternmatch.c
//  Pattern match process between a macromodel and the whole circuit
//
//  Lei Yang       May 5, 2002
//------------------------------------------------------------------

#include "pre.h"
#include "tellnp.h"
#include "node.h"
#include "branch.h"
#include "group.h"
#include "cgate.h"
#include "macromodel.h"
#include "patternmatch.h"
#include <malloc.h>
#include <stdio.h>
#include <string>
#include <cstring>

extern ResultList *ResultHead;

#define NLAlloc(type,num) (type *)calloc(num,sizeof(type))
#define ionumber 500

extern TransList *TransHead;

extern InoutList *InputHead, *OutputHead;

extern int device_biggest;
extern int node_biggest;
extern PMosList *hash_mos_list;
extern PNodeList *hash_node_list;

extern int num_of_gates;
extern int num_of_trans;
extern PCGate *hash_gate_list;
extern PTransList *hash_trans_list;

int gigi=0;


//////////////////// Patternmatch ///////////////////////
// construct class patternmatch

Patternmatch::Patternmatch(Macromodel *a_macromodel)
{
    macromodel = a_macromodel;
}

/////////////////// ~Patternmatch ///////////////////////
// destruct class patternmatch

Patternmatch::~Patternmatch()
{
    if(macromodel) delete macromodel;
}

// To tell a gate is a gate or transmission_gate
bool
Patternmatch::decide_a_gate(char *name)
{
    if(name[0]=='g' && name[1]=='a' && name[2]=='t' && name[3]=='e')
        return true;
    else
        return false;
}

// Given a node string name, I will go to hash_node_list to return a *node
int
Patternmatch::return_number(char *name)
{
	if(name!=NULL)
	{
        char *st = NLAlloc(char, strlen(name));
        strcpy(st, name);
        char *ret=NULL;
        ret = NLAlloc(char, strlen(st));
        int counter = 0;

        bool begin_f=false;
        bool end_f=false;

        int i=0;
        while(st[i]!='\0')
        {
                if(st[i]>=48 && st[i]<=57 && st[i]!='\0')
                {
                        begin_f = true;
                        ret[counter]=st[i];
                        counter+=1;
                }
                if(begin_f && st[i]!='\0' && ( st[i]<48 || st[i]>57 ))
                {
                        end_f = true;
                        break;
                }
                i+=1;
        }
        ret[counter] =  '\0';

        if(begin_f && !end_f)
        {
                return atoi(ret);
        }
        else
        {
                return -1;
        }
	}
	else
		return -1;
}

NodeList *
Patternmatch::get_node(char *nodename)
{
        int rval = return_number(nodename);
        if(rval != -1)
                return hash_node_list[rval];
        else
        {
                for(int i=node_biggest+1; i<(node_biggest+ionumber); i++)
                {
                  if(hash_node_list[i] != NULL)
                  {
                        if(!strcmp(nodename, hash_node_list[i]->nodename))
                                return hash_node_list[i];
                  }
                }
        }
}

/////////////////// Insert a pin in a pinlist ///////////////////////////////////
// Insert a pin in a pinlist
RGList *
Patternmatch::insert_return_gate(RGList *head, char *name, int type){
  RGList *current;
  if(head==NULL)
  {
        head=NLAlloc(RGList,1);
        head->gatename = NLAlloc(char,strlen(name));
        strcpy(head->gatename,name);
		head->type = type;
		head->unique = true;
// New added
		head->unique2 = true;
        head->next = NULL;
  }
  else {
        current=NLAlloc(RGList,1);
        current->gatename = NLAlloc(char,strlen(name));
        strcpy(current->gatename,name);
		current->type = type;
		current->unique = true;
// New added
		current->unique2 = true;
        current->next = head->next;
        head->next = current;
  }
  return head;
}

/////////////////// Search from a gate //////////////////////////////////////////
// Search from a gate, the output node of this gate will go to the input
// of other gates, these gates will be linked to be a gate_list, return this gate_list.
RGList *
Patternmatch::return_a_gate_list(char *gatename)
{
	RGList *glist = NULL;
		CGate *hgate = hash_gate_list[return_number(gatename)];
		NodeList *hnode = get_node((hgate->n_branch)->pn_node);
				for(PinList *cup=hnode->input_of_gate_list; cup!=NULL; cup=cup->next)
				{
					glist = insert_return_gate(glist, cup->NodeName, 0);
				}

				if(hnode->input_of_trans_name!=NULL)
					glist = insert_return_gate(glist, hnode->input_of_trans_name, 1);
				if(hnode->output_of_trans_name!=NULL)
					glist = insert_return_gate(glist, hnode->output_of_trans_name,1);
			/*
				for(PinList *cup2=hnode->input_of_trans_list; cup2!=NULL; cup2=cup2->next)
				{
					glist = insert_return_gate(glist, cup2->NodeName, 1);
				}
			*/
				return glist;
}
/********************** Added by Lei Yang ******************************/
/********************** Mar 30, 2003 ***********************************/
/**** Purpose: trace from trans gate to other gates ********************/
char * 
Patternmatch::query_name_in_nodestack(char *name)
{
	for(NodeStack *curr=macromodel->node_stack; curr!=NULL; curr=curr->next)
	{
		if(curr->recognized_flag && curr->circuit_name!=NULL)
		{
			if(!strcmp(curr->circuit_name, name))
				return curr->circuit_name;
		}
	}
	return NULL;
}

// If only input or only output has been found in nodestack, return it.
char *
Patternmatch::one_valid_terminal_of_trans(char *gatename)
{
	TransList *htrans = hash_trans_list[return_number(gatename)];
	if(query_name_in_nodestack(htrans->input)!=NULL && query_name_in_nodestack(htrans->output)==NULL)
		return htrans->output;
	else if(query_name_in_nodestack(htrans->input)==NULL && query_name_in_nodestack(htrans->output)!=NULL)
		return htrans->input;
	else 
		return NULL;
}

RGList *
Patternmatch::return_a_trans_rglist(char *gatename)
{
	RGList *glist = NULL;
	if(one_valid_terminal_of_trans(gatename)!=NULL)
	{
		char *here = NLAlloc(char, strlen(one_valid_terminal_of_trans(gatename)));
		strcpy(here, one_valid_terminal_of_trans(gatename));
		NodeList *tnode = get_node(here);

		for(PinList *cup=tnode->input_of_gate_list; cup!=NULL; cup=cup->next)
		{
			glist = insert_return_gate(glist, cup->NodeName, 0);
		}
		if(tnode->output_of_gate_name != NULL)
			glist = insert_return_gate(glist, tnode->output_of_gate_name, 0);

		if(tnode->input_of_trans_name != NULL)
			glist = insert_return_gate(glist, tnode->input_of_trans_name,1);
		if(tnode->output_of_trans_name != NULL)
			glist = insert_return_gate(glist, tnode->output_of_trans_name,1);
		return glist;
	}
	else
	{
		return NULL;
	}
}

/********************  End *************************************************/

RGList *
Patternmatch::return_a_trans_output_list(char *gatename)
{
	RGList *glist = NULL;
		TransList *htrans = hash_trans_list[return_number(gatename)];
		NodeList *tnode = get_node(htrans->output);
				for(PinList *cup=tnode->input_of_gate_list; cup!=NULL; cup=cup->next)
				{
					glist = insert_return_gate(glist, cup->NodeName, 0);
				}

				if(tnode->input_of_trans_name != NULL)
					glist = insert_return_gate(glist, tnode->input_of_trans_name,1);
				if(tnode->output_of_trans_name != NULL)
					glist = insert_return_gate(glist, tnode->output_of_trans_name,1);
			/*
				for(PinList *cup2=tnode->input_of_trans_list; cup2!=NULL; cup2=cup2->next)
				{
					glist = insert_return_gate(glist, cup2->NodeName, 1);
				}
			*/
				return glist;
}

RGList *
Patternmatch::return_a_trans_input_list(char *gatename)
{
	RGList *glist = NULL;
		TransList *htrans = hash_trans_list[return_number(gatename)];
		NodeList *tnode = get_node(htrans->input);
				for(PinList *cup=tnode->input_of_gate_list; cup!=NULL; cup=cup->next)
				{
					glist = insert_return_gate(glist, cup->NodeName, 0);
				}

				if(tnode->input_of_trans_name != NULL)
					glist = insert_return_gate(glist, tnode->input_of_trans_name,1);
				if(tnode->output_of_trans_name != NULL)
					glist = insert_return_gate(glist, tnode->output_of_trans_name,1);
			/*
				for(PinList *cup2=tnode->input_of_trans_list; cup2!=NULL; cup2=cup2->next)
				{
					glist = insert_return_gate(glist, cup2->NodeName, 1);
				}
			*/
				return glist;
}

void
Patternmatch::unique_rglist(RGList *glist)
{
  if(glist!=NULL)
  {
	for(RGList *curr=glist; curr!=NULL; curr=curr->next)
	{
		for(GateStack *curr_g=macromodel->gate_stack; curr_g!=NULL; curr_g=curr_g->next)
		{
		  if(curr_g->recognized_flag)
		  {
			if(!strcmp(curr_g->circuit_name, curr->gatename))
			{
					curr->unique = false;
// New added
					curr->unique2 = false;
				break;
			}
		  }
		}
	}
  if(glist->next!=NULL)
  {
	for(RGList *curr2=glist; curr2!=NULL; curr2=curr2->next)
	{
		if(curr2->unique)
		{
			for(RGList *curr3=curr2->next; curr3!=NULL; curr3=curr3->next)
			{
			  if(curr3->unique)
			  {
				if(!curr2->type && !curr3->type)
				{
					CGate *h1 = hash_gate_list[return_number(curr2->gatename)];
					CGate *h2 = hash_gate_list[return_number(curr3->gatename)];
				  if(h1!=NULL && h2!=NULL && (h1->n_branch)->headername!=NULL &&(h1->p_branch)->headername!=NULL && (h2->n_branch)->headername!=NULL && (h2->p_branch)->headername!=NULL)
				  {
					if(!strcmp((h1->n_branch)->headername, (h2->n_branch)->headername) && !strcmp((h1->p_branch)->headername, (h2->p_branch)->headername))
					{
						curr2->unique = false;
						curr3->unique = false;
// New added
						curr2->unique2 = true;
						curr3->unique2 = true;
					}
				  }
				  else
				  {
					cout<<" There is NULL gate in hash_gate_list! \n";
				  }
				}
				if(curr2->type && curr3->type)
				{
					curr2->unique = false;
					curr3->unique = false;
// New added
					curr2->unique2 = false;
					curr3->unique2 = false;
				}
			}
		  }
		}
	}
  }
  }
}

// If we find a gate in Macromodel matches one gate in Circuit, we will fill in their output node in
// the node_stack. 
void
Patternmatch::fill_gate_node_in_stack(char *mname, char *cname)
{
  if(decide_a_gate(mname) && decide_a_gate(cname))
  {
	MCGate *mgate = macromodel->hash_gate_list[return_number(mname)];
	CGate *cgate = hash_gate_list[return_number(cname)];
//	cout<<" Fill in: m "<<(mgate->n_branch)->pn_node<<" c "<<(cgate->n_branch)->pn_node<<endl;
	macromodel->insert_node_in_nodestack((mgate->n_branch)->pn_node, (cgate->n_branch)->pn_node);
// If the two gates are INV, fill their input into the nodestack.

	if(!strcmp((mgate->n_branch)->headername, "INV"))
	{
		if(!strcmp(((mgate->n_branch)->name_list)->gatename, ((mgate->p_branch)->name_list)->gatename) && !strcmp(((cgate->n_branch)->name_list)->gatename, ((cgate->p_branch)->name_list)->gatename))
		{
//				cout<<" x Fill in: m "<<((mgate->n_branch)->name_list)->gatename<<" c "<<((cgate->n_branch)->name_list)->gatename<<endl;
			macromodel->insert_node_in_nodestack(((mgate->n_branch)->name_list)->gatename, ((cgate->n_branch)->name_list)->gatename);
		}
		else
			cout<<"INV's N_branch and P_branch has different input name!\n";
	}

  }
}

void
Patternmatch::fill_trans_node_in_stack(char *mname, char *cname)
{
	TransList *mtrans = macromodel->hash_trans_list[return_number(mname)];
	TransList *ctrans = hash_trans_list[return_number(cname)];
	macromodel->insert_node_in_nodestack(mtrans->input, ctrans->input);
	macromodel->insert_node_in_nodestack(mtrans->output, ctrans->output);
	macromodel->insert_node_in_nodestack(mtrans->p_g, ctrans->p_g);
	macromodel->insert_node_in_nodestack(mtrans->n_g, ctrans->n_g);
}


// Based on the unique situation, fill them in the gate_stack
// If a gate in Macromodel_gatelist, if it is unique, find the corresponding one in circuit_gatelist,
// and fill them in gatestack.
// If a gate in the Macromodel_gatelist, we can't find it in circuit_gatelist, return false.

int
Patternmatch::count_num_of_rglist(RGList *glist)
{
	int num=0;
	for(RGList *curr=glist; curr!=NULL; curr=curr->next)
		num += 1;
	return num;
}

int
Patternmatch::count_num_of_a_list(PinList *plist)
{
	int num=0;
	for(PinList *curr=plist; curr!=NULL; curr=curr->next)
	{
		num+=1;
	}
	return num;
}

bool
Patternmatch::decide_two_pinlist_same(char *mnode_name, char *cnode_name)
{
// First check if mnode and cnode connect to trans, if one has trans, another
// one did not, return false.
	bool cyou = false;
	bool myou = false;

	bool cont = true;

	NodeList *cc = get_node(cnode_name);
	if(cc->input_of_trans_name!=NULL || cc->output_of_trans_name!=NULL)
		myou = true;

	for(NodeList *nl=macromodel->node_list; nl!=NULL; nl=nl->next)
	{
		if(!strcmp(mnode_name, nl->nodename))
		{
			if(nl->input_of_trans_name!=NULL || nl->output_of_trans_name!=NULL)
				cyou = true;
			break;
/*
			if(cc->input_of_trans_name!=NULL || cc->output_of_trans_name!=NULL)
			{
				if(nl->input_of_trans_name!=NULL || nl->output_of_trans_name!=NULL)
					cont = true;
				else
					cont = false;
			}
			else
			{
				if(nl->input_of_trans_name==NULL && nl->output_of_trans_name==NULL)
					cont = true;
				else
					cont = false;
			}
*/
		}
	}

	if(myou && cyou)
		cont=true;
	else if(!myou && !cyou)
		cont = true;
	else
		cont = false;

// Check the gate list match or not.
if(cont)
{
	PinList *mlist = macromodel->return_the_input_gate_list(mnode_name);
	PinList *clist = get_node(cnode_name)->input_of_gate_list;
	int mint = count_num_of_a_list(mlist);
	int cint = count_num_of_a_list(clist);
	if(mint != cint)
		return false;
	else
	{
		for(PinList *curr=mlist; curr!=NULL; curr=curr->next)
		{
			MCGate *mgate = macromodel->hash_gate_list[return_number(curr->NodeName)];
			for(PinList *currc=clist; currc!=NULL; currc=currc->next)
			{
				CGate *cgate = hash_gate_list[return_number(currc->NodeName)];
				if(strcmp((mgate->n_branch)->headername, (cgate->n_branch)->headername) || strcmp((mgate->p_branch)->headername, (cgate->p_branch)->headername))
					return false;
			}
		}
		return true;
	}
}
else
	return false;
}

bool
Patternmatch::decide_failed(RGList *mglist, RGList *cglist, int m_num, int c_num, bool ter)
{
  if(m_num==1)
		return true;
  else
  {
	char *repeat_n = NULL;
	char *repeat_p = NULL;
	bool exist = false;
	int record = 0;
	for(RGList *cr=mglist; cr!=NULL; cr=cr->next)
	{
	  if(decide_a_gate(cr->gatename))
	  {
		int count = 1;
		MCGate *mgate=macromodel->hash_gate_list[return_number(cr->gatename)];
		for(RGList *cr2=cr->next; cr2!=NULL; cr2=cr2->next)
		{
		  if(decide_a_gate(cr2->gatename))
		  {
			MCGate *mgate2 = macromodel->hash_gate_list[return_number(cr2->gatename)];
			if(!strcmp((mgate->n_branch)->headername, (mgate2->n_branch)->headername) && !strcmp((mgate->p_branch)->headername, (mgate2->p_branch)->headername))
			{
				repeat_n = NLAlloc(char, strlen((mgate->n_branch)->headername));
				strcpy(repeat_n, (mgate->n_branch)->headername);
				repeat_p = NLAlloc(char, strlen((mgate->p_branch)->headername));
				strcpy(repeat_p, (mgate->p_branch)->headername);
				count += 1;
			}
		  }
		}
		if(count>1)
		{
//			cout<<" count1: "<<count<<endl;
			exist = true;
			record = count;
			break;
		}
	  }
	}
	
  if(exist)
  {
	int num=0;
	for(RGList *cc=cglist; cc!=NULL; cc=cc->next)
	{
	  if(decide_a_gate(cc->gatename))
	  {
		CGate *cgate=hash_gate_list[return_number(cc->gatename)];
		if(cgate!=NULL)
		{
			if(!strcmp((cgate->n_branch)->headername, repeat_n) && !strcmp((cgate->p_branch)->headername, repeat_p))
				num+=1;
		}
	  }
	}
//			cout<<" count2: "<<num<<endl;
	if(record == num)
		return true;
	else
		return false;
  }
  else
	return true;
  }
}

/********************* Added by Lei, Apr 2, 2003 ************************************************/
// The above can not guarantee the right match, such as SDFH0 in CEGRP, so we should use this method: find the matched node in its inlist, decide if this node is only in P part or N part or both P & N part.
bool
Patternmatch::node_in_name_list(NameList *nlist, char *name)
{
	for(NameList *curr=nlist; curr!=NULL; curr=curr->next)
	{
		if(!strcmp(name, curr->gatename))
			return true;
	}
	return false;
}

bool 
Patternmatch::node_in_right_part(char *mnode, MCGate *mgate, char *cnode, CGate *cgate)
{
	if(node_in_name_list((mgate->n_branch)->name_list, mnode) && node_in_name_list((mgate->p_branch)->name_list, mnode))
	{
		if(node_in_name_list((cgate->n_branch)->name_list,cnode) && node_in_name_list((cgate->p_branch)->name_list,cnode))
			return true;
		else
			return false;
	}
	else if(node_in_name_list((mgate->n_branch)->name_list, mnode) && !node_in_name_list((mgate->p_branch)->name_list, mnode))
	{
		if(node_in_name_list((cgate->n_branch)->name_list,cnode) && !node_in_name_list((cgate->p_branch)->name_list,cnode))
			return true;
		else
			return false;
	}
	else if(!node_in_name_list((mgate->n_branch)->name_list, mnode) && node_in_name_list((mgate->p_branch)->name_list, mnode))
	{
		if(!node_in_name_list((cgate->n_branch)->name_list,cnode) && node_in_name_list((cgate->p_branch)->name_list,cnode))
			return true;
		else
			return false;
	}
	else
		return false;
}

bool
Patternmatch::match_inlist_node_in_P_or_N_part(MCGate *mgate, CGate *cgate)
{
	for(PinList *curr=mgate->inlist; curr!=NULL; curr=curr->next)
	{
		if(macromodel->return_circuit_name_from_nodestack(curr->NodeName)!=NULL)
		{
			char *corr = macromodel->return_circuit_name_from_nodestack(curr->NodeName);
			if(!node_in_right_part(curr->NodeName,mgate,corr,cgate))
				return false;
		}
	}
	return true;
}

/********************* Added by Lei, Apr 2, 2003 ************************************************/


bool
Patternmatch::fill_in_gate_stack(RGList *mglist, RGList *cglist, bool ter)
{
  int m_num=count_num_of_rglist(mglist);
  int c_num=count_num_of_rglist(cglist);

/*
  cout<<"ter :"<<ter<<" m_num: "<<m_num<<" c_num: "<<c_num<<endl;
  cout<<" MGlist ===========\n";
  for(RGList *cm=mglist; cm!=NULL; cm=cm->next)
	  cout<<" "<<cm->gatename<<" Unique: "<<cm->unique;
  cout<<endl;
  cout<<" CGlist ===========\n";
  for(RGList *cm=cglist; cm!=NULL; cm=cm->next)
	  cout<<" "<<cm->gatename<<" Unique: "<<cm->unique;
  cout<<endl;
*/

  if((!ter && m_num==c_num) || (ter && c_num>=m_num))
  {
   if(decide_failed(mglist, cglist, m_num, c_num, ter))
   {
	for(RGList *currm=mglist; currm!=NULL; currm=currm->next)
	{
// For every gate in macromodel_list, set a flag to see if the corresponding gate can be found in circuit.
		bool success = false;
		if(!currm->type)
		{
			MCGate *mgate = macromodel->hash_gate_list[return_number(currm->gatename)];
			if(currm->unique && currm->unique2)
			{
				for(RGList *currc=cglist; currc!=NULL; currc=currc->next)
				{
				  if(!currc->type)
				  {
					CGate *cgate = hash_gate_list[return_number(currc->gatename)];
				  if(mgate != NULL && cgate!=NULL )
				  {
					if(!strcmp((mgate->n_branch)->headername, (cgate->n_branch)->headername) && !strcmp((mgate->p_branch)->headername, (cgate->p_branch)->headername))
					{
						success = true;
						if(currc->unique)
						{
							macromodel->insert_gates_in_gatestack(currm->gatename, currc->gatename);
							fill_gate_node_in_stack(currm->gatename, currc->gatename);
//								cout<<" Fill in MGate: "<<currm->gatename<<" CGate: "<<currc->gatename<<endl;
						}
					}
				  }
				  else
				  {
					if(mgate == NULL)
						cout<<" Macro gate NULL: "<<endl;
					else if(cgate == NULL)
						cout<<" Circuit gate NULL: "<<endl;
					else
						;
				  }
				 }
				}
			}
			else if(!currm->unique && currm->unique2)
			{
				for(RGList *currc=cglist; currc!=NULL; currc=currc->next)
				{
				  if(!currc->type)
				  {
					CGate *cgate = hash_gate_list[return_number(currc->gatename)];
					if(!strcmp((mgate->n_branch)->headername, (cgate->n_branch)->headername) && !strcmp((mgate->p_branch)->headername, (cgate->p_branch)->headername))
					{
						success = true;
						if(!currc->unique && currc->unique2)
						{
							if(!macromodel->query_is_terminal((mgate->n_branch)->pn_node))
							{
// The following is, you have (mgate->n_branch)->pn_node's connected gates in macromodel, and you have 
// (cgate->b_branch)->pn_node's connected gates in circuit. Decide the two gate_list is the same.
								bool hflag = decide_two_pinlist_same((mgate->n_branch)->pn_node, (cgate->n_branch)->pn_node);
/********************* Added by Lei, Apr 2, 2003 ************************************************/
// The above can not guarantee the right match, such as SDFH0 in CEGRP, so we should use this method: find the matched node in its inlist, decide if this node is only in P part or N part or both P & N part.
								bool h2flag = match_inlist_node_in_P_or_N_part(mgate, cgate);
/********************* Added by Lei, Apr 2, 2003 ************************************************/
								if(hflag && h2flag)
								{
									macromodel->insert_gates_in_gatestack(currm->gatename, currc->gatename);
									fill_gate_node_in_stack(currm->gatename, currc->gatename);
//										cout<<" Fill in 2 MGate: "<<currm->gatename<<" CGate: "<<currc->gatename<<endl;
								}
							}
							else
							{
// Maybe we can do something here.
								;
							}
						}
					}
				  }
				}
			}
			else if(!currm->unique && !currm->unique2)
			{
				for(RGList *currc=cglist; currc!=NULL; currc=currc->next)
				{
				  if(!currc->type)
				  {
					CGate *cgate = hash_gate_list[return_number(currc->gatename)];
					if(!strcmp((mgate->n_branch)->headername, (cgate->n_branch)->headername) && !strcmp((mgate->p_branch)->headername, (cgate->p_branch)->headername))
					{
						success = true;
						break;
					}
				  }
				}
			}
			else
				;
		}
		else
		{
			if(currm->unique)
			{
				for(RGList *currc=cglist; currc!=NULL; currc=currc->next)
				{
					if(currc->type)
					{
						success = true;
						if(currc->unique)
						{
							macromodel->insert_gates_in_gatestack(currm->gatename, currc->gatename);
//							fill_trans_node_in_stack(currm->gatename, currc->gatename);
//							cout<<" Insert: "<<currm->gatename<<" "<<currc->gatename<<endl;
						}
					}
				}
			}
			else
			{
				for(RGList *currc=cglist; currc!=NULL; currc=currc->next)
				{
					if(currc->type)
					{
						success = true;
						break;
					}
				}
			}
		}
		
		if(!success)
			return false;
	}
	return true;
	}
	else
	{
		cout<<" decided fail \n";
		return false;
	}
  }
  else
  {
//	  cout<<"not match, exit! \n"<<endl;
	return false;
  }
}


//////////////////////////////// step ////////////////////////////
///// every step in the match process
int
Patternmatch::step()
{
    int stop = 1;
	GateStack *curr_gate=macromodel->gate_stack;
	while(macromodel->gatestack_search_to_end() && curr_gate!=NULL && stop)
    {
	    if(curr_gate->recognized_flag==1 && curr_gate->search_flag==0)
	    {
//		    cout<<" Search from M gate: "<<curr_gate->macromodel_name<<"C gate: "<<curr_gate->circuit_name<<endl;
			curr_gate->search_flag = 1;

		bool gate_or_trans = decide_a_gate(curr_gate->macromodel_name);

		if(gate_or_trans)
		{

			MCGate *tgate = macromodel->hash_gate_list[return_number(curr_gate->macromodel_name)];

			RGList *mglist = macromodel->return_a_gate_list(curr_gate->macromodel_name);	
			macromodel->unique_rglist(mglist);
/*
//if(gigi)
//{
//		if(!strcmp(macromodel->name, "DFFP"))
//		{
			cout<<"+++connect m gate: \n";
			for(RGList *ct=mglist; ct!=NULL; ct=ct->next)
			{
				MCGate *pgate = macromodel->hash_gate_list[return_number(ct->gatename)];
				cout<<" m_gate: "<<ct->gatename<<" unique: "<<ct->unique<<" unique2: "<<ct->unique2<<" Headername: "<<(pgate->n_branch)->headername<<endl;
			}
//		}
//}
*/
			
			

			RGList *cglist = return_a_gate_list(curr_gate->circuit_name);
			unique_rglist(cglist);		
/*
//if(gigi)
//{
//		if(!strcmp(macromodel->name, "DFFP"))
//		{
			cout<<"+++connect c gate: \n";
			if(cglist!=NULL)
			{
				for(RGList *ct=cglist; ct!=NULL; ct=ct->next)
				{
					CGate *qgate = hash_gate_list[return_number(ct->gatename)];
					if(qgate!=NULL )
						cout<<" c_gate: "<<ct->gatename<<" unique: "<<ct->unique<<" unique2: "<<ct->unique2<<" Headername: "<<(qgate->n_branch)->headername<<endl;
				}
			}
			else
				cout<<"Circuit rglist: NULL"<<endl;
//		}
//		gigi=0;
//}
*/
			bool ter = macromodel->query_is_terminal((tgate->n_branch)->pn_node);
			bool ht = fill_in_gate_stack(mglist, cglist, ter);

			if(ht)
				stop = 1;
			else
				stop = 0;
		}
		else
		{
/********************  Added by Lei Yang, Mar 30, 2003 **************************************/
/********    Trace from transmission gate to other gates ************************************/
		  if(macromodel->one_valid_terminal_of_trans(curr_gate->macromodel_name)!=NULL && one_valid_terminal_of_trans(curr_gate->circuit_name)!=NULL)
		  {
			RGList *mglist = macromodel->return_a_trans_rglist(curr_gate->macromodel_name);
			RGList *cglist = return_a_trans_rglist(curr_gate->circuit_name);
			macromodel->unique_rglist(mglist);
			unique_rglist(cglist);

/*
			cout<<"Search from M:"<<macromodel->one_valid_terminal_of_trans(curr_gate->macromodel_name)<<endl;

			cout<<"+++connect m gate: \n";
			for(RGList *ct=mglist; ct!=NULL; ct=ct->next)
			{
				MCGate *pgate = macromodel->hash_gate_list[return_number(ct->gatename)];
				cout<<" m_gate: "<<ct->gatename<<" unique: "<<ct->unique<<" unique2: "<<ct->unique2<<" Headername: "<<(pgate->n_branch)->headername<<endl;
			}


			cout<<"Search from C:"<<one_valid_terminal_of_trans(curr_gate->circuit_name)<<endl;


			cout<<"+++connect c gate: \n";
			if(cglist!=NULL)
			{
				for(RGList *ct=cglist; ct!=NULL; ct=ct->next)
				{
					CGate *qgate = hash_gate_list[return_number(ct->gatename)];
					if(qgate!=NULL )
						cout<<" c_gate: "<<ct->gatename<<" unique: "<<ct->unique<<" unique2: "<<ct->unique2<<" Headername: "<<(qgate->n_branch)->headername<<endl;
				}
			}
			else
				cout<<"Circuit rglist: NULL"<<endl;
*/


			if(mglist!=NULL && cglist!=NULL)
			{
				bool ter = macromodel->query_is_terminal(macromodel->one_valid_terminal_of_trans(curr_gate->macromodel_name));
				bool ht = fill_in_gate_stack(mglist, cglist, ter);

				if(ht)
					stop = 1;
				else
					stop = 0;
			}
		  }
/*
			TransList *htrans = macromodel->hash_trans_list[return_number(curr_gate->macromodel_name)];

			RGList *miglist = macromodel->return_a_trans_input_list(curr_gate->macromodel_name);
			macromodel->unique_rglist(miglist);

			RGList *moglist = macromodel->return_a_trans_input_list(curr_gate->macromodel_name);
			macromodel->unique_rglist(moglist);

			RGList *ciglist = return_a_trans_input_list(curr_gate->circuit_name);
			unique_rglist(ciglist);		

			RGList *coglist = return_a_trans_input_list(curr_gate->circuit_name);
			unique_rglist(coglist);		


			bool teri = macromodel->query_is_terminal(htrans->input);
			bool tero = macromodel->query_is_terminal(htrans->output);


			bool me1 = fill_in_gate_stack(miglist,ciglist,teri);
			bool me2 = fill_in_gate_stack(miglist,coglist,teri);
			if(me1 && me2)
				stop = 1;
			else
			{
				bool me3 = fill_in_gate_stack(moglist,ciglist,tero);
				bool me4 = fill_in_gate_stack(moglist,coglist,tero);
				if(me3 && me4)
					stop = 1;
				else
					stop = 0;
			}
*/
		}
	    }
	    curr_gate=curr_gate->next;
	}
  return stop;
}



////////////////////// deal_with_dead_loop ////////////////////////
// Sometimes, search from the output node of every gate is not successful. For this Case,
// 1) We will go to see every unfound gate
// 2) Case 1:
//    If output of the unfound gate connected to more than 1 found gate, great, this 
//    unfound gate can be located.
// 3) Case 2:
//    If output of the unfound gate connected to only 1 found gate, but this gate only
//    has one input(Either a INV, or a tranmission gate), great, this unfound gate can be located.
// 4) Case 3:
//    If output of the unfound gate connected to only 1 found gate, but this gate has more
//    than 1 input. Check if some inputs has been found in Nodestack, track the unfound input 
//    to see its input gate, compare them.


bool 
Patternmatch::query_in_pinlist(char *name, PinList *plist)
{
	for(PinList *curr=plist; curr!=NULL; curr=curr->next)
	{
		if(!strcmp(name, curr->NodeName))
			return true;
	}
	return false;
}

/*** This function is added by Lei Yang, Mar 30,2003 **********************/
/*** This is an extrem case **********************************************/
bool
Patternmatch::one_inlist_included_by_another_inlist(char *gate1, char *gate2)
{
	CGate *small_gate = hash_gate_list[return_number(gate1)];
	CGate *big_gate = hash_gate_list[return_number(gate2)];
	for(PinList *curr=small_gate->inlist; curr!=NULL; curr=curr->next)
	{
		for(PinList *curr2=big_gate->inlist; curr2!=NULL; curr2=curr2->next)
		{
			if(!strcmp(curr->NodeName, curr2->NodeName))
				return true;
		}
	}
	return false;
}
/********************* End add, Mar 30,2003*******************************/


char *
Patternmatch::case_one_gate(char *mgate_name,PinList *plist)
{
/***************** Added by Lei Yang, Mar 30, 2003, for extrem case ********************/
	bool extrem_case = false;
	char *mname = macromodel->query_macromodel_gate_name(plist->NodeName);
	extrem_case = macromodel->one_inlist_included_by_another_inlist(mgate_name, mname);
//	cout<<" gate1: "<<mgate_name<<" gate2: "<<mname<<" extrem_case: "<<extrem_case<<endl;
/********************* End of extrem case **********************************************/
	char *rname = NULL;
	char *rname2 = NULL;
	char *rname3 = NULL;
	char *rname4 = NULL;
	int counter = 0;
// Find the gate in macromodel which we will go to find its matching.
	MCGate *mmg = macromodel->hash_gate_list[return_number(mgate_name)]; 
	
// Find the gate in circuit which has been found.
	CGate *hgate = hash_gate_list[return_number(plist->NodeName)];

// Find all the gates connected with this gate.
  if(hgate != NULL)
  {
	for(PinList *curr=hgate->inlist; curr!=NULL; curr=curr->next)
	{
		NodeList *nnode = get_node(curr->NodeName);
		if(!macromodel->query_node_in_nodestack(curr->NodeName))
		{
			CGate *igate = hash_gate_list[return_number(nnode->output_of_gate_name)];
			{
			  if(mmg!=NULL && igate!=NULL)
			  {
				if(mmg->n_branch!=NULL && mmg->p_branch!=NULL && igate->n_branch!=NULL && igate->p_branch!=NULL)
				{
				if((mmg->n_branch)->headername!=NULL && (mmg->p_branch)->headername!=NULL && (igate->n_branch)->headername!=NULL && (igate->p_branch)->headername!=NULL)
				{
				if((!strcmp((mmg->n_branch)->headername, (igate->n_branch)->headername) && !strcmp((mmg->p_branch)->headername, (igate->p_branch)->headername))) 
				{
/* Original one
					rname = NLAlloc(char, strlen(nnode->output_of_gate_name));
					strcpy(rname, nnode->output_of_gate_name);
					counter += 1;
*/
/******************************* Added by Lei Yang, Mar 30, 2003 *****************************/
				  if(counter==0)
				  {
					rname = NLAlloc(char, strlen(nnode->output_of_gate_name));
					strcpy(rname, nnode->output_of_gate_name);
					counter += 1;
				  }
				  else if(counter==1)
				  {
					rname2 = NLAlloc(char, strlen(nnode->output_of_gate_name));
					strcpy(rname2, nnode->output_of_gate_name);
					counter += 1;
				  }
				  else if(counter==2)
				  {
					rname3 = NLAlloc(char, strlen(nnode->output_of_gate_name));
					strcpy(rname3, nnode->output_of_gate_name);
					counter += 1;
				  }
				  else if(counter==3)
				  {
					rname4 = NLAlloc(char, strlen(nnode->output_of_gate_name));
					strcpy(rname4, nnode->output_of_gate_name);
					counter += 1;
				  }
				  else
					cout<<"Enter this extrem case, Mar 30, 2003!"<<endl;
/************************End *****************************************************************/
				}
				}
				}
			  }
			}
		}
	}
  }
	if(counter == 1)
		return rname;
/********************* Original One *******************
	else
		return NULL;
*******************************************************/
// If counter is more than 1, that means that there is more than one gate matches, that is more than
// what we want, discard it.
/*************************************** New Added One *********************/
	else if(counter == 2)
	{
	  if(extrem_case)
	  {
		if(one_inlist_included_by_another_inlist(rname, plist->NodeName) && !one_inlist_included_by_another_inlist(rname2, plist->NodeName))
			return rname;
		else if(!one_inlist_included_by_another_inlist(rname, plist->NodeName) && one_inlist_included_by_another_inlist(rname2, plist->NodeName))
			return rname2;
		else
			return NULL;
	  }
	  else
		return NULL;
	}
	else if(counter == 3)
	{
	  if(extrem_case)
	  {
		if(one_inlist_included_by_another_inlist(rname, plist->NodeName) && !one_inlist_included_by_another_inlist(rname2, plist->NodeName) && !one_inlist_included_by_another_inlist(rname3, plist->NodeName))
			return rname;
		else if(!one_inlist_included_by_another_inlist(rname, plist->NodeName) && one_inlist_included_by_another_inlist(rname2, plist->NodeName) && !one_inlist_included_by_another_inlist(rname3, plist->NodeName))
			return rname2;
		else if(!one_inlist_included_by_another_inlist(rname, plist->NodeName) && !one_inlist_included_by_another_inlist(rname2, plist->NodeName) && one_inlist_included_by_another_inlist(rname3, plist->NodeName))
			return rname3;
		else
			return NULL;
	  }
	  else
		return NULL;
	}
	else if(counter == 4)
	{
	  if(extrem_case)
	  {
		if(one_inlist_included_by_another_inlist(rname, plist->NodeName) && !one_inlist_included_by_another_inlist(rname2, plist->NodeName) && !one_inlist_included_by_another_inlist(rname3, plist->NodeName) && !one_inlist_included_by_another_inlist(rname4, plist->NodeName))
			return rname;
		else if(!one_inlist_included_by_another_inlist(rname, plist->NodeName) && one_inlist_included_by_another_inlist(rname2, plist->NodeName) && !one_inlist_included_by_another_inlist(rname3, plist->NodeName) && !one_inlist_included_by_another_inlist(rname4, plist->NodeName))
			return rname2;
		else if(!one_inlist_included_by_another_inlist(rname, plist->NodeName) && !one_inlist_included_by_another_inlist(rname2, plist->NodeName) && one_inlist_included_by_another_inlist(rname3, plist->NodeName) && !one_inlist_included_by_another_inlist(rname4, plist->NodeName))
			return rname3;
		else if(!one_inlist_included_by_another_inlist(rname, plist->NodeName) && !one_inlist_included_by_another_inlist(rname2, plist->NodeName) && !one_inlist_included_by_another_inlist(rname3, plist->NodeName) && one_inlist_included_by_another_inlist(rname4, plist->NodeName))
			return rname4;
		else
			return NULL;
	  }
	  else
	      return NULL;
	}
	else
		return NULL;
/********************************** End of new *********************************/
}

char *
Patternmatch::case_more_than_one_gate(char *mgate_name, PinList *plist)
{
	MCGate *mmg = macromodel->hash_gate_list[return_number(mgate_name)];
// Get all the input_list of gates.
	PinList *nlist1=NULL;
	PinList *nlist2=NULL;
	PinList *nlist3=NULL;
	PinList *nlist4=NULL;

	int count=0;
	for(PinList *curr=plist; curr!=NULL; curr=curr->next)
	{
		CGate *hgate = hash_gate_list[return_number(curr->NodeName)];
		if(count==0)
		{
			nlist1 = hgate->inlist;
		}
		else if(count==1)
		{
			nlist2 = hgate->inlist;
		}
		else if(count==2)
		{
			nlist3 = hgate->inlist;
		}
		else if(count==3)
		{
			nlist4 = hgate->inlist;
		}
		else
			;
		count += 1;
	}


// Find the common node in nlist1, nlist2, nlist3, nlist4.
	char *common_node1 = NULL;
	char *common_node2 = NULL;
	int con=0;
	for(PinList *ver=nlist1; ver!=NULL; ver=ver->next)
	{
		if(query_in_pinlist(ver->NodeName, nlist2))
		{
			if(nlist3!=NULL)
			{
				if(query_in_pinlist(ver->NodeName, nlist3))
				{
					if(nlist4!=NULL)
					{
						if(query_in_pinlist(ver->NodeName, nlist4))
						{
// Just do the operation to fill the common_node in 1 or 2;
							if(con==0)
							{
								common_node1 = NLAlloc(char, strlen(ver->NodeName));
								strcpy(common_node1, ver->NodeName);
							}
							else if(con==1)
							{
								common_node2 = NLAlloc(char, strlen(ver->NodeName));
								strcpy(common_node2, ver->NodeName);
							}
							else
							{
								cout<<"The gates has more than 2 common node, exit! \n";
							}
							con += 1;
// Finish the operation.
						}
					}
					else
					{
// Just do the operation to fill the common_node in 1 or 2;
						if(con==0)
						{
							common_node1 = NLAlloc(char, strlen(ver->NodeName));
							strcpy(common_node1, ver->NodeName);
						}
						else if(con==1)
						{
							common_node2 = NLAlloc(char, strlen(ver->NodeName));
							strcpy(common_node2, ver->NodeName);
						}
						else
						{
							cout<<"The gates has more than 2 common node, exit! \n";
						}
						con += 1;
// Finish the operation.
					}
				}
			}
			else
			{
// Just do the operation to fill the common_node in 1 or 2;
				if(con==0)
				{
					common_node1 = NLAlloc(char, strlen(ver->NodeName));
					strcpy(common_node1, ver->NodeName);
				}
				else if(con==1)
				{
					common_node2 = NLAlloc(char, strlen(ver->NodeName));
					strcpy(common_node2, ver->NodeName);
				}
				else
				{
					cout<<"The gates has more than 2 common node, exit! \n";
				}
				con += 1;
// Finish the operation.
			}
		}
	}

	if(common_node1!=NULL && common_node2!=NULL)
	{
		NodeList *com1 = get_node(common_node1);
		CGate *comgate1 = hash_gate_list[return_number(com1->output_of_gate_name)];
		NodeList *com2 = get_node(common_node2);
		CGate *comgate2 = hash_gate_list[return_number(com2->output_of_gate_name)];

		if((!strcmp((mmg->n_branch)->headername, (comgate1->n_branch)->headername) && !strcmp((mmg->p_branch)->headername, (comgate1->p_branch)->headername))) 
		{
			if(!strcmp((mmg->n_branch)->headername, (comgate2->n_branch)->headername) && !strcmp((mmg->p_branch)->headername, (comgate2->p_branch)->headername))
				return NULL;
			else
				return com1->output_of_gate_name;
		}

		if((!strcmp((mmg->n_branch)->headername, (comgate2->n_branch)->headername) && !strcmp((mmg->p_branch)->headername, (comgate2->p_branch)->headername))) 
		{
			if(!strcmp((mmg->n_branch)->headername, (comgate1->n_branch)->headername) && !strcmp((mmg->p_branch)->headername, (comgate1->p_branch)->headername))
				return NULL;
			else
				return com2->output_of_gate_name;
		}
	}
	else if(common_node1!=NULL && common_node2==NULL)
	{
		NodeList *com1 = get_node(common_node1);
		CGate *comgate1 = hash_gate_list[return_number(com1->output_of_gate_name)];
		if((!strcmp((mmg->n_branch)->headername, (comgate1->n_branch)->headername) && !strcmp((mmg->p_branch)->headername, (comgate1->p_branch)->headername))) 
			return com1->output_of_gate_name;
		else
			return NULL;
	}
	else
		return NULL;
}

void
Patternmatch::deal_with_dead_loop()
{
	for(GateStack *curr=macromodel->gate_stack; curr!=NULL; curr=curr->next)
	{
		if(!curr->recognized_flag && decide_a_gate(curr->macromodel_name))
		{
//			cout<<" find gate: "<<curr->macromodel_name<<endl;
			PinList *plist = macromodel->return_found_gate_list(curr->macromodel_name);
/*
				for(PinList *cg=plist; cg!=NULL; cg=cg->next)
					printf("con: %s ", cg->NodeName);
				printf("\n");
*/
			int cnum = count_num_of_a_list(plist);
			if(cnum==0)
			{
				;
			}
			else if(cnum==1)
			{
				char *cname = case_one_gate(curr->macromodel_name, plist);
				if(cname!=NULL)
				{
					macromodel->insert_gates_in_gatestack(curr->macromodel_name, cname);
					fill_gate_node_in_stack(curr->macromodel_name, cname);
				}
			}
			else if(cnum>1 && cnum<5)
			{
				char *cname = case_more_than_one_gate(curr->macromodel_name, plist);
				if(cname!=NULL)
				{
					macromodel->insert_gates_in_gatestack(curr->macromodel_name, cname);
					fill_gate_node_in_stack(curr->macromodel_name, cname);
				}
			}
			else if(cnum>4)
			{
				cout<<"Unknown connected to 3 and more known gate, wired\n";
			}
		}
	}
}

///////////////////// From here, we should do the input and output matching.
////////////// Input_output_matching
// 1) Output matching
//    Because every output should be the output of one gate, it should have
//    matched in the block matching process.
// 2) Input matching
//	  i)   If the input is the input of a INV or TRANS, this is easy, and it
//		   should have been matched in the block matching.
//	  ii)  If the input is the input of several gates, this is also easy, 
//		   go to find the common input node of the several gates.
//	  iii) If the input only connect to one gate, and this gate have several
//		   inputs, which is hard to tell.


// This function, if a valid_gatename_list is given, we should return a pinname.
bool
Patternmatch::query_in_gatename_list(char *name, GateNameList *gnlist)
{
	for(GateNameList *curr=gnlist; curr!=NULL; curr=curr->next)
	{
		if(!strcmp(name, curr->name))
			return true;
	}
	return false;
}

char *
Patternmatch::compare_two_gatename_list(GateNameList *return_list, GateNameList *real_list)
{
	char *retname = NULL;
	if(query_in_gatename_list(return_list->name, real_list))
	{
		int counter = 0;
		for(GateNameList *curr=real_list; curr!=NULL; curr=curr->next)
		{
			if(!query_in_gatename_list(curr->name, return_list))
			{
				counter += 1;
				retname = NLAlloc(char, strlen(curr->name));
				strcpy(retname, curr->name);
			}
		}

		if(counter==1)	
			return retname;
		else if(counter==0)
			cout<<"Compare result: in the circuit gatename_list, no gate pin name have been found, exit!\n";
		else if(counter>1)
			cout<<"Compare result: in the circuit gatename_list, more than 1 pin name have been found, exit!\n";
		else
			;

	}
	else
		return NULL;
}

char *
Patternmatch::return_nbranch_matched_pinname(GateNameList *valid_list, CGate *hgate)
{
	for(RelationList *curr=(hgate->n_branch)->relation_list; curr!=NULL; curr=curr->next)
	{
		if(compare_two_gatename_list(valid_list, curr->gatename_list)!=NULL)
			return compare_two_gatename_list(valid_list, curr->gatename_list);
	}
	return NULL;
}

char *
Patternmatch::return_pbranch_matched_pinname(GateNameList *valid_list, CGate *hgate)
{
	for(RelationList *curr=(hgate->p_branch)->relation_list; curr!=NULL; curr=curr->next)
	{
		if(compare_two_gatename_list(valid_list, curr->gatename_list)!=NULL)
			return compare_two_gatename_list(valid_list, curr->gatename_list);
	}
	return NULL;
}

char *
Patternmatch::case_one_gate_for_a_pin(char *pinname,CGate *hgate)
{
	char *rname = NULL;
//	cout<<"pinname: "<<pinname<<endl;
//	cout<<" gatename: "<<hgate->gatename<<endl;
	
// Find the gate in circuit which has been found.
//	CGate *hgate = hash_gate_list[return_number(plist->NodeName)];

// Find the relation_list pinname belong to
	RelationList *nrlist = macromodel->return_n_relation_list(pinname);

	/*
		for(RelationList *cg=nrlist; cg!=NULL; cg=cg->next)
		{
			cout<<"type: "<<cg->type;
                        for(GateNameList *cn=cg->gatename_list; cn!=NULL; cn=cn->next)
				cout<<" name:  "<<cn->name;
			cout<<endl;
		}
	*/


	GateNameList *valid_gate_list = macromodel->relation_list_is_valid(pinname, nrlist);
//	cout<<" valid gate list\n";
/*
                        for(GateNameList *cn=valid_gate_list; cn!=NULL; cn=cn->next)
				cout<<" valid name:  "<<cn->name;
			cout<<endl;
*/
	if(valid_gate_list!=NULL)
	{
		if(return_nbranch_matched_pinname(valid_gate_list, hgate)!=NULL)
		{
//			cout<<" last: "<<return_nbranch_matched_pinname(valid_gate_list, hgate)<<endl;
			return return_nbranch_matched_pinname(valid_gate_list, hgate);
		}
		else
			cout<<"Return a NULL, unbelieveable\n";
	}
	else
	{
		RelationList *prlist = macromodel->return_p_relation_list(pinname);
		GateNameList *p_valid_gate_list = macromodel->relation_list_is_valid(pinname, prlist);
		if(p_valid_gate_list!=NULL)
		{
			if(return_pbranch_matched_pinname(p_valid_gate_list, hgate)!=NULL)
				return return_pbranch_matched_pinname(p_valid_gate_list, hgate);
			else
				cout<<"Return a NULL for P also, relation_pin method failed\n";
		}
		else
			cout<<"Can find this pinname use the relation_pin method!\n";
	}
	return NULL;
}


char *
Patternmatch::case_more_than_one_gate_for_a_pin(PinList *plist)
{
// Get all the input_list of gates.
	PinList *nlist1=NULL;
	PinList *nlist2=NULL;
	PinList *nlist3=NULL;
	PinList *nlist4=NULL;

	int count=0;
	for(PinList *curr=plist; curr!=NULL; curr=curr->next)
	{
		CGate *hgate = hash_gate_list[return_number(curr->NodeName)];
		if(count==0)
		{
			nlist1 = hgate->inlist;
		}
		else if(count==1)
		{
			nlist2 = hgate->inlist;
		}
		else if(count==2)
		{
			nlist3 = hgate->inlist;
		}
		else if(count==3)
		{
			nlist4 = hgate->inlist;
		}
		else
			cout<<"The pin connect to more than 4 gates, not acceptable\n";
		count += 1;
	}


// Find the common node in nlist1, nlist2, nlist3, nlist4.
	char *common_node1 = NULL;
	char *common_node2 = NULL;
	int con=0;
	for(PinList *ver=nlist1; ver!=NULL; ver=ver->next)
	{
		if(query_in_pinlist(ver->NodeName, nlist2))
		{
			if(nlist3!=NULL)
			{
				if(query_in_pinlist(ver->NodeName, nlist3))
				{
					if(nlist4!=NULL)
					{
						if(query_in_pinlist(ver->NodeName, nlist4))
						{
// Just do the operation to fill the common_node in 1 or 2;
							if(con==0)
							{
								common_node1 = NLAlloc(char, strlen(ver->NodeName));
								strcpy(common_node1, ver->NodeName);
							}
							else if(con==1)
							{
								common_node2 = NLAlloc(char, strlen(ver->NodeName));
								strcpy(common_node2, ver->NodeName);
							}
							else
							{
								cout<<"The gates has more than 2 common node, exit! \n";
							}
							con += 1;
// Finish the operation.
						}
					}
					else
					{
// Just do the operation to fill the common_node in 1 or 2;
						if(con==0)
						{
							common_node1 = NLAlloc(char, strlen(ver->NodeName));
							strcpy(common_node1, ver->NodeName);
						}
						else if(con==1)
						{
							common_node2 = NLAlloc(char, strlen(ver->NodeName));
							strcpy(common_node2, ver->NodeName);
						}
						else
						{
							cout<<"The gates has more than 2 common node, exit! \n";
						}
						con += 1;
// Finish the operation.
					}
				}
			}
			else
			{
// Just do the operation to fill the common_node in 1 or 2;
				if(con==0)
				{
					common_node1 = NLAlloc(char, strlen(ver->NodeName));
					strcpy(common_node1, ver->NodeName);
				}
				else if(con==1)
				{
					common_node2 = NLAlloc(char, strlen(ver->NodeName));
					strcpy(common_node2, ver->NodeName);
				}
				else
				{
					cout<<"The gates has more than 2 common node, exit! \n";
				}
				con += 1;
// Finish the operation.
			}
		}
	}

	if(common_node1!=NULL && common_node2!=NULL)
	{
				return NULL;
	}
	else if(common_node1!=NULL && common_node2==NULL)
	{
			return common_node1;
	}
	else
		return NULL;
}

void
Patternmatch::input_output_matching()
{
	for(NodeStack *curr=macromodel->node_stack; curr!=NULL; curr=curr->next)
	{
		if(curr->terminal_flag && !curr->recognized_flag)
		{
			cout<<" Unfound pin: "<<curr->macromodel_name<<endl;
			if(macromodel->query_in_outputlist(curr->macromodel_name))
			{
				cout<<"It's wired, because the output has not been found yet after the gate matching\n";
			}
			else
			{
				PinList *plist = macromodel->return_found_gate_list_for_a_pin(curr->macromodel_name);

				/*
				for(PinList *cp=plist; cp!=NULL; cp=cp->next)
					cout<<" gate: "<<cp->NodeName;
				cout<<endl;
				*/

				int cnum = count_num_of_a_list(plist);
//				cout<<" num of gates: "<<cnum<<endl;
				if(cnum==0)
				{
					cout<<"problem pin: "<<curr->macromodel_name<<endl;
					cout<<"This is impossible, because the pin did not connected to any gate!\n";
				}
				else if(cnum==1)
				{
					CGate *hgate = hash_gate_list[return_number(plist->NodeName)];
					char *cname = case_one_gate_for_a_pin(curr->macromodel_name, hgate);
					if(cname!=NULL)
					{
						macromodel->insert_node_in_nodestack(curr->macromodel_name, cname);
//						cout<<"m: "<<curr->macromodel_name<<" c: "<<cname<<endl;
					}
					else
						cout<<"Relation_pin method failed! \n";
				}
				else if(cnum>1 && cnum<5)
				{
					char *cname = case_more_than_one_gate_for_a_pin(plist);
					if(cname!=NULL)
					{
						macromodel->insert_node_in_nodestack(curr->macromodel_name, cname);
					}
// try the only one gate method.
					else
					{
						for(PinList *yp=plist; yp!=NULL; yp=yp->next)
						{
							CGate *ggate = hash_gate_list[return_number(plist->NodeName)];
							char *nname = case_one_gate_for_a_pin(curr->macromodel_name,ggate);
							if(nname!=NULL)
							{
								macromodel->insert_node_in_nodestack(curr->macromodel_name, nname);
								break;
							}
						}
					}
				}
				else if(cnum>4)
				{
					cout<<"Unknown connected to 3 and more known gate, wired! \n";
				}
				
			}
		}
	}
}

// If after the input_output_matching, there are still one pair of input pins can not be
// recognized, that may because that the two pins are symmetric, so match them randomly.
int
Patternmatch::count_num_of_gatename_list(GateNameList *gnlist)
{
	int num=0;
	for(GateNameList *curr=gnlist; curr!=NULL; curr=curr->next)
	{
		num += 1;
	}
	return num;
}

bool
Patternmatch::node_not_found_yet(char *name)
{
	for(NodeStack *curr=macromodel->node_stack; curr!=NULL; curr=curr->next)
	{
		if(curr->recognized_flag)
		{
			if(!strcmp(curr->circuit_name, name))
				return false;
		}
	}
	return true;
}

bool
Patternmatch::gatename_list_not_found_yet(GateNameList *gnlist)
{
	for(GateNameList *curr=gnlist; curr!=NULL; curr=curr->next)
	{
		if(!node_not_found_yet(curr->name))
			return false;
	}
	return true;
}


void
Patternmatch::related_pin_recognition()
{
	for(NodeStack *curr=macromodel->node_stack; curr!=NULL; curr=curr->next)
	{
// If one node is pin, and did not found yet, begin this new search algorithm.
		if(curr->terminal_flag && !curr->recognized_flag)
		{
			for(NodeList *cn=macromodel->node_list; cn!=NULL; cn=cn->next)
			{
				if(!strcmp(cn->nodename, curr->macromodel_name))
				{
					// In macromodel's node_list, find this node, this node has input_of_gate_list, get the first one as the target gate(hgate);
					MCGate *hgate = macromodel->hash_gate_list[return_number((cn->input_of_gate_list)->NodeName)];
					RelationList *rval = macromodel->locate_relation_list(curr->macromodel_name, (hgate->n_branch)->relation_list);
					bool mflag = macromodel->gatename_list_not_found_yet(rval->gatename_list);
					if(mflag)
					{
						// Find the corresponding cgate of target gate(hgate).
						char *cgatename = macromodel->find_circuit_gate_name(hgate->gatename);
						CGate *cgate = hash_gate_list[return_number(cgatename)];
						for(RelationList *cr=(cgate->n_branch)->relation_list; cr!=NULL; cr=cr->next)
						{
							if(count_num_of_gatename_list(cr->gatename_list) == count_num_of_gatename_list(rval->gatename_list))
							{
								if(gatename_list_not_found_yet(cr->gatename_list))
								{
									macromodel->insert_two_gatename_list(rval->gatename_list, cr->gatename_list);
								}
							}
						}
					}

				}
			}
		}
	}
}

bool
Patternmatch::query_in_inputlist(char *name)
{
	for(InoutList *curr=macromodel->inputhead; curr!=NULL; curr=curr->next)
	{
		if(!strcmp(name, curr->pinname))
		{
			return true;
		}
	}
	return false;
}

EqualNameList *
Patternmatch::insert_equal_name(EqualNameList *head, char *mname, char *cname)
{
  EqualNameList *current;
  if(head==NULL)
  {
        head=NLAlloc(EqualNameList,1);
        head->macromodel_name = NLAlloc(char,strlen(mname));
        strcpy(head->macromodel_name,mname);
        head->circuit_name = NLAlloc(char,strlen(cname));
        strcpy(head->circuit_name,cname);
        head->next = NULL;
  }
  else {
        current=NLAlloc(EqualNameList,1);
        current->macromodel_name = NLAlloc(char,strlen(mname));
        strcpy(current->macromodel_name,mname);
        current->circuit_name = NLAlloc(char,strlen(cname));
        strcpy(current->circuit_name,cname);
        current->next = head->next;
        head->next = current;
  }
  return head;
}

ResultList *
Patternmatch::generate_result_list(ResultList *head, char *name, EqualNameList *inputlist, EqualNameList *outputlist)
{
    ResultList *current;
    if(head==NULL)
    {
	head = NLAlloc(ResultList,1);
	head->libname = NLAlloc(char, strlen(name));
	strcpy(head->libname, name);
	head->input_list = inputlist;
	head->output_list = outputlist;
	head->next = NULL;
    }
    else
    {
	current = NLAlloc(ResultList,1);
	current->libname = NLAlloc(char, strlen(name));
	strcpy(current->libname, name);
	current->input_list = inputlist;
	current->output_list = outputlist;
	current->next = head->next;
	head->next = current;
    }
    return head;
}

PinList *
Patternmatch::insert_pin(PinList *head, char *name){
  PinList *current;
  if(head==NULL)
  {
        head=NLAlloc(PinList,1);
        head->NodeName = NLAlloc(char,strlen(name));
        strcpy(head->NodeName,name);
        head->next = NULL;
  }
  else {
        current=NLAlloc(PinList,1);
        current->NodeName = NLAlloc(char,strlen(name));
        strcpy(current->NodeName,name);
        current->next = head->next;
        head->next = current;
  }
  return head;
}


PinList *
Patternmatch::delete_one_in_pinlist(char *name, PinList *plist)
{
	PinList *rlist=NULL;
	for(PinList *curr=plist; curr!=NULL; curr=curr->next)
	{
		if(strcmp(name, curr->NodeName))
		{
			rlist = insert_pin(rlist, curr->NodeName);
		}
	}
	return rlist;
}

/*
void
Patternmatch::delete_one_in_pinlist(char *name, PinList *plist)
{
	if(plist->next == NULL)
	{
		if(!strcmp(name, plist->NodeName))
			plist = NULL;
	}
	else
	{
        PinList *curr, *prev;
        curr=prev=plist;
        while(curr!=NULL)
        {
          if(!strcmp(name, curr->NodeName))
          {
            if(curr==plist)
            {
                prev=curr->next;
                curr=curr->next;
                plist = curr;
            }
            else
            {
                prev->next = curr->next;
                curr = prev->next;
            }
            break;
          }
          else
          {
                prev=curr;
                curr=curr->next;
          }
        }
	}
}
*/

// If the pattern matching success, delete all the found gates
void
Patternmatch::delete_found_gates_trans()
{
// Delete all the gates and trans found
	for(GateStack *curr=macromodel->gate_stack; curr!=NULL; curr=curr->next)
	{
		bool gate_or_trans = decide_a_gate(curr->circuit_name);
		if(gate_or_trans)
		{
			CGate *hgate = hash_gate_list[return_number(curr->circuit_name)];
// Fist delete the output node and inlist node
/*
			NodeList *onode = get_node((hgate->n_branch)->pn_node);
			onode->output_of_gate_name = NULL;

			for(PinList *cpl=hgate->inlist; cpl!=NULL; cpl=cpl->next)
			{
				NodeList *pnode = get_node(cpl->NodeName);
				delete_one_in_pinlist(hgate->gatename, pnode->input_of_gate_list);
			}
*/
			hash_gate_list[return_number(curr->circuit_name)] = NULL;
		}
		else
		{
			TransList *htrans = hash_trans_list[return_number(curr->circuit_name)];
/*
		  if(htrans != NULL)
		  {
			NodeList *onode = get_node(htrans->input);
			if(onode->input_of_trans_name != NULL)
			{
				if(!strcmp(onode->input_of_trans_name, htrans->transname))
					onode->input_of_trans_name = NULL;
			}
			if(onode->output_of_trans_name != NULL)
			{
				if(!strcmp(onode->output_of_trans_name, htrans->transname))
					onode->output_of_trans_name = NULL;
			}

			NodeList *pnode = get_node(htrans->output);
			if(pnode->input_of_trans_name != NULL)
			{
				if(!strcmp(pnode->input_of_trans_name, htrans->transname))
					pnode->input_of_trans_name = NULL;
			}
			if(pnode->output_of_trans_name != NULL)
			{
				if(!strcmp(pnode->output_of_trans_name, htrans->transname))
					pnode->output_of_trans_name = NULL;
			}
		  }
*/
			hash_trans_list[return_number(curr->circuit_name)] = NULL;
		}
	}
}

void
Patternmatch::delete_the_connection_of_terminal(EqualNameList *inputlist, EqualNameList *outputlist)
{
	for(EqualNameList *curr=inputlist; curr!=NULL; curr=curr->next)
	{
		NodeList *hnode = get_node(curr->circuit_name);
		if(hnode->input_of_gate_list == NULL)
			;
		else
		{
			for(PinList *cp=hnode->input_of_gate_list; cp!=NULL; cp=cp->next)
			{
//					cout<<" delete this gate: "<<cp->NodeName<<endl;
				if(macromodel->query_gate_has_been_found(cp->NodeName))
					hnode->input_of_gate_list = delete_one_in_pinlist(cp->NodeName, hnode->input_of_gate_list);
			}
		}
	}
	for(EqualNameList *curr=outputlist; curr!=NULL; curr=curr->next)
	{
		NodeList *hnode = get_node(curr->circuit_name);
		if(hnode->output_of_gate_name != NULL)
		{
			if(macromodel->query_gate_has_been_found(hnode->output_of_gate_name))
				hnode->output_of_gate_name = NULL;
		}
		if(hnode->output_of_trans_name != NULL)
		{
			if(macromodel->query_gate_has_been_found(hnode->output_of_trans_name))
				hnode->output_of_trans_name = NULL;
		}
		if(hnode->input_of_trans_name != NULL)
		{
			if(macromodel->query_gate_has_been_found(hnode->input_of_trans_name))
				hnode->input_of_trans_name = NULL;
		}
	}
}

///////////////////// Pattern matching loop ////////////////////////
//pattern_match_process
void
Patternmatch::pattern_match_process()
{
	macromodel->initialize_gate_stack();
	macromodel->initialize_node_stack();
		cout<<" initialize \n";

/*
	cout<<" GateStack ..............................."<<endl;
	for(GateStack *curr=macromodel->gate_stack; curr!=NULL; curr=curr->next)
	{
		cout<<" M_name: "<<curr->macromodel_name<<" r_flag: "<<curr->recognized_flag<<" s_flag: "<<curr->search_flag<<endl;
	}
	cout<<" End of GateStack ......................."<<endl;
	cout<<" NodeStack ..............................."<<endl;
	for(NodeStack *curr=macromodel->node_stack; curr!=NULL; curr=curr->next)
	{
		cout<<" M_name: "<<curr->macromodel_name<<" r_flag: "<<curr->recognized_flag<<" s_flag: "<<curr->search_flag<<" t_flag: "<<curr->terminal_flag<<endl;
	}
	cout<<" End of NodeStack ......................."<<endl;
*/
	

	MCGate *m_gate = macromodel->locate_first_gate();
		cout<<" locate first gate \n";

	cout<<"first gate name:"<<m_gate->gatename<<" p logic: "<<(m_gate->p_branch)->headername<<" n logic: "<<(m_gate->n_branch)->headername<<endl;

//	cout<<"locate first gate success\n"<<endl;

  for(int ii=0; ii<num_of_gates; ii++)
  {
   if(hash_gate_list[ii]!=NULL)
   {
//	if(!strcmp(hash_gate_list[ii]->gatename, "gate661"))
//	{
	if((hash_gate_list[ii]->n_branch)->headername !=NULL && (hash_gate_list[ii]->p_branch)->headername !=NULL)
	{
	if(!strcmp((m_gate->n_branch)->headername, (hash_gate_list[ii]->n_branch)->headername) && !strcmp((m_gate->p_branch)->headername, (hash_gate_list[ii]->p_branch)->headername))
	{
//		cout<<" A new start point::::::::::::::::::::::  "<<hash_gate_list[ii]->gatename<<" headername: "<<((hash_gate_list[ii]->n_branch)->headername)<<endl;
/*
				NodeList *pnode=get_node((hash_gate_list[ii]->n_branch)->pn_node);
				for(PinList *curr=pnode->input_of_gate_list; curr!=NULL; curr=curr->next)
				{
					if(decide_a_gate(curr->NodeName))
					{
						CGate *pgate=hash_gate_list[return_number(curr->NodeName)];
						cout<<" After first gate: "<<" P log: "<<(pgate->p_branch)->headername<<" N log: "<<(pgate->n_branch)->headername<<endl;
					}
				}
*/				
		macromodel->insert_gates_in_gatestack(m_gate->gatename,hash_gate_list[ii]->gatename);
		fill_gate_node_in_stack(m_gate->gatename, hash_gate_list[ii]->gatename);

//		cout<<" macro: "<<m_gate->gatename<<"  circuit: "<<hash_gate_list[ii]->gatename<<endl;

// Gate matches
    int stop = 1;
	bool same_stack = false;
	GateStack *gatestack_old = NULL;
	GateStack *gatestack_new = NULL;
    while(macromodel->gatestack_search_to_end() && stop && !same_stack)
    {
		gatestack_old = macromodel->copy_gate_stack();
//			printf("************************************* one step**********************8\n");
		stop = step();

/*
//if(stop)
//{
//	gigi = 1;
cout<<" Good gate: "<<hash_gate_list[ii]->gatename<<endl;
//if(!strcmp(macromodel->name, "DFFP"))
//{
		cout<<" +++++status of Gate_stack "<<" stop: "<<stop<<endl;
			for(GateStack *ct=macromodel->gate_stack; ct!=NULL; ct=ct->next)
			{
				cout<<" Macro_name: "<<ct->macromodel_name;
				if(ct->circuit_name!=NULL)
					cout<<" Cir_name: "<<ct->circuit_name;
				else
					cout<<" Cir_name: (NULL)";
				cout<<" recog_flag: "<<ct->recognized_flag<<" search_flag: "<<ct->search_flag<<endl;
			}
		cout<<"++++ end of gate_stack"<<endl;
		cout<<endl;


		cout<<" +++++ status of Node_stack "<<endl;
			for(NodeStack *ct=macromodel->node_stack; ct!=NULL; ct=ct->next)
			{
				cout<<" Macro_name: "<<ct->macromodel_name;
				if(ct->circuit_name!=NULL)
					cout<<" Cir_name: "<<ct->circuit_name;
				else
					cout<<" Cir_name: (NULL)";
				cout<<" recog_flag: "<<ct->recognized_flag<<" search_flag: "<<ct->search_flag<<" ter_flag: "<<ct->terminal_flag<<endl;
			}
		cout<<"+++++++ end of node_stack"<<endl;
		cout<<endl;
//}
//}
*/


		macromodel->set_trans_search_flag_one();

		if(stop)
		{
//				cout<<" check dead loop: ..............."<<macromodel->check_dead_loop()<<endl;
			if(macromodel->check_dead_loop())
			{
//				cout<<"----------- check dead loop: "<<macromodel->check_dead_loop()<<endl;
				deal_with_dead_loop();
				macromodel->deal_with_dead_loop2();
/************ Added by Lei Yang, Mar 30, 2003 to search from transmission gate ******************/
				macromodel->set_trans_search_flag_zero();
				step();
/************ End of this ***********************************************************************/
			}
		}

	/*
		cout<<" +++++status of Gate_stack "<<" stop: "<<stop<<endl;
			for(GateStack *ct=macromodel->gate_stack; ct!=NULL; ct=ct->next)
				cout<<" Macro_name: "<<ct->macromodel_name<<" Cir_name: "<<ct->circuit_name<<" recog_flag: "<<ct->recognized_flag<<" search_flag: "<<ct->search_flag<<endl;
		cout<<"++++ end of gate_stack"<<endl;
		cout<<endl;


		cout<<" +++++ status of Node_stack "<<endl;
			for(NodeStack *ct=macromodel->node_stack; ct!=NULL; ct=ct->next)
				cout<<" Macro_name: "<<ct->macromodel_name<<" Cir_name: "<<ct->circuit_name<<" recog_flag: "<<ct->recognized_flag<<" search_flag: "<<ct->search_flag<<" ter_flag: "<<ct->terminal_flag<<endl;
		cout<<"+++++++ end of node_stack"<<endl;
		cout<<endl;
	*/
		gatestack_new = macromodel->copy_gate_stack();
		same_stack = macromodel->compare_two_gatestack(gatestack_old, gatestack_new);
    }
//    cout<<"stop: "<<stop<<endl;

// If the block matches, do the input output matches.






	if(!macromodel->gatestack_search_to_end())
	{
		cout<<" gate match done \n";
		for(GateStack *cg=macromodel->gate_stack; cg!=NULL; cg=cg->next)
		{
			if(cg->recognized_flag)
				fill_gate_node_in_stack(cg->macromodel_name, cg->circuit_name);
		}


		input_output_matching();
		if(!macromodel->gatestack_search_to_end())
		{
			cout<<" Begin Related_pin_recognition method! \n";
			cout<<" Begin Related_pin_recognition method! \n";
			related_pin_recognition();
		}

		cout<<" input_output match done \n";
	}


/*
	cout<<" GateStack ..............................."<<endl;
	for(GateStack *curr=macromodel->gate_stack; curr!=NULL; curr=curr->next)
	{
		cout<<" M_name: "<<curr->macromodel_name<<" C_name: "<<curr->circuit_name<<" r_flag: "<<curr->recognized_flag<<" s_flag: "<<curr->search_flag<<endl;
	}
	cout<<" End of GateStack ......................."<<endl;
		cout<<" +++++ status of Node_stack "<<endl;
			for(NodeStack *ct=macromodel->node_stack; ct!=NULL; ct=ct->next)
				cout<<" Macro_name: "<<ct->macromodel_name<<" Cir_name: "<<ct->circuit_name<<" recog_flag: "<<ct->recognized_flag<<" search_flag: "<<ct->search_flag<<" ter_flag: "<<ct->terminal_flag<<endl;
		cout<<"+++++++ end of node_stack"<<endl;
		cout<<endl;
*/

		int return_flag=0;
		if(!macromodel->nodestack_search_to_end() && !macromodel->gatestack_search_to_end())
		{
			cout<<" success..................................... \n"<<endl;
			// Generate the input and outputlist of the macromodelcell
		    EqualNameList *inputlist=NULL;
		    EqualNameList *outputlist=NULL;
			for(NodeStack *curr=macromodel->node_stack; curr!=NULL; curr=curr->next)
			{
				if(curr->terminal_flag && curr->recognized_flag)
				{
					if(query_in_inputlist(curr->macromodel_name))
					{
						inputlist = insert_equal_name(inputlist, curr->macromodel_name, curr->circuit_name);
//						cout<<" M Pin: "<<curr->macromodel_name<<"  C Pin: "<<curr->circuit_name<<endl;
					}
					else
					{
						outputlist = insert_equal_name(outputlist, curr->macromodel_name, curr->circuit_name);
//						cout<<" M Pin: "<<curr->macromodel_name<<"  C Pin: "<<curr->circuit_name<<endl;
					}
				}
			}
//			cout<<" input output done \n";
			// End of generating inputlist and outputlist
		    ResultHead = generate_result_list(ResultHead, macromodel->name, inputlist, outputlist);
//		    cout<<" reach here\n";

		    delete_found_gates_trans();
			delete_the_connection_of_terminal(inputlist, outputlist);
//		    cout<<" reach here 2\n";
		    return_flag=1;
		}
		else
		{
			// Do nothing, because the recognition failed.
//			cout<<"recognition failed\n";
			;
		}

		macromodel->reinitialize_macromodel();
	}
//				}
	}
   }
  }
}

