#ifndef _PRE
#define _PRE

typedef enum {RES, CAP, IND, VS,PULSE_VS, IS, E, H, G, F, D, M} ElementType;
typedef enum {TRAN,PRINT_TYPE,OPTION_TYPE,END_TYPE,DC_TYPE} ActionType;
typedef enum {DIOD_TYPE,MOSFET_TYPE} ModelType;
typedef enum {DP,GP,SP} PinType;
//typedef enum {SINGLE, DOUBLE} TransType;

typedef struct BRANCHNODELIST {
        char *nodename;
        struct BRANCHNODELIST *next;
        } BranchNodeList;

typedef struct TERMINALLIST{
        char *name;
        struct TERMINALLIST *next;
        } TerminalList;

typedef struct NAMELIST {
        char *gatename;
        char virtual_name;
        struct NAMELIST *next;
        } NameList;

typedef struct WIRELIST{
	char *wirename;
	struct WIRELIST *next;
	} WireList;

typedef struct TRANSLIST{
//	TransType type;
	char *transname;
	char *input;
	char *output;
	char *p_g;
	char *n_g;
	int ready_to_delete;
	struct TRANSLIST *next;
	} TransList;
	
typedef struct DFFLIST{
	char *dffname;
	char *d;
	char *q;
	char *clk;
	char *rst;
	struct DFFLIST *next;
	} DffList;

typedef struct PINLIST{
        char *NodeName;
        struct PINLIST *next;
        } PinList;

typedef struct INOUTLIST{
        char *pinname;
        struct INOUTLIST *next;
        } InoutList;


typedef struct ELEMENT{
	char *name;
	ElementType type;
	PinList *nodelist;
	float value;
        char *Model;
	float l_value;
	float w_value;
	struct ELEMENT *next;
	} Element;

typedef struct ACTION{
	ActionType type;
	float value1;
	float value2;
	int v;
	float start;
	float end;
	float step;
	char * name;
	struct ACTION *next;
	} Action;
	
typedef struct MODELLIST{
	ModelType type;
	char *name;
	char *npmos;
	float is;
	float vt;
	float k;
	float vth;
	float lambda;
	struct MODELLIST *next;
	} ModelList;	

typedef struct BIPOLARLIST {
        char *name;
        char *c;
        char *b;
        char *e;
        int nptype;
        char *modelname;
        struct BIPOLARLIST *next;
        } BipolarList;


typedef struct MOSLIST{
	int takeup; // This is to see if the mos has shown up in the list.
	char *mosname;
	char *d;
	char *g;
	char *s;
	int  nptype;
	int is_instead;
	char *b;
	char *modelname;
	float lvalue;
	float wvalue; 
	bool in_branch;
	struct MOSLIST *next;
	} MosList, *PMosList;

typedef struct DEVICELIST {
	char *devicename;
	int devicetype;
	char *d;
	char *g;
	char *s;
	PinType pintype;
	struct DEVICELIST *next;
	} DeviceList;

typedef struct NODELIST {
	char *nodename;
	bool terminal;
	bool wire;
	DeviceList *devicelist;
	int pn_node;
	PinList *p_list; // Add this I want to reduce the group time.
	PinList *n_list; // Add this I want to reduce the group time.
	char *connected_p_branch_name;
	char *connected_n_branch_name;
	char *output_of_gate_name;
	PinList *input_of_gate_list;
	char *output_of_trans_name;
//	char *input_name_of_trans;
//	PinList *input_of_trans_list;
	char *input_of_trans_name;
	int num_of_inputs;
	struct NODELIST *next;
	} NodeList, *PNodeList ;


typedef struct GATESTACK {
	int type; // gate=0, trans=1
        char *macromodel_name;
        char *circuit_name;
        bool recognized_flag;
        bool search_flag;
        struct GATESTACK *next;
        } GateStack;

typedef struct NODESTACK {
        char *macromodel_name;
        char *circuit_name;
        bool terminal_flag;
        bool recognized_flag;
        bool search_flag;
        struct NODESTACK *next;
        } NodeStack;

typedef struct RGLIST {
        char *gatename;
        int type; // gate=0, trans=1
        bool unique; // because of have found.
	bool unique2; // because of the same structure.
        struct RGLIST *next;
        } RGList;

typedef struct GATENAMELIST {
	char *name;
	struct GATENAMELIST *next;
	} GateNameList;

typedef struct RELATIONLIST {
	int type; //alonelist 0, parallel 1, series 2;
	GateNameList *gatename_list;
	struct RELATIONLIST *next;
	} RelationList;

typedef struct EQUALNAMELIST {
	char *macromodel_name;
	char *circuit_name;
	struct EQUALNAMELIST *next;
	} EqualNameList;
	
typedef struct RESULTLIST {
        char *libname;
        EqualNameList *input_list;
        EqualNameList *output_list;
        struct RESULTLIST *next;
        } ResultList;

typedef struct MACROMODELLIST{
        char *macromodel_name;
        TerminalList *terminal_list;
        Element *element_list;
        MosList *mos_list;
        BipolarList *bipolar_list;
        ModelList *model_list;
	int num_of_mos;
	bool done;
        int flag;

        struct MACROMODELLIST *next;
        } MacroModelList;


#endif
