#include "pre.h"
#include "branch.h"
#include "group.h"
#include <malloc.h>
#include <stdio.h>
#include <string.h>

#define NLAlloc(type,num) (type *)calloc(num,sizeof(type))

extern MosList *MosHead;
extern NodeList *NodeHead;
extern TransList *TransHead;
extern DffList *DffHead;
extern WireList *WireHead;
extern InoutList *InputHead;
extern InoutList *OutputHead;
int counter_flag = 0;
char *reset_signal;

Group::Group()
{
	p_branch_list = NULL;
	n_branch_list = NULL;
	pass_transistor_list = NULL;
	left_transistor_list = NULL;
}

Group::~Group()
{
	if(p_branch_list) delete p_branch_list;
	if(n_branch_list) delete n_branch_list;
	if(pass_transistor_list) delete pass_transistor_list;
	if(left_transistor_list) delete left_transistor_list;
}

char *
Group::get_vdd(MosList *moshead)
{
   char *vdd;
   MosList *curr = NULL;
   for(curr=moshead; curr!=NULL; curr=curr->next)
   {
        if(curr->nptype)
        {
          vdd = NLAlloc(char,strlen(curr->b));
          strcpy(vdd, curr->b);
          break;
        }
   }
   return vdd;
}

char *
Group::get_gnd(MosList *moshead)
{
   char *gnd;
   MosList *curr = NULL;
   for(curr=moshead; curr!=NULL; curr=curr->next)
   {
        if(!curr->nptype)
        {
          gnd = NLAlloc(char,strlen(curr->b));
          strcpy(gnd, curr->b);
          break;
        }
   }
   return gnd;
}

int
Group::query_n_branch(char *name)
{
   int exist_flag = 0;
   if(n_branch_list==NULL)
	return 0;
   for(Branch *cur_branch=n_branch_list; cur_branch!=NULL; cur_branch=cur_branch->next)
   {
	for(MosList *cur_mos=cur_branch->mos_list; cur_mos!=NULL; cur_mos=cur_mos->next)
	{
	   if(!strcmp(cur_mos->mosname, name))
	   {
		exist_flag = 1;
		return 1;
	   }
	}
   }
   
   if(!exist_flag)
	return 0;
}

int
Group::query_p_branch(char *name)
{
   int exist_flag = 0;
   if(p_branch_list==NULL)
	return 0;
   for(Branch *cur_branch=p_branch_list; cur_branch!=NULL; cur_branch=cur_branch->next)
   {
        for(MosList *cur_mos=cur_branch->mos_list; cur_mos!=NULL; cur_mos=cur_mos->next)
        {
           if(!strcmp(cur_mos->mosname, name))
	   {
              exist_flag = 1;
	      return 1;
	   }
        }
   }

   if(!exist_flag)
        return 0;
}

Branch *
Group::insert_branch(Branch *head, Branch *new_branch)
{
   Branch *current;
   if(head==NULL)
   {
	head = new_branch;
	head->next = NULL;
   }
   else
   {
	current = new_branch;
	current->next = head->next;
	head->next = current;
   }
   return head;
}

void
Group::generate_branch_list()
{
   char *gnd;
   gnd = get_gnd(MosHead);
   char *vdd;
   vdd = get_vdd(MosHead);

   NodeList *gnd_node, *vdd_node;

   for(NodeList *curr_node=NodeHead; curr_node!=NULL; curr_node=curr_node->next)
   {
	if(!strcmp(curr_node->nodename, gnd))
	   gnd_node = curr_node;
	if(!strcmp(curr_node->nodename, vdd))
	   vdd_node = curr_node;
   }


   for(DeviceList *cur_device=gnd_node->devicelist; cur_device!=NULL; cur_device=cur_device->next)
   {
//	printf("%s \n", cur_device->devicename);
	MosList *first_n_mos;
	if(!query_n_branch(cur_device->devicename))	
	{
	   Branch *new_branch;
	   new_branch = new Branch(0);	

	   for(MosList *cur_m=MosHead; cur_m!=NULL; cur_m=cur_m->next)
	   {
		if(!strcmp(cur_device->devicename, cur_m->mosname))
		{
		   first_n_mos = cur_m;
//		   printf("A new branch from this mos: %s \n", cur_device->devicename);
		   break;
		}
	   }

	   new_branch->generate_branch(first_n_mos);
/*
	   for(MosList *cur_b=new_branch->mos_list; cur_b!=NULL; cur_b=cur_b->next)
		printf("%s  ", cur_b->mosname);
	   printf("\n");	
*/
	   n_branch_list = insert_branch(n_branch_list, new_branch);
	}
   }

   for(DeviceList *cur_device=vdd_node->devicelist; cur_device!=NULL; cur_device=cur_device->next)
   {
        MosList *first_p_mos;
        if(!query_p_branch(cur_device->devicename))
        {
           Branch *new_branch;
           new_branch = new Branch(1);

           for(MosList *cur_m=MosHead; cur_m!=NULL; cur_m=cur_m->next)
           {
                if(!strcmp(cur_device->devicename, cur_m->mosname))
                {
                   first_p_mos = cur_m;
//	           printf("A new branch from this mos: %s \n", cur_device->devicename);
                   break;
                }
           }

           new_branch->generate_branch(first_p_mos);
/*
           for(MosList *cur_b=new_branch->mos_list; cur_b!=NULL; cur_b=cur_b->next)
                printf("%s  ", cur_b->mosname);
           printf("\n");
*/
           p_branch_list = insert_branch(p_branch_list, new_branch);
	}
   }
}

void
Group::group_branch_list()
{
   if(p_branch_list->next != NULL)
   {
     for(Branch *cur_branch=p_branch_list; cur_branch!=NULL; cur_branch=cur_branch->next)
     {
//		MosList *cu=cur_branch->mos_list;
//	        printf("Deal with: %s \n",cu->mosname); 
	Branch *curr, *prev, *temp;
	prev=cur_branch;
	curr=cur_branch->next;	
	while(curr!=NULL)
	{
	   if(!strcmp(curr->pn_node, cur_branch->pn_node))
	   {
		for(MosList *cur_mos=curr->mos_list; cur_mos!=NULL; cur_mos=cur_mos->next)
		   cur_branch->mos_list = cur_branch->insert_mos(cur_branch->mos_list,cur_mos);
		for(BranchNodeList *cur_node=curr->node_list; cur_node!=NULL; cur_node=cur_node->next)
		   cur_branch->node_list = cur_branch->insert_node(cur_branch->node_list,cur_node->nodename);
		prev->next = curr->next;
		curr = prev->next;
/*
		temp = prev->next;
		prev = temp;
		curr = temp->next;
*/
	   }
	   else
	   {
	   prev=curr;
	   curr=curr->next;
	   }
	}
/*
	for(Branch *cur_br=p_branch_list; cur_br!=NULL; cur_br=cur_br->next)
	{
		MosList *cu1=cur_br->mos_list;
	        printf("%s ",cu1->mosname); 
	}
	printf("\n");
*/
      }
   }



   if(n_branch_list->next != NULL)
   {
     for(Branch *cur_branch=n_branch_list; cur_branch!=NULL; cur_branch=cur_branch->next)
     {
        Branch *curr, *prev, *temp;
        prev=cur_branch;
        curr=cur_branch->next;
        while(curr!=NULL)
        {
           if(!strcmp(curr->pn_node, cur_branch->pn_node))
           {
                for(MosList *cur_mos=curr->mos_list; cur_mos!=NULL; cur_mos=cur_mos->next)
                   cur_branch->mos_list = cur_branch->insert_mos(cur_branch->mos_list,cur_mos);
		for(BranchNodeList *cur_node=curr->node_list; cur_node!=NULL; cur_node=cur_node->next)
                   cur_branch->node_list = cur_branch->insert_node(cur_branch->node_list,cur_node->nodename);
                prev->next = curr->next;
		curr = prev->next;
/*
                temp = prev->next;
                prev = temp;
                curr = temp->next;
*/
           }
           else
           {
           prev=curr;
           curr=curr->next;
           }
        }
/*
        for(Branch *cur_br=n_branch_list; cur_br!=NULL; cur_br=cur_br->next)
        {
                MosList *cu1=cur_br->mos_list;
                printf("%s ",cu1->mosname);
        }
        printf("\n");
*/
      } 
   }    
}

MosList *
Group::insert_mos(MosList *head, MosList *mos)
{
   MosList *current;
   if(head==NULL)
   {
        head = NLAlloc(MosList,1);
        head->mosname = NLAlloc(char,strlen(mos->mosname));
        strcpy(head->mosname,mos->mosname);
        head->d = NLAlloc(char,strlen(mos->d));
        strcpy(head->d,mos->d);
        head->g = NLAlloc(char,strlen(mos->g));
        strcpy(head->g,mos->g);
        head->s = NLAlloc(char,strlen(mos->s));
        strcpy(head->s,mos->s);
        head->nptype = mos->nptype;
	head->is_instead = mos->is_instead;
        head->b = NLAlloc(char,strlen(mos->b));
        strcpy(head->b,mos->b);
        head->modelname = NLAlloc(char,strlen(mos->modelname));
        strcpy(head->modelname,mos->modelname);
        head->lvalue = mos->lvalue;
        head->wvalue = mos->wvalue;
        head->next = NULL;
   }
   else
   {
        current = NLAlloc(MosList,1);
        current->mosname = NLAlloc(char, strlen(mos->mosname));
        strcpy(current->mosname,mos->mosname);
        current->d = NLAlloc(char, strlen(mos->d));
        strcpy(current->d,mos->d);
        current->g = NLAlloc(char, strlen(mos->g));
        strcpy(current->g,mos->g);
        current->s = NLAlloc(char, strlen(mos->s));
        strcpy(current->s,mos->s);
        current->nptype = mos->nptype;
	current->is_instead = mos->is_instead;
        current->b = NLAlloc(char, strlen(mos->b));
        strcpy(current->b,mos->b);
        current->modelname = NLAlloc(char, strlen(mos->modelname));
        strcpy(current->modelname,mos->modelname);
        current->lvalue = mos->lvalue;
        current->wvalue = mos->wvalue;
        current->next = head->next;
        head->next = current;
   }
   return head;
}


void
Group::exchange_d_s(MosList *mos)
{
    char *temp=NLAlloc(char, strlen(mos->d));
    strcpy(temp, mos->d);
    mos->d = mos->s;
    mos->s = temp;
}

int
Group::query_is_pnnode(char *name)
{
    for(NodeList *cur_n=NodeHead; cur_n!=NULL; cur_n=cur_n->next)
    {
	if(!strcmp(cur_n->nodename,name))
	{
	    return cur_n->pn_node;
	}
    }
}

int
Group::query_is_gnode(char *name)
{
    int p=0;
    for(NodeList *cur_n=NodeHead; cur_n!=NULL; cur_n=cur_n->next)
    {
        if(!strcmp(cur_n->nodename,name))
        {
            for(DeviceList *cur_d=cur_n->devicelist; cur_d!=NULL; cur_d=cur_d->next)
            {
                if(cur_d->pintype==GP)
                {
                    p=1;
                }
            }
        }
    }
    return p;
}


void
Group::generate_pass_left_list()
{
    int is_pn_1, is_g_1;
    int is_pn_2, is_g_2;

    for(MosList *curr_mos=MosHead; curr_mos!=NULL; curr_mos=curr_mos->next)
    {
	if(!query_n_branch(curr_mos->mosname) && !query_p_branch(curr_mos->mosname))
	{
	    if((query_is_pnnode(curr_mos->d) && query_is_gnode(curr_mos->s)) || (query_is_pnnode(curr_mos->s) && query_is_gnode(curr_mos->d)))
		pass_transistor_list = insert_mos(pass_transistor_list,curr_mos);
            else
                left_transistor_list = insert_mos(left_transistor_list, curr_mos);
	}
    }	
}

/*
	    printf("%s  ", curr_mos->mosname);
	    for(NodeList *curr_node=NodeHead; curr_node!=NULL; curr_node=curr_node->next)
	    {
		if(!strcmp(curr_mos->d, curr_node->nodename))
		{
		    if(curr_node->pn_node)
		    {
			for(NodeList *curr_node2=NodeHead; curr_node2!=NULL; curr_node2=curr_node2->next)
			{
			    if(!strcmp(curr_mos->s, curr_node2->nodename))
			    {
				DeviceList *temp=curr_node2->devicelist;
				if(temp->next==NULL && temp->pintype==GP)
					pass_transistor_list = insert_mos(pass_transistor_list,curr_mos);
				else
					left_transistor_list = insert_mos(left_transistor_list, curr_mos);
				break;
			    }
			}
		    }
		    else
			left_transistor_list = insert_mos(left_transistor_list, curr_mos);
		    break;
		}

		if(!strcmp(curr_mos->s, curr_node->nodename))
                {
                    if(curr_node->pn_node)
                    {
                        for(NodeList *curr_node3=NodeHead; curr_node3!=NULL; curr_node3=curr_node3->next)
                        {
                            if(!strcmp(curr_mos->d, curr_node3->nodename))
                            {
                                DeviceList *temp=curr_node3->devicelist;
                                if(temp->next==NULL && temp->pintype==GP)
				{
					exchange_d_s(curr_mos);
                                        pass_transistor_list = insert_mos(pass_transistor_list,curr_mos);
				}
                                else
                                        left_transistor_list = insert_mos(left_transistor_list, curr_mos);
                                break;
                            }
                        }
                    }
                    else
                        left_transistor_list = insert_mos(left_transistor_list, curr_mos);
                    break;
                }

	    }
	}
	else
	    left_transistor_list = insert_mos(left_transistor_list, curr_mos);
    }
}
*/

TransList *
Group::insert_trans(TransList *head, TransType type, char *input, char *output, char *p_g, char *n_g)
{
    TransList *current;
    if(head==NULL)
    {
	head = NLAlloc(TransList, 1);
	head->type = type;
	head->input = NLAlloc(char, strlen(input));
	strcpy(head->input, input);
	head->output = NLAlloc(char, strlen(output));
	strcpy(head->output, output);
	head->p_g = NLAlloc(char, strlen(p_g));
	strcpy(head->p_g, p_g);
	head->n_g = NLAlloc(char, strlen(n_g));
	strcpy(head->n_g, n_g);
	head->next = NULL;
    }
    else
    {
	current = NLAlloc(TransList,1);
	current->type = type;
	current->input = NLAlloc(char, strlen(input));
	strcpy(current->input, input);
	current->output = NLAlloc(char, strlen(output));
	strcpy(current->output, output);
	current->p_g = NLAlloc(char, strlen(p_g));
	strcpy(current->p_g, p_g);
	current->n_g = NLAlloc(char, strlen(n_g));	
	strcpy(current->n_g, n_g);
	current->next = head->next;
	head->next = current;
    }
    return head;
}

void
Group::generate_trans_list()
{
    for(MosList *curr_mos=pass_transistor_list; curr_mos!=NULL; curr_mos=curr_mos->next)
    {
	int flag = 0;
        MosList *curr, *prev;
        prev=curr_mos;
        curr=curr_mos->next;
	while(curr!=NULL)
        {
	    if((!strcmp(curr_mos->d, curr->d) && !strcmp(curr_mos->s, curr->s)) || (!strcmp(curr_mos->d, curr->s) && !strcmp(curr_mos->s,curr->d)))
	    {
		flag = 1;
		if(curr_mos->nptype==1 && curr->nptype==0)
		    TransHead = insert_trans(TransHead, DOUBLE, curr_mos->d, curr_mos->s, curr_mos->g, curr->g);
		if(curr_mos->nptype==0 && curr->nptype==1)
		    TransHead = insert_trans(TransHead, DOUBLE, curr_mos->d, curr_mos->s, curr->g, curr_mos->g);
		prev->next = curr->next;
                curr = prev->next;
                break;
	    }
	    else
	    {
		prev=curr;
                curr=curr->next;
            }
	}
	if(!flag)
	{
	    TransHead = insert_trans(TransHead, SINGLE, curr_mos->d, curr_mos->s, curr_mos->g, curr->g);
	}
    }
}
	    
int
Group::counter(Branch *branch)
{
   int num = 0;
   for(MosList *cur_mos=branch->mos_list; cur_mos!=NULL; cur_mos=cur_mos->next)
   {
	num = num+1;
   }
   return num;
}

char *
Group::generate_add(char g1, char g2)
{
    char *new_string = NLAlloc(char, 6);
    strcpy(new_string, "a | b");
    new_string[0] = g1;
    new_string[1] = ' ';
    new_string[2] = '|';
    new_string[3] = ' ';
    new_string[4] = g2;
    new_string[5] = '\0';
    return new_string;
}

char *
Group::generate_add(char *g1, char g2)
{
    char *new_string = NLAlloc(char, strlen(g1)+4);
    strcpy(new_string,g1);
    int k = strlen(new_string);
    new_string[k] = ' ';
    new_string[k+1] = '|';
    new_string[k+2] = ' ';
    new_string[k+3] = g2;
    new_string[k+4] = '\0';
    return new_string;
/*
    const char *add = " | ";
    int len = strlen(g1)+4;
    char *new_string = NLAlloc(char, len);
    strcpy(new_string,g1);
    strcat(new_string, add);
    int k = strlen(new_string);
    new_string[k] = g2;
    new_string[k+1] = '\0';
    return new_string;
*/
}

char *
Group::generate_add(char *g1, char *g2)
{
    const char *add = " | ";
    int len1 = strlen(g1);
    int len2 = strlen(g2);
    char *new_string = NLAlloc(char, len1+len2+3);
    strcpy(new_string,g1);
    strcat(new_string, add);
    strcat(new_string, g2);
    return new_string;
}

char *
Group::generate_mult(char g1, char g2)
{
    char *new_string = NLAlloc(char, 6);
    strcpy(new_string, "a & b");
    new_string[0] = g1;
    new_string[1] = ' ';
    new_string[2] = '&';
    new_string[3] = ' ';
    new_string[4] = g2;
    new_string[5] = '\0';
    return new_string;
}

char *
Group::generate_mult(char *g1, char g2)
{
    char *new_string = NLAlloc(char, strlen(g1)+4);
    strcpy(new_string,g1);
    int k = strlen(new_string);
    new_string[k] = ' ';
    new_string[k+1] = '&';
    new_string[k+2] = ' ';
    new_string[k+3] = g2;
    new_string[k+4] = '\0';
    return new_string;
/*
    const char *mult = " & ";
    char *new_string = NLAlloc(char, strlen(g1)+4);
    strcpy(new_string,g1);
    strcpy(new_string,mult);
    int k = strlen(new_string);
    new_string[k] = g2;
    new_string[k+1] = '\0';
    return new_string;
*/
}

char *
Group::generate_mult(char *g1, char *g2)
{
    const char *mult = " & ";
    int len1 = strlen(g1);
    int len2 = strlen(g2);
    char *new_string = NLAlloc(char, len1+len2+3);
    strcpy(new_string,g1);
    strcat(new_string,mult);
    strcat(new_string,g2);
    return new_string;
}


char *
Group::generate_parenthes(char *g)
{
    const char *left = "(";
    const char *right = ")";
    char *new_string=NLAlloc(char,strlen(g)+2);
    strcpy(new_string,left);
    strcat(new_string, g);
    strcat(new_string, right);
    return new_string;
}

char *
Group::generate_mark(char *g)
{
    const char *mark = "~";
    char *new_string=NLAlloc(char, strlen(g)+1);
    strcpy(new_string,mark);
    strcat(new_string, g);
    return new_string;
}

int
Group::is_series_node(char *nodename, Branch *curr_branch)
{
    int flag = 0;
    int counter=0;
    MosList *mos=curr_branch->mos_list;
    for(; mos!=NULL; mos=mos->next)
    {
	if(!strcmp(nodename, mos->d) || !strcmp(nodename, mos->s))
	    counter += 1;
    }

    if(counter==2)
	flag = 1;

    return flag;
}

char 
Group::change_g(Branch *n_branch, MosList *mos1, MosList *mos2, char turn, int add_or_flag)
{
    char local = turn;

    if(mos1->is_instead && mos2->is_instead)
    {
	if(add_or_flag)
	    mos1->g = generate_add(mos1->g, mos2->g);
	else
	    mos1->g = generate_mult(mos1->g, mos2->g);
    }

    if(mos1->is_instead && !mos2->is_instead)
    {
	n_branch->name_list = n_branch->insert_name(n_branch->name_list, mos2->g, local);
	if(add_or_flag)
	    mos1->g = generate_add(mos1->g, local);
	else
	    mos1->g = generate_mult(mos1->g, local);
	local+=1;
    }

    if(!mos1->is_instead && mos2->is_instead)
    {
	n_branch->name_list = n_branch->insert_name(n_branch->name_list, mos1->g, local);
	if(add_or_flag)
	    mos1->g = generate_add(mos2->g, local);
	else
	    mos1->g = generate_mult(mos2->g, local);
	local+=1;
	mos1->is_instead = 1;
    }

    if(!mos1->is_instead && !mos2->is_instead)
    {
	n_branch->name_list = n_branch->insert_name(n_branch->name_list, mos1->g, local);
	n_branch->name_list = n_branch->insert_name(n_branch->name_list, mos2->g, local+1);
	char local1 = local+1;
	if(add_or_flag)
	    mos1->g = generate_add(local, local1);
	else
	    mos1->g = generate_mult(local, local1);
	local+=2;
	mos1->is_instead = 1;
    }
    
    return local;
}
    

char
Group::parallel_recognition(Branch *n_branch, char turn)
{
    counter_flag = 0;
    int local_counter;
    char retur=turn;
    for(MosList *curr_mos=n_branch->mos_list; curr_mos!=NULL; curr_mos=curr_mos->next)
    {
	local_counter = 0;
	int flag =0;
	MosList *curr_pointer, *prev_pointer;
	prev_pointer = curr_mos;
	curr_pointer = curr_mos->next;
	while(curr_pointer!=NULL)
	{
	    if((!strcmp(curr_mos->d, curr_pointer->d) && !strcmp(curr_mos->s, curr_pointer->s)) || (!strcmp(curr_mos->d, curr_pointer->s) && !strcmp(curr_mos->s, curr_pointer->d)))
	    {
		local_counter +=1;
		flag = 1;

		retur = change_g(n_branch, curr_mos, curr_pointer, retur, 1);

		prev_pointer->next = curr_pointer->next;
		curr_pointer = prev_pointer->next;
	    }
	    else
	    {
		prev_pointer=curr_pointer;
		curr_pointer=curr_pointer->next;
	    }
	}
	if(flag){
	curr_mos->g=generate_parenthes(curr_mos->g);
	}
	if(local_counter)
  	    counter_flag = counter_flag + local_counter + 1;
    }
    return retur;
}

char
Group::series_recognition(Branch *n_branch, char turn)
{
    counter_flag = 0;
    int local_counter = 0;
    char *GND;
    GND = get_gnd(MosHead);
    char retur=turn;
    for(MosList *curr_mos=n_branch->mos_list; curr_mos!=NULL; curr_mos=curr_mos->next)
    {
	local_counter = 0;
        int flag =0;
        MosList *curr_pointer, *prev_pointer;
        prev_pointer = curr_mos;
        curr_pointer = curr_mos->next;
        while(curr_pointer!=NULL)
        {
	    if((!strcmp(curr_mos->d, curr_pointer->d) && strcmp(curr_mos->d, GND) && strcmp(curr_mos->d, n_branch->pn_node) && strcmp(curr_mos->s, curr_pointer->s) && is_series_node(curr_mos->d,n_branch)) || (!strcmp(curr_mos->s, curr_pointer->s) && strcmp(curr_mos->s, GND) && strcmp(curr_mos->s, n_branch->pn_node) && strcmp(curr_mos->d, curr_pointer->d) && is_series_node(curr_mos->s, n_branch)) || (!strcmp(curr_mos->d, curr_pointer->s) && strcmp(curr_mos->d, GND) && strcmp(curr_mos->d, n_branch->pn_node) && strcmp(curr_mos->s, curr_pointer->d) && is_series_node(curr_mos->d, n_branch)) || (!strcmp(curr_mos->s, curr_pointer->d) && strcmp(curr_mos->s, GND) && strcmp(curr_mos->s, n_branch->pn_node) && is_series_node(curr_mos->s, n_branch) && strcmp(curr_mos->d, curr_pointer->s)))	
            {
		local_counter+=1;
		flag=1;
	        if(!strcmp(curr_mos->d, curr_pointer->d) && strcmp(curr_mos->s, curr_pointer->s))
		     strcpy(curr_mos->d, curr_pointer->s);
	        if(!strcmp(curr_mos->s, curr_pointer->s) && strcmp(curr_mos->d, curr_pointer->d))
		     strcpy(curr_mos->s, curr_pointer->d);
	        if(!strcmp(curr_mos->d, curr_pointer->s) && strcmp(curr_mos->s, curr_pointer->d)) 
	       	     strcpy(curr_mos->d, curr_pointer->d);
	        if(!strcmp(curr_mos->s, curr_pointer->d) && strcmp(curr_mos->d, curr_pointer->s))
		     strcpy(curr_mos->s, curr_pointer->s);

		retur = change_g(n_branch, curr_mos, curr_pointer, retur, 0);

                prev_pointer->next = curr_pointer->next;
                curr_pointer = prev_pointer->next;
            }
            else
            {
                prev_pointer=curr_pointer;
                curr_pointer=curr_pointer->next;
            }
        }
	if(flag){
	curr_mos->g=generate_parenthes(curr_mos->g);
	}
	if(local_counter)
  	    counter_flag = counter_flag + local_counter + 1;
    }
    return retur;
}

void
Group::branch_recognition()
{
    const char *model_name = "model_";
    const char *inv = "INV";
    char model_add = 'a';
    for(Branch *curr_n=n_branch_list; curr_n!=NULL; curr_n=curr_n->next)
    {
	int *decide = new int[4];
	for(int i=0; i<4; ++i)
	     decide[i]=0; 
        for(Branch *curr_p=p_branch_list; curr_p!=NULL; curr_p=curr_p->next)
	{
	    if(!strcmp(curr_p->pn_node, curr_n->pn_node))
	    {
		if(counter(curr_p)!=counter(curr_n))
		{
 		    curr_p->success = 0;
		    curr_n->success = 0;
   		}
		else
   		{
	   	    curr_p->success = 1;
	   	    curr_n->success = 1;
		    if(counter(curr_n) == 1)
		    {
			MosList *mos;
			mos=curr_n->mos_list;
			curr_n->name_list = curr_n->insert_name(curr_n->name_list, mos->g, 'a');
			mos->g=generate_mark("a");	
			curr_n->headername = NLAlloc(char, strlen(inv));
			strcpy(curr_n->headername, inv);
		    }
		    else
		    {
			char input_name;
			input_name='a';
			MosList *list=curr_n->mos_list;
			int counter_number = 0;
			while(counter(curr_n)!=1)
			{
			    input_name = parallel_recognition(curr_n,input_name); 
				if(counter_number==0)
				    decide[0]=counter_flag;
				else if(counter_number==1)
				    decide[2]=counter_flag;
				else
					;

			    input_name = series_recognition(curr_n, input_name);
				if(counter_number==0)
				    decide[1]=counter_flag;
				else if(counter_number==1)
				    decide[3]=counter_flag;
				else
					;
			}
			list->g = generate_mark(list->g);
		    }
		}
                if(counter(curr_n) == 2)
                {
                    if(decide[0]==2 && decide[1]==0)
                    {
                        curr_n->headername = NLAlloc(char, 6);
                        strcpy(curr_n->headername, "NOR2");
                    }
                    else if(decide[0]==0 && decide[1]==2)
                    {
                        curr_n->headername = NLAlloc(char, 6);
                        strcpy(curr_n->headername, "NAND2");
                    }
                }
                else if(counter(curr_n) == 3)
                {
                    if(decide[0]==3 && decide[1]==0)
                    {
                        curr_n->headername = NLAlloc(char, 6);
                        strcpy(curr_n->headername, "NOR3");
                    }
                    else if(decide[0]==0 && decide[1]==3)
                    {
                        curr_n->headername = NLAlloc(char, 6);
                        strcpy(curr_n->headername, "NAND3");
                    }
                    else if(decide[0]==2 && decide[1]==2)
                    {
                        curr_n->headername = NLAlloc(char, 6);
                        strcpy(curr_n->headername, "OAI12");
                    }
                    else if(decide[0]==0 && decide[1]==2 && decide[2]==2)
                    {
                        curr_n->headername = NLAlloc(char, 6);
                        strcpy(curr_n->headername, "AOI12");
                    }
                    else
                        ;
                }
                else if(counter(curr_n) == 4)
                {
                    if(decide[0]==4 && decide[1]==0 && decide[2]==0 && decide[3]==0)
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "NOR4");
                    }
                    else if(decide[0]==0 && decide[1]==4 && decide[2]==0 && decide[3]==0)
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "NAND4");
                    }
                    else if(decide[0]==0 && decide[1]==4 && decide[2]==2 && decide[3]==0)
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "AOI22");
                    }
                    else if(decide[0]==4 && decide[1]==2 && decide[2]==0 && decide[3]==0)
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "OAI22");
                    }
                    else if(decide[0]==0 && decide[1]==3 && decide[2]==2 && decide[3]==0)
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "AOI13");
                    }
                    else if(decide[0]==3 && decide[1]==2 && decide[2]==0 && decide[3]==0)
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "OAI13");
                    }
                    else if(decide[0]==2 && decide[1]==2 && decide[2]==2 && decide[3]==0)
                    {
                        MosList *mos;
                        mos=curr_n->mos_list;
                        if(!strcmp(mos->g,"~(((a | b) & c) | d)"))
                        {
                            curr_n->headername = NLAlloc(char, 8);
                            strcpy(curr_n->headername, "OAOI112");
                        }
                        else
                        {
                            curr_n->headername = NLAlloc(char, 8);
                            strcpy(curr_n->headername, "AOOI112");
                        }
                    }
                    else if(decide[0]==2 && decide[1]==3 && decide[2]==0 && decide[3]==0)
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "OAAI112");
                    }
                    else if(decide[0]==0 && decide[1]==2 && decide[2]==2 && decide[3]==3)
                    {
                        curr_n->headername = NLAlloc(char, 8);
                        strcpy(curr_n->headername, "AOAI112");
                    }
                    else
                        ;
                }
                else if(counter(curr_n)>4)
                {
                    curr_n->greatthan4 = 1;
                    curr_n->headername = NLAlloc(char, strlen(model_name)+2);
                    strcpy(curr_n->headername, model_name);
                    int k = strlen(curr_n->headername);
                    curr_n->headername[k]=model_add;
                    curr_n->headername[k+1]='\0';
                    model_add+=1;
                }
                else ;
	    }
	    break;
	}


	printf("begin:\n");
	printf("%s\n", curr_n->pn_node);
	MosList *mos=curr_n->mos_list;
	printf("%s->\n", mos->g);
	printf("%d, %d, %d, %d\n", decide[0],decide[1],decide[2],decide[3]);
	printf("end:\n");
	printf("\n");
    }
}
				
DffList *
Group::insert_dff(DffList *head, char *dffname, char *d, char *q, char *clk, char *rst)
{
    DffList *current;
    if(head==NULL)
    {
	head = NLAlloc(DffList,1);
	head->dffname = NLAlloc(char, strlen(dffname));
	strcpy(head->dffname, dffname);
	head->d = NLAlloc(char, strlen(d));
	strcpy(head->d, d);
	head->q = NLAlloc(char, strlen(q));	
	strcpy(head->q, q);
	head->clk = NLAlloc(char, strlen(clk));
	strcpy(head->clk, clk);
	head->rst = NLAlloc(char, strlen(rst));
	strcpy(head->rst, rst);
	head->next = NULL;
    }
    else
    {
	current= NLAlloc(DffList,1);
        current->dffname = NLAlloc(char, strlen(dffname));
        strcpy(current->dffname, dffname);
        current->d = NLAlloc(char, strlen(d));
        strcpy(current->d, d);
        current->q = NLAlloc(char, strlen(q));     
        strcpy(current->q, q);
        current->clk = NLAlloc(char, strlen(clk));
        strcpy(current->clk, clk);
        current->rst = NLAlloc(char, strlen(rst));
        strcpy(current->rst, rst);
	current->next = head->next;
	head->next = current;
    }
    return head;
}

char *
Group::find_module(char *node, char *module)
{
    char *flag=NULL;
    for(Branch *curr=n_branch_list; curr!=NULL; curr=curr->next)
    {
	if(!strcmp(curr->headername,module))
	{
	    for(NameList *curr_name=curr->name_list; curr_name!=NULL; curr_name=curr_name->next)
	    {
	        if(!strcmp(curr_name->gatename, node)) 
	        {
		    flag = NLAlloc(char, strlen(curr->pn_node));
		    strcpy(flag, curr->pn_node);
		    curr->ready_to_delete = 1;
	        }
		else
		{
		    reset_signal = NLAlloc(char, strlen(curr_name->gatename));
		    strcpy(reset_signal, curr_name->gatename);
		}
	    }
	}
    }

    if(flag==NULL)
	return NULL;
    else
	return flag;
}	

char *
Group::find_inv(char *node)
{
    char *flag=NULL;
    for(Branch *curr=n_branch_list; curr!=NULL; curr=curr->next)
    {
        if(!strcmp(curr->headername,"INV"))
        {
                NameList *first=curr->name_list;
                if(!strcmp(first->gatename, node))
                {
                    flag = NLAlloc(char, strlen(curr->pn_node));
                    strcpy(flag, curr->pn_node);
		    curr->ready_to_delete=1;
                }
        }
    }

    if(flag==NULL)
        return NULL;
    else
        return flag;
}

TransList *
Group::find_trans(char *node)
{
    TransList *ret;
    int flag = 0;
    for(TransList *curr_t=TransHead; curr_t!=NULL; curr_t=curr_t->next)
    {
	if(!strcmp(curr_t->input, node))
	{
	    flag=1;
	    ret = curr_t;
	    curr_t->ready_to_delete=1;
	    break;
	}
    }
    if(flag)
	return ret;
    else
	return NULL;
}

int
Group::query_module(char *node1, char *node2, char *module)
{
    int ret_flag = 0;
    int flag = 0;
    for(Branch *curr=n_branch_list; curr!=NULL; curr=curr->next)
    {
	if(!strcmp(curr->headername, module))
	{
	   if(!strcmp(curr->pn_node, node1)) 
		flag += 1;
	   for(NameList *curr_name=curr->name_list; curr_name!=NULL; curr_name=curr_name->next)
	   {
		if(!strcmp(curr_name->gatename, node2))
                    flag += 1;
	   }
	   if(flag==2)
	   {
		ret_flag = 1;
		curr->ready_to_delete=1;
		break;
	   }
	}
    }

    return ret_flag;
}

char *
Group::backtrace_inv(char *node)
{
    char *flag = NULL;
    for(Branch *curr=n_branch_list; curr!=NULL; curr=curr->next)
    {
	if(!strcmp(curr->headername, "INV") && !strcmp(curr->pn_node, node))
	{
	    NameList *first = curr->name_list;
	    flag = NLAlloc(char, strlen(first->gatename));
	    strcpy(flag, first->gatename);
	    curr->ready_to_delete = 1;
	    break;
	}
    }
    if(flag==NULL)
	return NULL;
    else
	return flag;
}

void
Group::delete_branch()
{
    Branch *curr, *prev;
    curr=n_branch_list;
    prev=n_branch_list;
    while(curr!=NULL)
    {
	if(curr->ready_to_delete)
	{
	    if(curr==n_branch_list)
	    {
		prev=curr->next;
		curr=curr->next;
		n_branch_list = curr;
	    }
	    else
	    {
		prev->next = curr->next;
		curr = prev->next;
	    }
	}
	else
	{
	    prev=curr;
	    curr=curr->next;
	}
    }
}
		
void
Group::delete_trans()
{
    TransList *curr, *prev;
    curr=TransHead;
    prev=TransHead;
    while(curr!=NULL)
    {
	if(curr->ready_to_delete)
	{
	    if(curr==TransHead)
	    {
		prev=curr->next;
		curr=curr->next;
		TransHead = curr;
	    }
	    else
	    {
		prev->next = curr->next;
		curr = prev->next;
	    }
	}
	else
	{
	    prev=curr;
	    curr=curr->next;
	}
    }
}

void
Group::recover_branch()
{
    for(Branch *curr=n_branch_list; curr!=NULL; curr=curr->next)
    {
	if(curr->ready_to_delete == 1)
	    curr->ready_to_delete = 0;
    }
}

void
Group::recover_trans()
{
    for(TransList *curr=TransHead; curr!=NULL; curr=curr->next)
    {
	if(curr->ready_to_delete == 1)
	    curr->ready_to_delete = 0;
    }
}

void
Group::dff_recognization_dff12()
{
    int success=0;
    char *node1, *node2, *node3, *node4;
    char *trans1_p, *trans1_n, *trans2_p, *trans2_n;
    char *rst, *q, *d, *clk;
    for(TransList *curr_trans=TransHead; curr_trans!=NULL; curr_trans=curr_trans->next)
    {
	if(curr_trans->type==DOUBLE)				//Search from a dual-trans
	{
	    d = NLAlloc(char, strlen(curr_trans->input));
	    strcpy(d, curr_trans->input);
	    trans1_p = NLAlloc(char, strlen(curr_trans->p_g));
	    strcpy(trans1_p, curr_trans->p_g);
	    trans1_n = NLAlloc(char, strlen(curr_trans->n_g));
	    strcpy(trans1_n, curr_trans->n_g);
	    node1 = NLAlloc(char, strlen(curr_trans->output));
	    strcpy(node1, curr_trans->output);
	    node2 = NLAlloc(char, strlen(find_inv(node1)));	//Find a INV after the dual-trans
	    strcpy(node2, find_inv(node1));
	    if(node2!=NULL)
	    {
		curr_trans->ready_to_delete = 1;
		if(query_module(node1, node2, "OAI12"))		//OAI12 exist between node1, 2
		{
		    TransList *trnow=find_trans(node2);
		    if(trnow!=NULL)
		    {
	    	        node3 = NLAlloc(char, strlen(trnow->output)); //Dual exist after node2
	    	        strcpy(node3, trnow->output);
	    		trans2_p = NLAlloc(char, strlen(trnow->p_g));
	    		strcpy(trans2_p, trnow->p_g);
	    		trans2_n = NLAlloc(char, strlen(trnow->n_g));
	    		strcpy(trans2_n, trnow->n_g);
			node4=NLAlloc(char, strlen(find_module(node3,"NOR2"))); //Find a NOR2
			strcpy(node4, find_module(node3, "NOR2"));
			if(node4!=NULL)
			{
			    rst = NLAlloc(char, strlen(reset_signal));
			    strcpy(rst, reset_signal);
			    if(query_module(node3,node4,"NAND2"))  //NAND2 exist between node3,4
			    {
				q=NLAlloc(char, strlen(find_inv(node3)));  //Find a INV at q
				strcpy(q, find_inv(node3));
				if(q!=NULL)
				{
				    if(!strcmp(trans1_n, backtrace_inv(trans1_p)) && !strcmp(trans2_n, backtrace_inv(trans1_n)) && !strcmp(trans2_p, backtrace_inv(trans2_n)))
				    {				
					clk=NLAlloc(char, strlen(backtrace_inv(trans2_p)));
					strcpy(clk, backtrace_inv(trans2_p));
					if(clk!=NULL)
					{
					    DffHead=insert_dff(DffHead,"DFF1",d,q,clk,rst);
					    success=1;
					    delete_branch();
					    delete_trans();
					}
				    }
				    if(!strcmp(trans1_p, backtrace_inv(trans1_n)) && !strcmp(trans2_p, backtrace_inv(trans1_p)) && !strcmp(trans2_n, backtrace_inv(trans2_p)))
				    {
					clk=NLAlloc(char, strlen(backtrace_inv(trans2_n)));
					strcpy(clk, backtrace_inv(trans2_n));
					if(clk!=NULL)
					{
					    DffHead=insert_dff(DffHead,"DFF2",d,q,clk,rst);
					    success=1;
					    delete_branch();
					    delete_trans();
					}
				    }
				}
			    }
			}
		    }
		}
	    }
	}
	if(!success)
	{
	    recover_branch();
	    recover_trans();
	}
    }
}
							    
void
Group::dff_recognization_dff34()
{
    int success=0;
    char *node1, *node2, *node3, *node4;
    char *trans1_p, *trans1_n, *trans2_p, *trans2_n;
    char *rst, *q, *d, *clk;
    for(TransList *curr_trans=TransHead; curr_trans!=NULL; curr_trans=curr_trans->next)
    {
        if(curr_trans->type==DOUBLE)                            //Search from a dual-trans
        {
            d = NLAlloc(char, strlen(curr_trans->input));
            strcpy(d, curr_trans->input);
            trans1_p = NLAlloc(char, strlen(curr_trans->p_g));
            strcpy(trans1_p, curr_trans->p_g);
            trans1_n = NLAlloc(char, strlen(curr_trans->n_g));
            strcpy(trans1_n, curr_trans->n_g);
            node1 = NLAlloc(char, strlen(curr_trans->output));
            strcpy(node1, curr_trans->output);
            node2 = NLAlloc(char, strlen(find_module(node1,"NAND2")));     //Find a INV after the dual-trans
            strcpy(node2, find_module(node1,"NAND2"));
            if(node2!=NULL)
            {
                rst = NLAlloc(char, strlen(reset_signal));
                strcpy(rst, reset_signal);
                curr_trans->ready_to_delete = 1;
                if(query_module(node1, node2, "NAND2"))         //NAND2 exist between node1, 2
                {
                    TransList *trnow=find_trans(node2);
                    if(trnow!=NULL)
                    {
                        node3 = NLAlloc(char, strlen(trnow->output)); //Dual exist after node2
                        strcpy(node3, trnow->output);
                        trans2_p = NLAlloc(char, strlen(trnow->p_g));
                        strcpy(trans2_p, trnow->p_g);
                        trans2_n = NLAlloc(char, strlen(trnow->n_g));
                        strcpy(trans2_n, trnow->n_g);
                        node4=NLAlloc(char, strlen(find_inv(node3))); //Find a INV
                        strcpy(node4, find_inv(node3));
                        if(node4!=NULL)
                        {
                            if(query_module(node3,node4,"NAND3"))  //NAND3 exist between node3,4
                            {
                                q=NLAlloc(char, strlen(find_inv(node3)));  //Find a INV at q
                                strcpy(q, find_inv(node3));
                                if(q!=NULL)
                                {
                                    if(!strcmp(trans1_n, backtrace_inv(trans1_p)) && !strcmp(trans2_n, backtrace_inv(trans1_n)) && !strcmp(trans2_p, backtrace_inv(trans2_n)))
                                    {
                                        clk=NLAlloc(char, strlen(backtrace_inv(trans2_p)));
                                        strcpy(clk, backtrace_inv(trans2_p));
                                        if(clk!=NULL)
                                        {
                                            DffHead=insert_dff(DffHead,"DFF3",d,q,clk,rst);
                                            success=1;
                                            delete_branch();
                                            delete_trans();
                                        }
                                    }
                                    if(!strcmp(trans1_p, backtrace_inv(trans1_n)) && !strcmp(trans2_p, backtrace_inv(trans1_p)) && !
strcmp(trans2_n, backtrace_inv(trans2_p)))
                                    {
                                        clk=NLAlloc(char, strlen(backtrace_inv(trans2_n)));
                                        strcpy(clk, backtrace_inv(trans2_n));
                                        if(clk!=NULL)
                                        {
                                            DffHead=insert_dff(DffHead,"DFF4",d,q,clk,rst);
                                            success=1;
                                            delete_branch();
                                            delete_trans();
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        if(!success)
        {
            recover_branch();
            recover_trans();
        }
    }
}

int
Group::query_inoutput(char *name)
{
    int flag = 0;
    for(InoutList *curr=InputHead; curr!=NULL; curr=curr->next)
    {
	if(!strcmp(curr->pinname, name))
	{
	    flag=1;
	    return 1;
	}
    }
    for(InoutList *curr2=OutputHead; curr2!=NULL; curr2=curr2->next)
    {
	if(!strcmp(curr2->pinname, name))
	{
	    flag=1;
	    return 1;
	}
    }

    if(!flag)
	return 0;
}

WireList *
Group::insert_wire(WireList *head, char *name)
{
    WireList *current;
    if(head==NULL)
    {
	head=NLAlloc(WireList,1);
	head->wirename=NLAlloc(char, strlen(name));
	strcpy(head->wirename,name);
	head->next = NULL;
    }
    else
    {
	current=NLAlloc(WireList,1);
	current->wirename=NLAlloc(char, strlen(name));
	strcpy(current->wirename,name);
	current->next = head->next;
	head->next = current;
    }
    return head;
}

int
Group::query_wire(char *wire)
{
    int flag=0;
    for(WireList *curr_w=WireHead; curr_w!=NULL; curr_w=curr_w->next)
    {
	if(!strcmp(curr_w->wirename, wire))
	{
	    flag=1;
	    break;
	}
    }

    if(!flag)
	return 0;
    else
	return 1;
}



